/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"

#define PIOVER180 0.0174532925f

Sith::Res::Model::Model(std::string SystemName) : Sith::VFS::Resource(SystemName)
{
	return;
};

// expects 1024 byte out
int Sith::Res::Model::GetLine(std::stringstream &buffer, char *out)
{
	// why does getline return \r when not in binary mode
/*	do
	{
		if (buffer.eof())
			return 0;
		buffer.getline(out, 1024);
	} while (out[0] == '#');
	return 1;*/
	do
	{
		if (buffer.eof())
			return 0;
		int p = -1;
		do
		{
			p++;
			out[p] = buffer.get();
		} while (out[p] != '\n' && out[p] != '\r' && out[p] != 0 && p < 1024 - 1);
		out[p] = 0;
	} while (out[0] == 0 || out[0] == '#' || out[0] == '\r' || out[0] == '\n');
	return 1;
}

int Sith::Res::Model::GetTo(std::stringstream &buffer, char *out, const char *val)
{
	if (!GetLine(buffer, out))
		return 0;
	while (stricmp(out, val) != 0)
	{
		if (!GetLine(buffer, out))
			return 0;
	}
	return 1;
}

Sith::SITHERROR Sith::Res::Model::LoadFromFile (const char* FileBuffer, std::string ColormapName)
{
	Transparency = 0;

	std::stringstream buffer(std::string(FileBuffer), std::stringstream::in);
	char line[1024];
	char temp[1024];

	if (!GetTo(buffer, line, "SECTION: HEADER"))
	{
		sprintf(temp, "ERROR: Premature end of file while searching for %s\n", "SECTION: HEADER");
		ConOut(temp, ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	GetLine(buffer, line);
	sscanf(line, "3DO %s", temp);
	Version = temp;

	if (!GetTo(buffer, line, "SECTION: MODELRESOURCE"))
	{
		sprintf(temp, "ERROR: Premature end of file while searching for %s\n", "SECTION: MODELRESOURCE");
		ConOut(temp, ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	GetLine(buffer, line);
	int materials = 0;
	sscanf(line, "MATERIALS %d", &materials);
	MaterialName.resize(materials);
	Material.resize(materials);
	for (int n = 0; n < materials; n++)
	{
		GetLine(buffer, line);
		sscanf(line, "%*d: %s", temp);
		MaterialName[n] = temp;
		Material[n] = VFS::ResourceManager::Get().GetMaterial(temp, ColormapName);
	}

	if (!GetTo(buffer, line, "SECTION: GEOMETRYDEF"))
	{
		sprintf(temp, "ERROR: Premature end of file while searching for %s\n", "SECTION: GEOMETRYDEF");
		ConOut(temp, ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	GetLine(buffer, line);
	sscanf(line, "RADIUS %f", &Radius);
	GetLine(buffer, line);
	sscanf(line, "INSERT OFFSET %f %f %f", &InsertOffset.x, &InsertOffset.y, &InsertOffset.z);
	int geosets = 0;
	GetLine(buffer, line);
	sscanf(line, "GEOSETS %d", &geosets);
	GeoSet.resize(geosets);
	for (int n = 0; n < geosets; n++)
	{
		GetLine(buffer, line);
		GetLine(buffer, line); // skip GEOSET n
		int meshes = 0;
		sscanf(line, "MESHES %d", &meshes);
		GeoSet[n].Mesh.resize(meshes);
		for (int m = 0; m < meshes; m++)
		{
			GetLine(buffer, line);
			GetLine(buffer, line); // skip MESH m
			sscanf(line, "NAME %s", temp);
			GeoSet[n].Mesh[m].Name = temp;
			GetLine(buffer, line);
			sscanf(line, "RADIUS %f", &GeoSet[n].Mesh[m].Radius);
			GetLine(buffer, line);
			sscanf(line, "GEOMETRYMODE %d", &GeoSet[n].Mesh[m].GeoMode);
			GetLine(buffer, line);
			sscanf(line, "LIGHTINGMODE %d", &GeoSet[n].Mesh[m].LightMode);
			GetLine(buffer, line);
			sscanf(line, "TEXTUREMODE %d", &GeoSet[n].Mesh[m].TexMode);

			int vertices = 0;
			GetLine(buffer, line);
			sscanf(line, "VERTICES %d", &vertices);
			GeoSet[n].Mesh[m].Vertex.resize(vertices);
			GeoSet[n].Mesh[m].VertexNormal.resize(vertices);
			for (int v = 0; v < vertices; v++)
			{
				GetLine(buffer, line);
				sscanf(line, "%*d: %f %f %f %*f", &GeoSet[n].Mesh[m].Vertex[v].x, &GeoSet[n].Mesh[m].Vertex[v].y, &GeoSet[n].Mesh[m].Vertex[v].z);
			}

			int texvertices = 0;
			GetLine(buffer, line);
			sscanf(line, "TEXTURE VERTICES %d", &texvertices);
			GeoSet[n].Mesh[m].TexVertex.resize(texvertices);
			for (int v = 0; v < texvertices; v++)
			{
				GetLine(buffer, line);
				sscanf(line, "%*d: %f %f", &GeoSet[n].Mesh[m].TexVertex[v].u, &GeoSet[n].Mesh[m].TexVertex[v].v);
			}

			GetLine(buffer, line); // skip VERTEX NORMALS
			for (int v = 0; v < vertices; v++)
			{
				GetLine(buffer, line);
				sscanf(line, "%*d: %f %f %f", &GeoSet[n].Mesh[m].VertexNormal[v].x, &GeoSet[n].Mesh[m].VertexNormal[v].y, &GeoSet[n].Mesh[m].VertexNormal[v].z);
			}

			int faces = 0;
			GetLine(buffer, line);
			sscanf(line, "FACES %d", &faces);
			GeoSet[n].Mesh[m].Face.resize(faces);
			for (int f = 0; f < faces; f++)
			{
				GetLine(buffer, line);
				int facevertices = 0;
				int pos = 0, readcount = 0;
				sscanf(line, "%*d: %d %x %d %d %d %f %d%n", &GeoSet[n].Mesh[m].Face[f].Material,
					&GeoSet[n].Mesh[m].Face[f].Type, &GeoSet[n].Mesh[m].Face[f].GeoMode,
					&GeoSet[n].Mesh[m].Face[f].LightMode, &GeoSet[n].Mesh[m].Face[f].TexMode,
					&GeoSet[n].Mesh[m].Face[f].ExtraLight, &facevertices, &pos);
				GeoSet[n].Mesh[m].Face[f].VertexList.resize(facevertices);
				GeoSet[n].Mesh[m].Face[f].TexVertexList.resize(facevertices);
				for (int v = 0; v < facevertices; v++)
				{
					sscanf(&line[pos], " %d, %d%n", &GeoSet[n].Mesh[m].Face[f].VertexList[v], &GeoSet[n].Mesh[m].Face[f].TexVertexList[v], &readcount);
					pos += readcount;
				}

				if (GeoSet[n].Mesh[m].Face[f].Type & 0x2)
					Transparency = 1;
			}
			GetLine(buffer, line); // skip FACE NORMALS
			for (int f = 0; f < faces; f++)
			{
				GetLine(buffer, line);
				sscanf(line, "%*d: %f %f %f", &GeoSet[n].Mesh[m].Face[f].Normal.x, &GeoSet[n].Mesh[m].Face[f].Normal.y, &GeoSet[n].Mesh[m].Face[f].Normal.z);
			}
		}
	}
	if (!GetTo(buffer, line, "SECTION: HIERARCHYDEF"))
	{
		sprintf(temp, "ERROR: Premature end of file while searching for %s\n", "SECTION: HIERARCHYDEF");
		ConOut(temp, ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	GetLine(buffer, line);
	int hnodes = 0;
	sscanf(line, "HIERARCHY NODES %d", &hnodes);
	Node.resize(hnodes);
	for (int n = 0; n < hnodes; n++)
	{
		GetLine(buffer, line);
		sscanf(line, "%*d: %x %x %d %d %d %d %d %f %f %f %f %f %f %f %f %f %s", &Node[n].Flags, &Node[n].Type, &Node[n].Mesh, &Node[n].Parent, &Node[n].Child, &Node[n].Sibling, &Node[n].NumChildren, &Node[n].Pos.x, &Node[n].Pos.y, &Node[n].Pos.z, &Node[n].Rot.x, &Node[n].Rot.y, &Node[n].Rot.z, &Node[n].Pivot.x, &Node[n].Pivot.y, &Node[n].Pivot.z, temp);
		Node[n].Name = temp;
	}


	// process nodes to get transform matrices
	for (int n = 0; n < hnodes; n++)
	{
		Sith::Math::Quaternion q, q2;
		Sith::Math::Matrix m, m2;

		m.Translate(Node[n].Pivot.x, Node[n].Pivot.y, Node[n].Pivot.z);

		q.Set(Math::VectorSet(1, 0, 0), Node[n].Rot.x * PIOVER180);
		q2.Set(Math::VectorSet(0, 0, 1), Node[n].Rot.y * PIOVER180);
		q *= q2;
		q2.Set(Math::VectorSet(0, 1, 0), Node[n].Rot.z * PIOVER180);
		q *= q2;
		q.Matrix(m2.m);
		m = m2 * m;

		m2.Translate(Node[n].Pos.x, Node[n].Pos.y, Node[n].Pos.z);
		m = m2 * m;

		int p = Node[n].Parent;
		while (p != -1)
		{
			q.Set(Math::VectorSet(1, 0, 0), Node[p].Rot.x * PIOVER180);
			q2.Set(Math::VectorSet(0, 0, 1), Node[p].Rot.y * PIOVER180);
			q *= q2;
			q2.Set(Math::VectorSet(0, 1, 0), Node[p].Rot.z * PIOVER180);
			q *= q2;
			q.Matrix(m2.m);
			m = m2 * m;

			m2.Translate(Node[p].Pos.x, Node[p].Pos.y, Node[p].Pos.z);
			m = m2 * m;

			p = Node[p].Parent;
		}

		Node[n].Transform = m;
	}



	// vertex arrays
	for (int g = 0; g < GeoSet.size(); g++)
	{
		int p;
		std::vector<ModelMesh>::iterator m;
		for (m = GeoSet[g].Mesh.begin(); m != GeoSet[g].Mesh.end(); ++m)
		{
			m->vertices = 0;
			for (int f = 0; f < m->Face.size(); f++)
			{
				m->vertices += m->Face[f].VertexList.size();
			}
			m->vertex = new Sith::Win32::CompiledVertex [m->vertices];

			p = 0;
			float uscale, vscale;
			for (f = 0; f < m->Face.size(); f++)
			{
				if (m->Face[f].Material >= 0)
				{
					uscale = (1.0f / Material[m->Face[f].Material]->GetHBM(0).GetGeometry().xdim) / Material[m->Face[f].Material]->xscale;
					vscale = (1.0f / Material[m->Face[f].Material]->GetHBM(0).GetGeometry().ydim) / Material[m->Face[f].Material]->yscale;
				}
				for (int v = 0; v < m->Face[f].VertexList.size(); v++)
				{
					m->vertex[p].pos[0] = m->Vertex[m->Face[f].VertexList[v]].x;
					m->vertex[p].pos[1] = m->Vertex[m->Face[f].VertexList[v]].y;
					m->vertex[p].pos[2] = m->Vertex[m->Face[f].VertexList[v]].z;
					m->vertex[p].tangent[0] = 0;
					m->vertex[p].tangent[1] = 0;
					m->vertex[p].tangent[2] = 0;
					m->vertex[p].binormal[0] = 0;
					m->vertex[p].binormal[1] = 0;
					m->vertex[p].binormal[2] = 0;
					m->vertex[p].normal[0] = m->VertexNormal[m->Face[f].VertexList[v]].x;
					m->vertex[p].normal[1] = m->VertexNormal[m->Face[f].VertexList[v]].y;
					m->vertex[p].normal[2] = m->VertexNormal[m->Face[f].VertexList[v]].z;
					if (m->TexVertex.size())
					{
						m->vertex[p].tex[0] = m->TexVertex[m->Face[f].TexVertexList[v]].u * uscale;
						m->vertex[p].tex[1] = m->TexVertex[m->Face[f].TexVertexList[v]].v * vscale;
					}
					else
					{
						m->vertex[p].tex[0] = 0;
						m->vertex[p].tex[1] = 0;
					}
					if (m->Face[f].Type & 0x2)
					{
						m->vertex[p].color[0] = 255;
						m->vertex[p].color[1] = 255;
						m->vertex[p].color[2] = 255;
						m->vertex[p].color[3] = 90;
					}
					else
					{
						m->vertex[p].color[0] = 255;
						m->vertex[p].color[1] = 255;
						m->vertex[p].color[2] = 255;
						m->vertex[p].color[3] = 255;
					}
					p++;
				}
			}

			m->drawbins = Material.size();
			m->drawsize = new int [m->drawbins];
			memset(m->drawsize, 0, sizeof(int) * m->drawbins);
			for (f = 0; f < m->Face.size(); f++)
			{
				m->drawsize[m->Face[f].Material] += (m->Face[f].VertexList.size() - 2) * 3;
				if (m->Face[f].Type & 0x1) // two-sided
					m->drawsize[m->Face[f].Material] += (m->Face[f].VertexList.size() - 2) * 3;
			}
			m->drawindex = new unsigned int* [m->drawbins];
			for (int n = 0; n < m->drawbins; n++)
				m->drawindex[n] = new unsigned int [m->drawsize[n]];

			int *dp;
			dp = new int [m->drawbins];
			memset(dp, 0, sizeof(int) * m->drawbins);
			int vp = 0, texindex;
			for (f = 0; f < m->Face.size(); f++)
			{
				texindex = m->Face[f].Material;
				if (texindex < 0 || !m->Face[f].GeoMode)
				{
					vp += m->Face[f].VertexList.size();
					continue;
				}
				for (int v = 2; v < m->Face[f].VertexList.size(); v++)
				{
					m->drawindex[texindex][dp[texindex]] = vp;
					dp[texindex]++;
					m->drawindex[texindex][dp[texindex]] = vp + v - 1;
					dp[texindex]++;
					m->drawindex[texindex][dp[texindex]] = vp + v;
					dp[texindex]++;
				}
				if (m->Face[f].Type & 0x1) // two-sided
				{
					for (v = 2; v < m->Face[f].VertexList.size(); v++)
					{
						m->drawindex[texindex][dp[texindex]] = vp + v;
						dp[texindex]++;
						m->drawindex[texindex][dp[texindex]] = vp + v - 1;
						dp[texindex]++;
						m->drawindex[texindex][dp[texindex]] = vp;
						dp[texindex]++;
					}
				}
				vp += m->Face[f].VertexList.size();
			}
			delete [] dp;

			Sith::Win32::glGenBuffersARB(1, &m->drawbuffer);
			Sith::Win32::glBindBufferARB(GL_ARRAY_BUFFER_ARB, m->drawbuffer);
			Sith::Win32::glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(Sith::Win32::CompiledVertex) * m->vertices, m->vertex, GL_STATIC_DRAW_ARB);

			m->drawindexbuffer = new unsigned int [m->drawbins];
			for (n = 0; n < m->drawbins; n++)
			{
				if (m->drawsize[n] > 0)
				{
					Sith::Win32::glGenBuffersARB(1, &m->drawindexbuffer[n]);
					Sith::Win32::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, m->drawindexbuffer[n]);
					Sith::Win32::glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof(unsigned int) * m->drawsize[n], m->drawindex[n], GL_STATIC_DRAW_ARB);
				}
				else
				{
					m->drawindexbuffer[n] = 0;
				}
			}
		}
	}

	Sith::Win32::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
	Sith::Win32::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);

	return SE_OK;
}


Sith::Res::Model::~Model()
{
	for (int g = 0; g < GeoSet.size(); g++)
	{
		std::vector<ModelMesh>::iterator m;
		for (m = GeoSet[g].Mesh.begin(); m != GeoSet[g].Mesh.end(); ++m)
		{
			if (m->vertex)
				delete [] m->vertex;
			m->vertex = 0;
			if (m->drawindex)
			{
				for (int n = 0; n < m->drawbins; n++)
					delete [] m->drawindex[n];
				delete [] m->drawindex;
			}
			m->drawindex = 0;
			if (m->drawsize)
				delete [] m->drawsize;
			m->drawsize = 0;

			if (m->drawbuffer)
				Sith::Win32::glDeleteBuffersARB(1, &m->drawbuffer);
			m->drawbuffer = 0;
			if (m->drawindexbuffer)
			{
				for (int n = 0; n < m->drawbins; n++)
					if (m->drawindexbuffer[n])
						Sith::Win32::glDeleteBuffersARB(1, &m->drawindexbuffer[n]);
				delete [] m->drawindexbuffer;
			}
			m->drawindexbuffer = 0;
		}
	}
}


Sith::VFS::WisePointer<Sith::Res::Model> Sith::VFS::ResourceManager::GetModel(std::string SystemName, std::string ColormapName)
{
	// search the Model list for a name matching this one
	for(std::list<Res::Model>::iterator it = ModelList.begin(); it != ModelList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Model>(&(*it));
		};
	};

	// still here? Not loaded yet. So load it lollers.
	ModelList.push_back(Res::Model(SystemName));

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("3do\\" + SystemName);

	if(!FileBuffer)
	{
		ModelList.pop_back();
		ConOut("WARNING: Model file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Warning);
		return GetModel("dflt.3do"); // return null pointer
	};

	if(Failed(ModelList.back().LoadFromFile(FileBuffer, ColormapName)))
	{
		delete[] FileBuffer;
		ModelList.pop_back();
		ConOut("WARNING: Failed to load Model \'" + SystemName + "\'.\n", ConOut::Warning);
		return GetModel("dflt.3do"); // return null pointer
	};

	delete[] FileBuffer;
	return VFS::WisePointer<Res::Model>(&ModelList.back());
};