/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _3DOFILE_H_INCLUDED_
#define _3DOFILE_H_INCLUDED_

#include "ogl_render_compiled.h"

namespace Sith {
	namespace Res {

		struct ModelTexCoord
		{
			float u, v;
		};

		struct ModelFace
		{
			int Material;
			int Type;
			int GeoMode;
			int LightMode;
			int TexMode;
			float ExtraLight;
			std::vector<int> VertexList;
			std::vector<int> TexVertexList;
			Math::Vector Normal;
		};

		struct ModelHierarchy
		{
			Math::Matrix Transform;

			int Flags;
			int Type;
			int Mesh;
			int Parent;
			int Child;
			int Sibling;
			int NumChildren;
			Math::Vector Pos, Rot, Pivot;
			std::string Name;
		};

		struct ModelMesh
		{
			std::string	Name;
			float		Radius;
			int			GeoMode;
			int			LightMode;
			int			TexMode;
			std::vector<Math::Vector> Vertex;
			std::vector<ModelTexCoord> TexVertex;
			std::vector<Math::Vector> VertexNormal;
			std::vector<ModelFace> Face;

			Sith::Win32::CompiledVertex *vertex;
			unsigned int **drawindex;
			int *drawsize;
			int drawbins;
			int vertices;
			unsigned int drawbuffer, *drawindexbuffer;
		};

		struct ModelGeoSet
		{
			std::vector<ModelMesh> Mesh;
		};

		class Model : public VFS::Resource {
		public:
			std::string		Version;

			std::vector<std::string>					MaterialName;
			std::vector<VFS::WisePointer<Material> >	Material;

			float			Radius;
			Math::Vector	InsertOffset;

			std::vector<ModelGeoSet>					GeoSet;

			std::vector<ModelHierarchy>					Node;


			int				Transparency; // not zero if it has transparency


			Model (std::string FileName);
			~Model();
			SITHERROR LoadFromFile (const char* FileBuffer, std::string ColormapFile="dflt.cmp");
		private:
			int GetLine(std::stringstream &buffer, char *out);
			int GetTo(std::stringstream &buffer, char *out, const char *val);
		};
	};
};

#endif