/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _ANIMCONTEXT_H_INCLUDED_
#define _ANIMCONTEXT_H_INCLUDED_

namespace Sith
{
	namespace Gameplay
	{
		class Animation : public System::Task
		{
			friend class System::KernelProc;
		private:
			Sith::Win32::Timer theTimer;

			static Animation* StaticInterface;

			class AnimContext
			{
			private:
				static int AnimCount;
				int Index;
			public:
				virtual SITHERROR Update(double dT)=0;
				AnimContext();
				inline int GetIndex()
				{
					return Index;
				};

				virtual inline int GetSurface()
				{
					return -1;
				};
			};

			class CeilingSkyAnim : public AnimContext
			{
				float dU, dV;
			public:
				CeilingSkyAnim(float u_speed, float v_speed);
				SITHERROR Update(double dT);
			};

			class HorizonSkyAnim : public AnimContext
			{
				float dU, dV;
			public:
				HorizonSkyAnim(float u_speed, float v_speed);
				SITHERROR Update(double dT);
			};

			class SurfaceSlideAnim : public AnimContext
			{
				Sith::Math::Vector Direction;
				float Speed;

				float dU, dV;

				int Surface;
			public:
				SurfaceSlideAnim(int _Surface, Sith::Math::Vector _Direction, float _Speed);
				SITHERROR Update(double dT);

				inline int GetSurface()
				{
					return Surface;
				};
			};

			class SurfaceCelAnim : public AnimContext
			{
				int Flags;
				double Frametime;
				double TimeSinceLastFrame;
				int Surface;
				int Framecount;
				int CurrentFrame;

			public:
				SurfaceCelAnim(int _Surface, float _Framerate, int _Flags);
				SITHERROR Update(double dT);

				inline int GetSurface()
				{
					return Surface;
				};
			};

			std::list<AnimContext*> AnimList;

			Animation();
			~Animation();

			SITHERROR Process();
		public:

			static inline Animation& Get()
			{
				return *StaticInterface;
			};

			SITHERROR StopAll();
			SITHERROR StopAnim(int Anim);

			// COG VM BINDINGS:
			COGLIBAPI Stop(COG::VMType* Index)
			{
				Animation::Get().StopAnim(Index->GetInt());
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI StopSurfaceAnim(COG::VMType* Surface)
			{
				for(std::list<AnimContext*>::iterator it = Get().AnimList.begin(); it != Get().AnimList.end();)
				{
					if((*it)->GetSurface() == Surface->GetInt())
					{
						Get().StopAnim((*it)->GetIndex());
						it = Get().AnimList.begin();
						continue;
					};

					++it;
				};

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI SlideCeilingSky(COG::VMType* u_speed, COG::VMType* v_speed)
			{
				Animation::Get().AnimList.push_back(new CeilingSkyAnim(u_speed->GetFloat(), v_speed->GetFloat()));
				return new COG::VMType(int(Animation::Get().AnimList.back()->GetIndex()));
			};

			COGLIBAPI SlideHorizonSky(COG::VMType* u_speed, COG::VMType* v_speed)
			{
				Animation::Get().AnimList.push_back(new HorizonSkyAnim(u_speed->GetFloat(), v_speed->GetFloat()));
				return new COG::VMType(int(Animation::Get().AnimList.back()->GetIndex()));
			};

			COGLIBAPI SlideSurface(COG::VMType* surface, COG::VMType* vec, COG::VMType* speed)
			{
				Animation::Get().AnimList.push_back(new SurfaceSlideAnim(surface->GetInt(), vec->GetVector(), speed->GetFloat()));
				return new COG::VMType(int(Animation::Get().AnimList.back()->GetIndex()));
			};

			COGLIBAPI SurfaceAnim(COG::VMType* surface, COG::VMType* framerate, COG::VMType* flags)
			{
				Animation::Get().AnimList.push_back(new SurfaceCelAnim(surface->GetInt(), framerate->GetFloat(), flags->GetInt()));
				return new COG::VMType(int(Animation::Get().AnimList.back()->GetIndex()));
			};

			COGLIBAPI SetSurfaceCel(COG::VMType* surface, COG::VMType* celnum);
			COGLIBAPI GetSurfaceCel(COG::VMType* surface);

			COGLIBAPI AddDynamicTint(COG::VMType* thing, COG::VMType* red, COG::VMType* green, COG::VMType* blue)
			{
				// TODO: Implement
				return new COG::VMType();
			};

			COGLIBAPI NewColorEffect(COG::VMType* R1, COG::VMType* G1, COG::VMType* B1, COG::VMType* R2, COG::VMType* G2, COG::VMType* B2, COG::VMType* R3, COG::VMType* G3, COG::VMType* B3, COG::VMType* K)
			{
				return new COG::VMType(1);
			};

			COGLIBAPI FreeColorEffect(COG::VMType* AnimNum)
			{
				// TODO: Something
				return new COG::VMType();
			};
		};
	};
};

#endif // _ANIMCONTEXT_H_INCLUDED_