/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "ogl_render.h"

Sith::Res::Bitmap::Bitmap (std::string FileName) : Sith::VFS::Resource(FileName) {
	Offset = 0;
	return;
}

Sith::Res::Bitmap::~Bitmap () {
	if (Images.size()) {
		std::vector<Image>::iterator it;
		for (it = Images.begin(); it != Images.end(); it++) {
			if (it->Data) {
				delete [] it->Data;
			}
			if (it->RawData) {
				delete[] it->RawData;
			}
		}
	}
	return;
}

Sith::SITHERROR Sith::Res::Bitmap::LoadFromFile (const char *FileBuffer) {
	// oh god i love binary files they are beautiful
	memcpy (&BMHeader, (FileBuffer + Offset), sizeof(Header));
	Offset += sizeof(Header);

	if (!(BMHeader.FileType[0] == 'B' && BMHeader.FileType[1] == 'M' && BMHeader.FileType[2] == ' ')) {
		ConOut ("ERROR: Invalid BM file type in header.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}

	Images.resize (BMHeader.NumImages);
	if (BMHeader.NumBits == 8) {			// 8-bit BM
		std::vector<Image>::iterator it;
		for (it = Images.begin(); it != Images.end(); it++) {
			memcpy (&(*it), FileBuffer + Offset, sizeof(unsigned int) * 2);
			Offset += sizeof(unsigned int) * 2;
			it->Data = new unsigned char[it->Width * it->Height];		// Data is an array of chars
			memcpy (it->Data, FileBuffer + Offset,  sizeof (unsigned char) * it->Width * it->Height);
			Offset +=  sizeof (unsigned char) * it->Width * it->Height;
		}
	} else if (BMHeader.NumBits == 16) {	// 16-bit BM
		std::vector<Image>::iterator it;
		for (it = Images.begin(); it != Images.end(); it++) {
			memcpy (&(*it), FileBuffer + Offset, sizeof(unsigned int) * 2);
			Offset += sizeof(unsigned int) * 2;
			it->Data = new unsigned short[it->Width * it->Height];	// Data is an array of shorts
			memcpy (it->Data, FileBuffer + Offset,  sizeof (unsigned short) * it->Width * it->Height);
			Offset +=  sizeof (unsigned short) * it->Width * it->Height;

			it->RawData = new __int32 [it->Width * it->Height];

			for (int n = 0; n < it->Width * it->Height; n++)
			{
				it->RawData[n] = (int)((float)(((unsigned short *)it->Data)[n] & 0x001F) * (255.0f / 31.0f)) & 0x000000FF
						| (int)((float)((((unsigned short *)it->Data)[n] & 0x07E0) >> 5) * (255.0f / 63.0f)) << 8 & 0x0000FF00
						| (int)((float)((((unsigned short *)it->Data)[n] & 0xF800) >> 11) * (255.0f / 31.0f)) << 16 & 0x00FF0000
						| 0xFF000000;
					// transparency
					if (((unsigned short *)it->Data)[n] == BMHeader.TransparentColor) {
						it->RawData[n] &= 0x00FFFFFF;
					}
			}
		}
	} else {
		ConOut ("ERROR: BM is neither 8- nor 16-bit.\n", ConOut::Error);
		return SE_FILE_CORRUPT;
	}

	if (BMHeader.NumBits == 8) {
		if (BMHeader.PaletteIncluded != 1) {	// if a palette is included
			memcpy (Palette, FileBuffer + Offset, sizeof(Palette));
			std::vector<Image>::iterator it;
			for (it = Images.begin(); it != Images.end(); it++) {
				it->RawData = new __int32 [it->Width * it->Height];
				for (int i = 0; i < it->Width * it->Height; i++) {
					it->RawData[i] = (Palette[((unsigned char *)it->Data)[i]].b << 24) | (Palette[((unsigned char *)it->Data)[i]].b << 16) | (Palette[((unsigned char *)it->Data)[i]].r << 8) | (((unsigned char *)it->Data)[i] == BMHeader.TransparentColor ? 0x00 : 0xFF);
					if (((unsigned char *)it->Data)[i] == BMHeader.TransparentColor) {
						it->RawData[i] &= 0x00FFFFFF;
					}
				}
			}
		} else {
			VFS::WisePointer<Res::Colormap> UIColormap = Sith::VFS::ResourceManager::Get().GetColormap ("uicolormap.cmp");
			if (!UIColormap.isValid()) {
				ConOut ("ERROR: Could not load uicolormap.cmp for BM.\n", ConOut::Error);
				return SE_FILE_CORRUPT;
			}
			std::vector<Image>::iterator it;
			for (it = Images.begin(); it != Images.end(); it++) {
				it->RawData = new __int32 [it->Width * it->Height];
				for (int i = 0; i < it->Width * it->Height; i++) {
					it->RawData[i] = UIColormap->Get(((unsigned char *)it->Data)[i]);
					if (((unsigned char *)it->Data)[i] == BMHeader.TransparentColor) {
						it->RawData[i] &= 0x00FFFFFF;
					}
				}
			}
		}
	}

	// Create the OpenGL Texture
	std::vector<Image>::iterator it;
	for (it = Images.begin(); it != Images.end(); it++) {
		glGenTextures(1, &(it->OpenGLTexture));
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, it->OpenGLTexture);
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8, it->Width, it->Height, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, it->RawData);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}

	return SE_OK;
}

