/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _BMFILE_H_INCLUDED_
#define _BMFILE_H_INCLUDED_

namespace Sith {
	namespace Res {
		class Bitmap : public VFS::Resource {
		public:
			unsigned int		Offset;

			struct Header {
				char			FileType[3];		// "BM "
				char			Version;			// $20
				int				Unknown1;			// $46
				int				Unknown2;			// 0
				int				PaletteIncluded;	// 0, 1, 2; 2 = palette included
				unsigned int	NumImages;			// number of images
				unsigned int	XOffset;
				unsigned int	YOffset;
				unsigned int	TransparentColor;
				unsigned int	Unknown3;			// 1 in 16-bit mat, 0 in 8-bit mat
				unsigned int	NumBits;			// 8 or 16
				unsigned int	BlueBits;			// 5 for 16 bit; else 0
				unsigned int	GreenBits;			// 6 for 16 bit
				unsigned int	RedBits;			// 5 for 16 bit
				unsigned int	Unknown4;			// 11 in 16-bit, else 0
				unsigned int	Unknown5;			// 5 in 16-bit
				unsigned int	Unknown6;			// 0
				unsigned int	Unknown7;			// 3 in 16-bit
				unsigned int	Unknown8;			// 2 in 16-bit
				unsigned int	Unknown9;			// 2 in 16-bit
				unsigned int	Padding[13];		// 0
			};

			struct Image {
				unsigned int	Width;
				unsigned int	Height;
				void			*Data;			// size = w*h and char if 8 bit, 2*w*h and short int if 16 bit 
				__int32			*RawData;		// raw data for opengl in BGRA
				unsigned int	OpenGLTexture;	// opengl texture index
			};

			struct PaletteEntry {
				unsigned char	r, g, b;
			};

			Header				BMHeader;
			std::vector<Image>	Images;
			PaletteEntry		Palette[256];

			SITHERROR LoadFromFile (const char* FileBuffer);
			Bitmap(std::string FileName);
			~Bitmap();
		};
	};
};

#endif