/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* COG YACC PARSER PROFILE */

%{
#include "stdafx.h"

#include "vmtype.h"
#include "opcodes.h"
#include "instruction.h"
#include "vmheap.h"
#include "stackmachine.h"
#include "coglib.h"
#include "vector.h"

#include "cog.y.hpp"

using namespace Sith::COG;

int yyerror(char *s);
int yylex(void);

extern char *yytext;

extern int COGlineno;

VMType yyVMTbuf;

std::stack<int> BackPatchStack;

int yyArgCtr = 0;
std::stack<int> RecurArgCtr;

Sith::COG::VMHeap::SymbolExtensionTable ExtTab;

extern StackMachine* GlobalStackMachine;

%}

%union { int Integer; float Float; double Double; bool Boolean; char *String; Sith::COG::VMTypeCode TypeCode; }

%token  <String> VMTT_IDENTIFIER VMTT_PLUS_IDENTIFIER
%type	<String> identifier

%token	VMTT_KEYWORD_BREAK VMTT_KEYWORD_CALL VMTT_KEYWORD_ELSE VMTT_KEYWORD_FOR VMTT_KEYWORD_FOR VMTT_KEYWORD_IF
%token	VMTT_KEYWORD_RETURN VMTT_KEYWORD_STOP VMTT_KEYWORD_WHILE VMTT_KEYWORD_DO VMTT_KEYWORD_SYMBOLS VMTT_KEYWORD_CONTINUE
%token	VMTT_KEYWORD_CODE VMTT_KEYWORD_END

%token	VMTT_KEYWORD_LOCAL VMTT_KEYWORD_DESC VMTT_KEYWORD_MASK VMTT_KEYWORD_LINKID VMTT_KEYWORD_NOLINK

%token	VMTT_KEYWORD_GETPARAM VMTT_KEYWORD_GETSENDERID VMTT_KEYWORD_GETSENDERREF VMTT_KEYWORD_GETSENDERTYPE
%token	VMTT_KEYWORD_GETSOURCEREF VMTT_KEYWORD_GETSOURCETYPE VMTT_KEYWORD_RETURNEX VMTT_KEYWORD_SLEEP

%token	<TypeCode> VMTT_TYPE_AI VMTT_TYPE_BOOL VMTT_TYPE_COG VMTT_TYPE_FLEX VMTT_TYPE_FLOAT VMTT_TYPE_INT VMTT_TYPE_KEYFRAME
%token	<TypeCode> VMTT_TYPE_MATERIAL VMTT_TYPE_MESSAGE VMTT_TYPE_MODEL VMTT_TYPE_SECTOR VMTT_TYPE_SOUND
%token	<TypeCode> VMTT_TYPE_SURFACE VMTT_TYPE_TEMPLATE VMTT_TYPE_THING VMTT_TYPE_VECTOR

%token	VMTT_PUNC_ADD VMTT_PUNC_SUB VMTT_PUNC_MUL VMTT_PUNC_DIV VMTT_PUNC_MOD VMTT_PUNC_AND VMTT_PUNC_OR VMTT_PUNC_XOR
%token	VMTT_PUNC_INC VMTT_PUNC_DEC VMTT_PUNC_SHL VMTT_PUNC_SHR VMTT_PUNC_APOS

%token	VMTT_PUNC_ASSIGN VMTT_PUNC_ASSIGN_ADD VMTT_PUNC_ASSIGN_SUB VMTT_PUNC_ASSIGN_MUL VMTT_PUNC_ASSIGN_DIV
%token	VMTT_PUNC_ASSIGN_MOD VMTT_PUNC_ASSIGN_AND VMTT_PUNC_ASSIGN_OR VMTT_PUNC_ASSIGN_XOR VMTT_PUNC_ASSIGN_SHL VMTT_PUNC_ASSIGN_SHR

%token	VMTT_PUNC_EQUALITY_EQ VMTT_PUNC_EQUALITY_NEQ VMTT_PUNC_EQUALITY_GT VMTT_PUNC_EQUALITY_LT
%token	VMTT_PUNC_EQUALITY_GTE VMTT_PUNC_EQUALITY_LTE VMTT_PUNC_EQUALITY_AND VMTT_PUNC_EQUALITY_OR VMTT_PUNC_EQUALITY_NOT

%token	VMTT_PUNC_LPAREN VMTT_PUNC_RPAREN VMTT_PUNC_LBRACKET VMTT_PUNC_RBRACKET VMTT_PUNC_LBRACE VMTT_PUNC_RBRACE

%token	VMTT_PUNC_COMMA VMTT_PUNC_COLON VMTT_PUNC_ENDSTMT

%token	VMTT_PUNC_DBLCOLON

%token	<Integer> VMTT_LITERAL_INTEGER VMTT_LITERAL_INTEGER_HEX
%type	<Integer> integer_literal
%token	<Float> VMTT_LITERAL_FLOAT
%type	<Float> float_literal vector_literal_element
%token	<Double> VMTT_LITERAL_DOUBLE
%token	<String> VMTT_LITERAL_STRING VMTT_LITERAL_FILENAME VMTT_LITERAL_TEMPLATE
%type	<String> string_literal
%token	<Boolean> VMTT_LITERAL_BOOLEAN
%type	<Boolean> boolean_literal

%type	<Integer> if_block_header if_block

%start cog_script

%%

cog_script
	: statement_list VMTT_KEYWORD_END
	| statement_list
	|
	;

/* CODE SECTION */

integer_literal
	: VMTT_LITERAL_INTEGER { $$ = $1; }
	| VMTT_LITERAL_INTEGER_HEX { $$ = $1; }
	;
	
float_literal
	: VMTT_LITERAL_FLOAT { $$ = $1; }
	| VMTT_LITERAL_DOUBLE { $$ = $1; }
	;
	
string_literal
	: VMTT_LITERAL_STRING { $$ = $1; }
	;
	
boolean_literal
	: VMTT_LITERAL_BOOLEAN { $$ = $1; }
	;
	
vector_literal_element
	: float_literal { $$ = $1; }
	| integer_literal { $$ = (float)$1; }
	;
	
literal
	: integer_literal { yyVMTbuf = VMType((int)$1); }
	| boolean_literal { yyVMTbuf = VMType((bool)$1); }
	| float_literal { yyVMTbuf = VMType((float)$1); }
	| string_literal
		{
			std::string StrTmp = std::string($1);
			std::string StrBuf;
			// Translate and output STRING LITERALS:
			int i = 1;
			while(StrTmp[i] != '\"' && StrTmp[i] != '\0')
			{
				if(StrTmp[i] == '\\')
				{
					++i;
					if(StrTmp[i] == 'n')
					{
						StrBuf.push_back('\n');
					}
					else if(StrTmp[i] == '\"')
					{
						StrBuf.push_back('\"');
					}
					else if(StrTmp[i] == 't')
					{
						StrBuf.push_back('\t');
					}
					else if(StrTmp[i] == '\\')
					{
						StrBuf.push_back('\\');
					}
					else if(StrTmp[i] == '\0')
					{
						break;
					}
					
					++i;
					continue;
				}
				else
				{
					StrBuf.push_back(StrTmp[i++]);
					continue;
				};
			};
			yyVMTbuf = VMType(StrBuf);
			free($1);
		}
	| VMTT_PUNC_APOS vector_literal_element vector_literal_element vector_literal_element VMTT_PUNC_APOS
		{
			yyVMTbuf = VMType(Sith::Math::Vector($2, $3, $4));
		}
	;
	
identifier
	: VMTT_IDENTIFIER { $$ = $1 }
	;

primary_expression
	: identifier
						{
							int HeapAddr = GlobalStackMachine->Heap.TranslateLocalHeapName($1);
							if(HeapAddr == -1)
							{
								yyerror("Unknown Identifier");
								free($1);
								return 1;
							};
							GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
							free($1);
						}
	| identifier VMTT_PUNC_DBLCOLON identifier
						{
							int Scope = GlobalStackMachine->Heap.TranslateNamespaceName($1);
							if(Scope == -1)
							{
								// Scope does not exist; add it
								Sith::COG::CogLib::InjectNamespace($1);
								Scope = GlobalStackMachine->Heap.TranslateNamespaceName($1);						
							}
							int Value = GlobalStackMachine->Heap.TranslateNamespaceHeapName(Scope, $3);
							if(Value == -1)
							{
								// Variable does not exist; add it
								std::stringstream VarName;
								VarName << $1 << "::" << $3;
								Sith::COG::CogLib::AddSetting(VarName.str(), Sith::COG::VMType(Sith::COG::VMTC_VOID));
								
								Value = GlobalStackMachine->Heap.TranslateNamespaceHeapName(Scope, $3);
							};
							GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, Value, Scope), COGlineno);
							free($1);
							free($3);
						}
	| literal { GlobalStackMachine->AddInstruction(Opcodes::PUSH, VMType(yyVMTbuf), COGlineno); }
	| VMTT_PUNC_LPAREN expression VMTT_PUNC_RPAREN
	;
	
postfix_expression
	: primary_expression
	| identifier VMTT_PUNC_LBRACKET expression VMTT_PUNC_RBRACKET
						{
							GlobalStackMachine->AddInstruction(Opcodes::ILOD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
						}
	| VMTT_KEYWORD_GETSENDERID VMTT_PUNC_LPAREN VMTT_PUNC_RPAREN
						{
							GlobalStackMachine->AddInstruction(Opcodes::SYSCALL, VMType(VMTC_SYSTEM_CALL, 2, -1), COGlineno);
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_KEYWORD_GETSENDERREF VMTT_PUNC_LPAREN VMTT_PUNC_RPAREN
						{
							GlobalStackMachine->AddInstruction(Opcodes::SYSCALL, VMType(VMTC_SYSTEM_CALL, 0, -1), COGlineno);
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_KEYWORD_GETSOURCEREF VMTT_PUNC_LPAREN VMTT_PUNC_RPAREN
						{
							GlobalStackMachine->AddInstruction(Opcodes::SYSCALL, VMType(VMTC_SYSTEM_CALL, 1, -1), COGlineno);
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_KEYWORD_GETPARAM VMTT_PUNC_LPAREN { RecurArgCtr.push(yyArgCtr); yyArgCtr = 0; } argument_expression_list VMTT_PUNC_RPAREN
						{
							if(yyArgCtr != 1)
							{
								std::ostringstream Stream;
								Stream << "WARNING: System call \'getparam\' does not take " << yyArgCtr << " arguments.\n";
								ConOut(Stream.str(), ConOut::Warning);
								YYABORT;
							};
							
							yyArgCtr = RecurArgCtr.top();
							RecurArgCtr.pop();
							
							GlobalStackMachine->AddInstruction(Opcodes::SYSCALL, VMType(VMTC_SYSTEM_CALL, 4, -1), COGlineno);
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_KEYWORD_RETURNEX VMTT_PUNC_LPAREN { RecurArgCtr.push(yyArgCtr); yyArgCtr = 0; } argument_expression_list VMTT_PUNC_RPAREN
						{
							if(yyArgCtr != 1)
							{
								std::ostringstream Stream;
								Stream << "WARNING: System call \'returnex\' does not take " << yyArgCtr << " arguments.\n";
								ConOut(Stream.str(), ConOut::Warning);
								YYABORT;
							};
							
							yyArgCtr = RecurArgCtr.top();
							RecurArgCtr.pop();
							
							GlobalStackMachine->AddInstruction(Opcodes::SYSCALL, VMType(VMTC_SYSTEM_CALL, 5, -1), COGlineno);
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_KEYWORD_SLEEP VMTT_PUNC_LPAREN { RecurArgCtr.push(yyArgCtr); yyArgCtr = 0; } argument_expression_list VMTT_PUNC_RPAREN
						{
							if(yyArgCtr != 1)
							{
								std::ostringstream Stream;
								Stream << "WARNING: System call \'sleep\' does not take " << yyArgCtr << " arguments.\n";
								ConOut(Stream.str(), ConOut::Warning);
								YYABORT;
							};
							
							yyArgCtr = RecurArgCtr.top();
							RecurArgCtr.pop();
							
							GlobalStackMachine->AddInstruction(Opcodes::SYSCALL, VMType(VMTC_SYSTEM_CALL, 3, -1), COGlineno);
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_IDENTIFIER VMTT_PUNC_LPAREN { RecurArgCtr.push(yyArgCtr); yyArgCtr = 0; } VMTT_PUNC_RPAREN
						{
							int LibOff = CogLib::TranslateLibraryName($1, yyArgCtr);
							if(LibOff == -1)
							{
								std::ostringstream Stream;
								Stream << "WARNING: Library function \'" << $1 << "\' does not exist or it does not take " << yyArgCtr << " arguments.\n";
								ConOut(Stream.str(), ConOut::Warning);
								YYABORT;
							}
							else
							{
								GlobalStackMachine->AddInstruction(Opcodes::LIBCALL, VMType(VMTC_LIBRARY_CALL, LibOff, -1), COGlineno);
							};
						
							yyArgCtr = RecurArgCtr.top();
							RecurArgCtr.pop();
							
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_IDENTIFIER VMTT_PUNC_LPAREN { RecurArgCtr.push(yyArgCtr); yyArgCtr = 0; } argument_expression_list VMTT_PUNC_RPAREN
						{
							int LibOff = CogLib::TranslateLibraryName($1, yyArgCtr);
							if(LibOff == -1)
							{
								std::ostringstream Stream;
								Stream << "WARNING: Library function \'" << $1 << "\' does not exist or it does not take " << yyArgCtr << " arguments.\n";
								ConOut(Stream.str(), ConOut::Warning);
								YYABORT;
							}
							else
							{
								GlobalStackMachine->AddInstruction(Opcodes::LIBCALL, VMType(VMTC_LIBRARY_CALL, LibOff, -1), COGlineno);
							};
														
							yyArgCtr = RecurArgCtr.top();
							RecurArgCtr.pop();
							
							GlobalStackMachine->AddInstruction(Opcodes::RPSH, VMType(VMTC_VOID), COGlineno);
						}
	| VMTT_IDENTIFIER VMTT_PUNC_INC
				{
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::INC, VMType(VMTC_VOID), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::STOR, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
				}
	| VMTT_IDENTIFIER VMTT_PUNC_DEC
				{
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::DEC, VMType(VMTC_VOID), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::STOR, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
				}
	| postfix_expression VMTT_PUNC_INC { /* eat it */ }
	| postfix_expression VMTT_PUNC_DEC { /* eat it */ }
	;
	
argument_expression_list
	: assignment_expression { ++yyArgCtr; }
	| argument_expression_list VMTT_PUNC_COMMA assignment_expression { ++yyArgCtr; }
	;
	
unary_expression
	: postfix_expression
	| VMTT_PUNC_SUB multiplicative_expression
				{
					GlobalStackMachine->AddInstruction(Opcodes::PUSH, VMType(-1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::MUL, VMType(VMTC_VOID), COGlineno);
				}
	| VMTT_PUNC_EQUALITY_NOT multiplicative_expression { GlobalStackMachine->AddInstruction(Opcodes::NOT, VMType(VMTC_VOID), COGlineno); }
	| VMTT_PUNC_INC VMTT_IDENTIFIER
				{
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::INC, VMType(VMTC_VOID), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::STOR, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
				}
	| VMTT_PUNC_DEC VMTT_IDENTIFIER
				{
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::DEC, VMType(VMTC_VOID), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::STOR, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
					GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
				}
	;
	
multiplicative_expression
	: unary_expression
	| multiplicative_expression VMTT_PUNC_MUL unary_expression { GlobalStackMachine->AddInstruction(Opcodes::MUL, VMType(VMTC_VOID), COGlineno) }
	| multiplicative_expression VMTT_PUNC_DIV unary_expression { GlobalStackMachine->AddInstruction(Opcodes::DIV, VMType(VMTC_VOID), COGlineno) }
	| multiplicative_expression VMTT_PUNC_MOD unary_expression { GlobalStackMachine->AddInstruction(Opcodes::MOD, VMType(VMTC_VOID), COGlineno) }
	;
	
additive_expression
	: multiplicative_expression
	| additive_expression VMTT_PUNC_ADD multiplicative_expression { GlobalStackMachine->AddInstruction(Opcodes::ADD, VMType(VMTC_VOID), COGlineno) }
	| additive_expression VMTT_PUNC_SUB multiplicative_expression { GlobalStackMachine->AddInstruction(Opcodes::SUB, VMType(VMTC_VOID), COGlineno) }
	;
	
shift_expression
	: additive_expression
	| shift_expression VMTT_PUNC_SHL additive_expression { GlobalStackMachine->AddInstruction(Opcodes::SHL, VMType(VMTC_VOID), COGlineno) }
	| shift_expression VMTT_PUNC_SHR additive_expression { GlobalStackMachine->AddInstruction(Opcodes::SHR, VMType(VMTC_VOID), COGlineno) }
	;
	
relational_expression
	: shift_expression
	| relational_expression VMTT_PUNC_EQUALITY_LT shift_expression { GlobalStackMachine->AddInstruction(Opcodes::CLT, VMType(VMTC_VOID), COGlineno) }
	| relational_expression VMTT_PUNC_EQUALITY_GT shift_expression { GlobalStackMachine->AddInstruction(Opcodes::CGT, VMType(VMTC_VOID), COGlineno) }
	| relational_expression VMTT_PUNC_EQUALITY_LTE shift_expression { GlobalStackMachine->AddInstruction(Opcodes::CLTE, VMType(VMTC_VOID), COGlineno) }
	| relational_expression VMTT_PUNC_EQUALITY_GTE shift_expression { GlobalStackMachine->AddInstruction(Opcodes::CGTE, VMType(VMTC_VOID), COGlineno) }
	;
	
equality_expression
	: relational_expression
	| equality_expression VMTT_PUNC_EQUALITY_EQ relational_expression { GlobalStackMachine->AddInstruction(Opcodes::CE, VMType(VMTC_VOID), COGlineno) }
	| equality_expression VMTT_PUNC_EQUALITY_NEQ relational_expression { GlobalStackMachine->AddInstruction(Opcodes::CNE, VMType(VMTC_VOID), COGlineno) }
	;
	
and_expression
	: equality_expression
	| and_expression VMTT_PUNC_AND equality_expression { GlobalStackMachine->AddInstruction(Opcodes::AND, VMType(VMTC_VOID), COGlineno) }
	;
	
exclusive_or_expression
	: and_expression
	| exclusive_or_expression VMTT_PUNC_XOR and_expression { GlobalStackMachine->AddInstruction(Opcodes::XOR, VMType(VMTC_VOID), COGlineno) }
	;
	
inclusive_or_expression
	: exclusive_or_expression
	| inclusive_or_expression VMTT_PUNC_OR exclusive_or_expression { GlobalStackMachine->AddInstruction(Opcodes::OR, VMType(VMTC_VOID), COGlineno) }
	;
	
logical_and_expression
	: inclusive_or_expression
	| logical_and_expression VMTT_PUNC_EQUALITY_AND inclusive_or_expression { GlobalStackMachine->AddInstruction(Opcodes::CAND, VMType(VMTC_VOID), COGlineno) }
	;
	
logical_or_expression
	: logical_and_expression
	| logical_or_expression VMTT_PUNC_EQUALITY_OR logical_and_expression { GlobalStackMachine->AddInstruction(Opcodes::COR, VMType(VMTC_VOID), COGlineno) }
	;
	
assignment_expression
	: logical_or_expression
	| identifier VMTT_PUNC_LBRACKET expression VMTT_PUNC_RBRACKET VMTT_PUNC_ASSIGN assignment_expression
							{
								GlobalStackMachine->AddInstruction(Opcodes::ISTR, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
							}
	| identifier VMTT_PUNC_ASSIGN assignment_expression
							{
								GlobalStackMachine->AddInstruction(Opcodes::STOR, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
								GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1), COGlineno);
							}
	| identifier VMTT_PUNC_DBLCOLON identifier VMTT_PUNC_ASSIGN assignment_expression
							{
								int Scope = GlobalStackMachine->Heap.TranslateNamespaceName($1);
								if(Scope == -1)
								{
									yyerror("Scope does not exist");
									return 1;
								}
								int Value = GlobalStackMachine->Heap.TranslateNamespaceHeapName(Scope, $3);
								if(Value == -1)
								{
									yyerror("Unknown Identifier");
									return 1;
								};
								GlobalStackMachine->AddInstruction(Opcodes::STOR, VMType(VMTC_HEAP_ADDRESS, Value, Scope), COGlineno);
								GlobalStackMachine->AddInstruction(Opcodes::LOAD, VMType(VMTC_HEAP_ADDRESS, Value, Scope), COGlineno);
							}
	;
	
expression
	: assignment_expression
	| expression VMTT_PUNC_COMMA assignment_expression
	;
	
statement
	: labeled_statement
	| compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	;
	
message_label
	: identifier VMTT_PUNC_COLON
			{
				int LocalHeapAddr = GlobalStackMachine->Heap.TranslateLocalHeapName($1);
				if(LocalHeapAddr == -1)
				{
					GlobalStackMachine->Heap.AddLocal(std::string($1), VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1));
				}
				else
				{
					GlobalStackMachine->Heap.Set(GlobalStackMachine->Heap.TranslateLocalHeapName($1), -1, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1));
				};
			}
	;
	
labeled_statement
	: message_label statement
	;
	
compound_statement
	: VMTT_PUNC_LBRACE VMTT_PUNC_RBRACE
	| VMTT_PUNC_LBRACE statement_list VMTT_PUNC_RBRACE
	;
	
statement_list
	: statement
	| statement_list statement
	;
	
expression_statement
	: VMTT_PUNC_ENDSTMT
	| expression VMTT_PUNC_ENDSTMT
			{
				Opcode Temp = GlobalStackMachine->Code[GlobalStackMachine->InstructionCount - 1].OPCODE;
				if((Temp == Opcodes::PUSH) || (Temp == Opcodes::LOAD) || (Temp == Opcodes::RPSH))
				{
					GlobalStackMachine->RemoveInstruction(); /* Remove last PUSH/LOAD/RPSH instruction */
				}
			}
	;
	
if_block_header
	: VMTT_KEYWORD_IF VMTT_PUNC_LPAREN expression VMTT_PUNC_RPAREN
			{
				$$ = GlobalStackMachine->AddInstruction(Opcodes::JNT, VMType(VMTC_VOID), COGlineno); // Placeholder, return back placeholder address
			}
	;
	
if_block
	: if_block_header statement
			{
				BackPatchStack.push(GlobalStackMachine->InstructionCount);
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_VOID), COGlineno);	// Placeholder for jump to end
				BackPatchStack.push(1);	// Back patch counter
				$$ = $1; // return back JNT
			}
	;
	
else_block
	: if_block VMTT_KEYWORD_ELSE
			{
				GlobalStackMachine->Code[$1] = Instruction(Opcodes::JNT, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[$1].LineNumber);
			}
	;

selection_statement
	: if_block
			{
				GlobalStackMachine->Code[$1] = Instruction(Opcodes::JNT, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount - 1, -1), GlobalStackMachine->Code[$1].LineNumber);
				int BackPatchCount = BackPatchStack.top(); BackPatchStack.pop();
				
				for(int i = 0; i < BackPatchCount; ++i)
				{
					GlobalStackMachine->Code[BackPatchStack.top()] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount - 1, -1), GlobalStackMachine->Code[BackPatchStack.top()].LineNumber);
					BackPatchStack.pop();
				};
				
				GlobalStackMachine->RemoveInstruction(); // Remove last JMP
			}
	| else_block statement
			{
				int BackPatchCount = BackPatchStack.top(); BackPatchStack.pop();
				
				for(int i = 0; i < BackPatchCount; ++i)
				{
					GlobalStackMachine->Code[BackPatchStack.top()] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[BackPatchStack.top()].LineNumber);
					BackPatchStack.pop();
				};
			}
	;
	
while_statement_header
	: VMTT_KEYWORD_WHILE { BackPatchStack.push(GlobalStackMachine->InstructionCount); } // return start of the loop, before the conditional
	;
	
while_statement
	: while_statement_header VMTT_PUNC_LPAREN expression VMTT_PUNC_RPAREN
			{
				BackPatchStack.push(GlobalStackMachine->InstructionCount); // Push back address of JNT
				BackPatchStack.push(0); // Add the end-patch counter
				GlobalStackMachine->AddInstruction(Opcodes::JNT, VMType(VMTC_VOID), COGlineno); // Placeholder
			}
	;
	
do_statement
	: VMTT_KEYWORD_DO
			{
				BackPatchStack.push(GlobalStackMachine->InstructionCount); // return start of loop
				BackPatchStack.push(0); // add end-patch counter
			}
	;
	
for_statement_header
	: VMTT_KEYWORD_FOR VMTT_PUNC_LPAREN expression_statement
			{
				BackPatchStack.push(GlobalStackMachine->InstructionCount); // return start of loop before conditional
			}
	;
	
for_statement_block
	: for_statement_header expression_statement
			{
				BackPatchStack.push(GlobalStackMachine->InstructionCount); // return JNT offset
				GlobalStackMachine->AddInstruction(Opcodes::JNT, VMType(VMTC_VOID), COGlineno); // placeholder
				BackPatchStack.push(GlobalStackMachine->InstructionCount); // return JMP offset
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_VOID), COGlineno); // this should jump to EXECUTION STATEMENT
				BackPatchStack.push(GlobalStackMachine->InstructionCount); // return address of first instruction of iteration statement
			}
	;
	
for_statement
	: for_statement_block expression VMTT_PUNC_RPAREN
			{
				int ItOff = BackPatchStack.top(); BackPatchStack.pop();
				int JMPOff = BackPatchStack.top(); BackPatchStack.pop();
				int JNTOff = BackPatchStack.top(); BackPatchStack.pop();
				int StartOff = BackPatchStack.top(); BackPatchStack.pop();
				
				// Add jump to beginning right after expression
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, StartOff, -1), COGlineno);
				
				// Execution block is immediately after this instruction
				
				// Patch back execution address
				GlobalStackMachine->Code[JMPOff] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[JMPOff].LineNumber);
				
				// Push remaining values back onto the back-patch stack
				BackPatchStack.push(ItOff);
				BackPatchStack.push(JNTOff);
				BackPatchStack.push(0); // Add the back-patch counter
			}
	| for_statement_block VMTT_PUNC_RPAREN
			{
				// Pop off pointless backpatch
				BackPatchStack.pop();
				
				// Execution block is immediately after this instruction
				
				// Patch back execution address
				GlobalStackMachine->Code[BackPatchStack.top()] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[BackPatchStack.top()].LineNumber);
				BackPatchStack.pop();
				
				BackPatchStack.push(0); // Add the back-patch counter
			}
	;
	
iteration_statement
	: while_statement statement
			{
				int EndCnt = BackPatchStack.top(); BackPatchStack.pop();
				std::stack<int> EndPatchStack;
				int i;
				for(i = 0; i < EndCnt; ++i)
				{
					EndPatchStack.push(BackPatchStack.top());
					BackPatchStack.pop();
				};
				
				int JNTOff = BackPatchStack.top(); BackPatchStack.pop();
				
				// Add jump back to start of loop
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, BackPatchStack.top(), -1), COGlineno); BackPatchStack.pop();
				
				// Set JNT to end of loop
				GlobalStackMachine->Code[JNTOff] = Instruction(Opcodes::JNT, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[JNTOff].LineNumber);
				
				for(i = 0; i < EndCnt; ++i)
				{
					// Back-patch breaks
					GlobalStackMachine->Code[EndPatchStack.top()] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[EndPatchStack.top()].LineNumber);
					EndPatchStack.pop();
				};
			}
	| do_statement statement VMTT_KEYWORD_WHILE VMTT_PUNC_LPAREN expression VMTT_PUNC_RPAREN VMTT_PUNC_ENDSTMT
			{
				int EndCnt = BackPatchStack.top(); BackPatchStack.pop();
				std::stack<int> EndPatchStack;
				int i;
				for(i = 0; i < EndCnt; ++i)
				{
					EndPatchStack.push(BackPatchStack.top());
					BackPatchStack.pop();
				};
				
				// Add jump back to start of loop
				GlobalStackMachine->AddInstruction(Opcodes::JT, VMType(VMTC_INSTRUCTION_ADDRESS, BackPatchStack.top(), -1), COGlineno); BackPatchStack.pop();
				
				for(i = 0; i < EndCnt; ++i)
				{
					// Back-patch breaks
					GlobalStackMachine->Code[EndPatchStack.top()] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[EndPatchStack.top()].LineNumber);
					EndPatchStack.pop();
				};
			}
	| for_statement statement
			{
				int EndCnt = BackPatchStack.top(); BackPatchStack.pop();
				std::stack<int> EndPatchStack;
				int i;
				for(i = 0; i < EndCnt; ++i)
				{
					EndPatchStack.push(BackPatchStack.top());
					BackPatchStack.pop();
				};
				
				int JNTOff = BackPatchStack.top(); BackPatchStack.pop();
				
				// Add jump back to start of iterator
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, BackPatchStack.top(), -1), COGlineno); BackPatchStack.pop();
				
				// Back patch JNT instruction
				GlobalStackMachine->Code[JNTOff] = Instruction(Opcodes::JNT, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[JNTOff].LineNumber);
				
				// Back patch breaks
				for(i = 0; i < EndCnt; ++i)
				{
					GlobalStackMachine->Code[EndPatchStack.top()] = Instruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, GlobalStackMachine->InstructionCount, -1), GlobalStackMachine->Code[EndPatchStack.top()].LineNumber);
					EndPatchStack.pop();
				};
			}
	;
	
jump_statement
	: VMTT_KEYWORD_CONTINUE VMTT_PUNC_ENDSTMT
			{
				int EndCnt = BackPatchStack.top(); BackPatchStack.pop();
				std::stack<int> EndPatchStack;
				int i;
				for(i = 0; i < EndCnt; ++i)
				{
					EndPatchStack.push(BackPatchStack.top());
					BackPatchStack.pop();
				}; // Take the end-patches off the backpatchstack
				
				int JNTOff = BackPatchStack.top(); BackPatchStack.pop();
				
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_INSTRUCTION_ADDRESS, BackPatchStack.top(), -1), COGlineno);
				
				BackPatchStack.push(JNTOff);
				
				// Put the end-patches back on the backpatchstack
				for(i = 0; i < EndCnt; ++i)
				{
					BackPatchStack.push(EndPatchStack.top());
					EndPatchStack.pop();
				};
				
				BackPatchStack.push(EndCnt); // Put the endpatch counter back on
			}
	| VMTT_KEYWORD_BREAK VMTT_PUNC_ENDSTMT
			{
				int EndCnt = BackPatchStack.top(); BackPatchStack.pop();
				BackPatchStack.push(GlobalStackMachine->InstructionCount);
				GlobalStackMachine->AddInstruction(Opcodes::JMP, VMType(VMTC_VOID), COGlineno);
				BackPatchStack.push(++EndCnt);
			}
	| VMTT_KEYWORD_RETURN VMTT_PUNC_ENDSTMT { GlobalStackMachine->AddInstruction(Opcodes::RETURN, VMType(VMTC_VOID), COGlineno); }
	| VMTT_KEYWORD_RETURN expression VMTT_PUNC_ENDSTMT { GlobalStackMachine->AddInstruction(Opcodes::RETURN, VMType(VMTC_VOID), COGlineno); }
	| VMTT_KEYWORD_CALL VMTT_IDENTIFIER VMTT_PUNC_ENDSTMT
			{
				int LocalHeapAddr = GlobalStackMachine->Heap.TranslateLocalHeapName($2);
				if(LocalHeapAddr == -1)
				{
					GlobalStackMachine->Heap.AddLocal(std::string($2), VMType(yyVMTbuf), ExtTab);
				};
				
				GlobalStackMachine->AddInstruction(Opcodes::MSGCALL, VMType(VMTC_HEAP_ADDRESS, GlobalStackMachine->Heap.TranslateLocalHeapName($2), -1), COGlineno);
			}
	;
	
/* END OF CODE SECTION */

%%

int yyerror(char* s)
{
	extern int COGlineno;	// defined in lex.c
	
	std::stringstream ErrorString;
	ErrorString << "ERROR: \"" << std::string(s) << "\" on line " << COGlineno << "\n";
	ConOut(ErrorString.str(), ConOut::Error);
	
	while(!RecurArgCtr.empty())
	{
		RecurArgCtr.pop();
	};
	
	yyArgCtr = 0;
	
	while(!BackPatchStack.empty())
	{
		BackPatchStack.pop();
	};
	
	return 1;
};