/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "di_input.h"

#include "ogl_render.h"

Sith::Win32::Input* Sith::Win32::Input::StaticInterface = 0;

BOOL CALLBACK DIEnumDevicesCallback(LPCDIDEVICEINSTANCE lpddi, LPVOID pvRef)
{
	Sith::Win32::Input *input = (Sith::Win32::Input *)pvRef;

	if (FAILED(input->g_DI->CreateDevice(lpddi->guidInstance, &input->g_Joy, NULL)))
	{
		input->g_Joy = 0;
		return DIENUM_STOP;
	}
	input->g_Joy->SetCooperativeLevel(Sith::Win32::Renderer::Get().hMainWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
	input->g_Joy->SetDataFormat(&c_dfDIJoystick);
	input->g_Joy->Acquire();

	return DIENUM_STOP;
}

Sith::Win32::Input::Input() : Task("DirectX Input Handler")
{
	// Set static interface
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple input objects for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;


	Loaded = 0;

	if (FAILED(DirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, IID_IDirectInput8, (void**)&g_DI, NULL)))
	{
		return;
	}

	if (FAILED(g_DI->CreateDevice(GUID_SysKeyboard, &g_KDIDev, NULL)))
		return;
	g_KDIDev->SetDataFormat(&c_dfDIKeyboard);
	g_KDIDev->SetCooperativeLevel(Renderer::Get().hMainWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
	g_KDIDev->Acquire();
	DIPROPDWORD  dipdw;
	dipdw.diph.dwSize = sizeof(DIPROPDWORD);
	dipdw.diph.dwHeaderSize = sizeof(DIPROPHEADER);
	dipdw.diph.dwObj = 0;
	dipdw.diph.dwHow = DIPH_DEVICE;
	dipdw.dwData = 10;
	g_KDIDev->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph);

	if (FAILED(g_DI->CreateDevice(GUID_SysMouse, &g_pMouse, NULL)))
		return;
	g_pMouse->SetDataFormat(&c_dfDIMouse2);
	g_pMouse->SetCooperativeLevel(Renderer::Get().hMainWnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE);
	g_pMouse->Acquire();


	g_Joy = 0;
	g_DI->EnumDevices(DI8DEVCLASS_GAMECTRL, DIEnumDevicesCallback, this, DIEDFL_ATTACHEDONLY);


	MapInput();

	Loaded = 1;

	COG::CogLib::InjectLibraryFunction("bind", 2, &bind);
	COG::CogLib::InjectLibraryFunction("bind", 3, &verbind);

	return;
};

Sith::Win32::Input::~Input()
{
	Loaded = 0;

	if (g_DI)
    {
        if (g_KDIDev)
        {
            g_KDIDev->Unacquire();
            g_KDIDev->Release();
            g_KDIDev = NULL;
        }
		if (g_pMouse)
		{
			g_pMouse->Unacquire();
			g_pMouse->Release();
			g_pMouse = NULL;
		}
		if (g_Joy)
		{
			g_Joy->Unacquire();
			g_Joy->Release();
			g_Joy = NULL;
		}
        g_DI->Release();
        g_DI = NULL;
    }

	return;
};

Sith::SITHERROR Sith::Win32::Input::Process()
{
	if (!Loaded)
	{
		ConOut("CRITICAL ERROR: Failed to create input device.\n", ConOut::Critical);
		Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	}

	if (GetForegroundWindow() != Renderer::Get().hMainWnd)
	{
		memset(&dims, 0, sizeof(DIMOUSESTATE2));
		memset(keybuffer, 0, sizeof(keybuffer));
		memset(&dijs, 0, sizeof(DIJOYSTATE));
		dijs.lRx = 32767;
		dijs.lRy = 32767;
		dijs.lRz = 32767;
		dijs.lX = 32767;
		dijs.lY = 32767;
		dijs.lZ = 32767;
		dijs.rgdwPOV[0] = 1;
		dijs.rgdwPOV[1] = 1;
		dijs.rgdwPOV[2] = 1;
		dijs.rgdwPOV[3] = 1;
		dijs.rglSlider[0] = 32767;
		dijs.rglSlider[1] = 32767;
		return SE_OK;
	}

	g_pMouse->Acquire();
	g_pMouse->GetDeviceState(sizeof(DIMOUSESTATE2), &dims);
	g_KDIDev->Acquire();
	g_KDIDev->GetDeviceState(sizeof(keybuffer), &keybuffer);

	if (g_Joy)
	{
		g_Joy->Acquire();
		g_Joy->Poll();
		g_Joy->GetDeviceState(sizeof(DIJOYSTATE), &dijs);
	}

	return SE_OK;
};

float Sith::Win32::Input::GetState(std::string action)
{
	std::pair<std::multimap<std::string, std::pair<std::string, float> >::iterator, std::multimap<std::string, std::pair<std::string, float> >::iterator> bounds;
	bounds = Bindings.equal_range(action);

	float deadzone = .1f;

	float rval = 0;
	for (std::multimap<std::string, std::pair<std::string, float> >::iterator i = bounds.first; i != bounds.second; ++i)
	{
		float t;
		int inputval = InputMap[(*i).second.first];
		switch (inputval)
		{
		case 300:
		case 301:
		case 302:
		case 303:
		case 304:
		case 305:
		case 306:
		case 307: // mouse buttons
			rval += (dims.rgbButtons[inputval - 300] > 0) * (*i).second.second;
			break;
		case 308: // mousex
			rval += -dims.lX * (*i).second.second;
			break;
		case 309: // mousey
			rval += -dims.lY * (*i).second.second;
			break;
		case 310: // mousez - this has to be handled differently, since return values are weird
			if (dims.lZ > 0)
				rval += 1 * (*i).second.second;
			else if (dims.lZ < 0)
				rval += -1 * (*i).second.second;
			break;

		case 400:
		case 401:
		case 402:
		case 403:
		case 404:
		case 405:
		case 406:
		case 407:
		case 408:
		case 409:
		case 410:
		case 411:
		case 412:
		case 413:
		case 414:
		case 415:
		case 416:
		case 417:
		case 418:
		case 419:
		case 420:
		case 421:
		case 422:
		case 423:
		case 424:
		case 425:
		case 426:
		case 427:
		case 428:
		case 429:
		case 430:
		case 431: // joystick buttons
			rval += (dijs.rgbButtons[inputval - 400] > 0) * (*i).second.second;
			break;

		// pov 0
		case 432:
			if (dijs.rgdwPOV[0] == 1 || LOWORD(dijs.rgdwPOV[0]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[0] <= 4500 || dijs.rgdwPOV[0] > 31500)
				rval += (*i).second.second;
			break;
		case 433:
			if (dijs.rgdwPOV[0] == 1 || LOWORD(dijs.rgdwPOV[0]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[0] > 4500 && dijs.rgdwPOV[0] <= 13500)
				rval += (*i).second.second;
			break;
		case 434:
			if (dijs.rgdwPOV[0] == 1 || LOWORD(dijs.rgdwPOV[0]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[0] > 13500 && dijs.rgdwPOV[0] <= 22500)
				rval += (*i).second.second;
			break;
		case 435:
			if (dijs.rgdwPOV[0] == 1 || LOWORD(dijs.rgdwPOV[0]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[0] > 22500 && dijs.rgdwPOV[0] <= 31500)
				rval += (*i).second.second;
			break;

		// pov 1
		case 436:
			if (dijs.rgdwPOV[1] == 1 || LOWORD(dijs.rgdwPOV[1]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[1] <= 4500 || dijs.rgdwPOV[1] > 31500)
				rval += (*i).second.second;
			break;
		case 437:
			if (dijs.rgdwPOV[1] == 1 || LOWORD(dijs.rgdwPOV[1]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[1] > 4500 && dijs.rgdwPOV[1] <= 13500)
				rval += (*i).second.second;
			break;
		case 438:
			if (dijs.rgdwPOV[1] == 1 || LOWORD(dijs.rgdwPOV[1]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[1] > 13500 && dijs.rgdwPOV[1] <= 22500)
				rval += (*i).second.second;
			break;
		case 439:
			if (dijs.rgdwPOV[1] == 1 || LOWORD(dijs.rgdwPOV[1]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[1] > 22500 && dijs.rgdwPOV[1] <= 31500)
				rval += (*i).second.second;
			break;

		// pov 2
		case 440:
			if (dijs.rgdwPOV[2] == 1 || LOWORD(dijs.rgdwPOV[2]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[2] <= 4500 || dijs.rgdwPOV[2] > 31500)
				rval += (*i).second.second;
			break;
		case 441:
			if (dijs.rgdwPOV[2] == 1 || LOWORD(dijs.rgdwPOV[2]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[2] > 4500 && dijs.rgdwPOV[2] <= 13500)
				rval += (*i).second.second;
			break;
		case 442:
			if (dijs.rgdwPOV[2] == 1 || LOWORD(dijs.rgdwPOV[2]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[2] > 13500 && dijs.rgdwPOV[2] <= 22500)
				rval += (*i).second.second;
			break;
		case 443:
			if (dijs.rgdwPOV[2] == 1 || LOWORD(dijs.rgdwPOV[2]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[2] > 22500 && dijs.rgdwPOV[2] <= 31500)
				rval += (*i).second.second;
			break;

		// pov 3
		case 444:
			if (dijs.rgdwPOV[3] == 1 || LOWORD(dijs.rgdwPOV[3]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[3] <= 4500 || dijs.rgdwPOV[3] > 31500)
				rval += (*i).second.second;
			break;
		case 445:
			if (dijs.rgdwPOV[3] == 1 || LOWORD(dijs.rgdwPOV[3]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[3] > 4500 && dijs.rgdwPOV[3] <= 13500)
				rval += (*i).second.second;
			break;
		case 446:
			if (dijs.rgdwPOV[3] == 1 || LOWORD(dijs.rgdwPOV[3]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[3] > 13500 && dijs.rgdwPOV[3] <= 22500)
				rval += (*i).second.second;
			break;
		case 447:
			if (dijs.rgdwPOV[3] == 1 || LOWORD(dijs.rgdwPOV[3]) == 65535) // centered
				break;
			if (dijs.rgdwPOV[3] > 22500 && dijs.rgdwPOV[3] <= 31500)
				rval += (*i).second.second;
			break;

		case 448: // sliders
			t = 1.0f - dijs.rglSlider[0] / 32767.0f;
			if(fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;

		case 449: // sliders
			t = 1.0f - dijs.rglSlider[1] / 32767.0f;
			if(fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;

		case 450: // axes
			t = 1.0f - dijs.lX / 32767.0f;
			if (fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;
		case 451:
			t = 1.0f - dijs.lY / 32767.0f;
			if (fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;
		case 452:
			t = 1.0f - dijs.lZ / 32767.0f;
			if (fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;
		case 453: // rotation axes
			t = 1.0f - dijs.lRx / 32767.0f;
			if (fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;
		case 454:
			t = 1.0f - dijs.lRy / 32767.0f;
			if (fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;
		case 455:
			t = 1.0f - dijs.lRz / 32767.0f;
			if (fabs(t) < deadzone)
				t = 0;
			rval += t * (*i).second.second;
			break;

		default: // keyboard
			if (inputval < 1 || inputval > 221)
				break;
			if (keybuffer[inputval] & 0x80)
			{
				rval += (*i).second.second;
			}
			break;
		};
	}

	return rval;
}

Sith::SITHERROR Sith::Win32::Input::BindInput(std::string keyname, std::string action, float scale)
{
	Bindings.insert(std::pair<std::string, std::pair<std::string, float> >(action, std::pair<std::string, float>(keyname, scale)));
	return SE_OK;
}

void Sith::Win32::Input::MapInput()
{
	InputMap["escape"]	= 1;
	InputMap["1"]	= 2;
	InputMap["2"]	= 3;
	InputMap["3"]	= 4;
	InputMap["4"]	= 5;
	InputMap["5"]	= 6;
	InputMap["6"]	= 7;
	InputMap["7"]	= 8;
	InputMap["8"]	= 9;
	InputMap["9"]	= 10;
	InputMap["0"]	= 11;
	InputMap["minus"]	= 12;
	InputMap["equals"]	= 13;
	InputMap["back"]	= 14;
	InputMap["tab"]	= 15;
	InputMap["q"]	= 16;
	InputMap["w"]	= 17;
	InputMap["e"]	= 18;
	InputMap["r"]	= 19;
	InputMap["t"]	= 20;
	InputMap["y"]	= 21;
	InputMap["u"]	= 22;
	InputMap["i"]	= 23;
	InputMap["o"]	= 24;
	InputMap["p"]	= 25;
	InputMap["lbracket"]	= 26;
	InputMap["rbracket"]	= 27;
	InputMap["return"]	= 28;
	InputMap["lcontrol"]	= 29;
	InputMap["a"]	= 30;
	InputMap["s"]	= 31;
	InputMap["d"]	= 32;
	InputMap["f"]	= 33;
	InputMap["g"]	= 34;
	InputMap["h"]	= 35;
	InputMap["j"]	= 36;
	InputMap["k"]	= 37;
	InputMap["l"]	= 38;
	InputMap["semicolon"]	= 39;
	InputMap["apostrophe"]	= 40;
	InputMap["grave"]	= 41;
	InputMap["lshift"]	= 42;
	InputMap["backslash"]	= 43;
	InputMap["z"]	= 44;
	InputMap["x"]	= 45;
	InputMap["c"]	= 46;
	InputMap["v"]	= 47;
	InputMap["b"]	= 48;
	InputMap["n"]	= 49;
	InputMap["m"]	= 50;
	InputMap["comma"]	= 51;
	InputMap["period"]	= 52;
	InputMap["slash"]	= 53;
	InputMap["rshift"]	= 54;
	InputMap["multiply"]	= 55;
	InputMap["lalt"]	= 56;
	InputMap["space"]	= 57;
	InputMap["capital"]	= 58;
	InputMap["f1"]	= 59;
	InputMap["f2"]	= 60;
	InputMap["f3"]	= 61;
	InputMap["f4"]	= 62;
	InputMap["f5"]	= 63;
	InputMap["f6"]	= 64;
	InputMap["f7"]	= 65;
	InputMap["f8"]	= 66;
	InputMap["f9"]	= 67;
	InputMap["f10"]	= 68;
	InputMap["numlock"]	= 69;
	InputMap["scroll"]	= 70;
	InputMap["numpad7"]	= 71;
	InputMap["numpad8"]	= 72;
	InputMap["numpad9"]	= 73;
	InputMap["subtract"]	= 74;
	InputMap["numpad4"]	= 75;
	InputMap["numpad5"]	= 76;
	InputMap["numpad6"]	= 77;
	InputMap["add"]	= 78;
	InputMap["numpad1"]	= 79;
	InputMap["numpad2"]	= 80;
	InputMap["numpad3"]	= 81;
	InputMap["numpad0"]	= 82;
	InputMap["decimal"]	= 83;
	InputMap["f11"]	= 87;
	InputMap["f12"]	= 88;
	InputMap["numpadenter"]	= 156;
	InputMap["rcontrol"]	= 157;
	InputMap["divide"]		= 181;
	InputMap["sysrq"]		= 183;
	InputMap["ralt"]		= 184;
	InputMap["pause"]		= 197;
	InputMap["home"]		= 199;
	InputMap["up"]			= 200;
	InputMap["pgup"]		= 201;
	InputMap["left"]		= 203;
	InputMap["right"]		= 205;
	InputMap["end"]			= 207;
	InputMap["down"]		= 208;
	InputMap["pgdn"]		= 209;
	InputMap["insert"]		= 210;
	InputMap["delete"]		= 211;
	InputMap["lwin"]		= 219;
	InputMap["rwin"]		= 220;
	InputMap["apps"]		= 221;

	// mouse input
	InputMap["mouse0"] = 300;
	InputMap["mouse1"] = 301;
	InputMap["mouse2"] = 302;
	InputMap["mouse3"] = 303;
	InputMap["mouse4"] = 304;
	InputMap["mouse5"] = 305;
	InputMap["mouse6"] = 306;
	InputMap["mouse7"] = 307;
	InputMap["mousex"] = 308;
	InputMap["mousey"] = 309;
	InputMap["mousez"] = 310;

	// joystick input
	if (g_Joy)
	{
		InputMap["joy0"] = 400;
		InputMap["joy1"] = 401;
		InputMap["joy2"] = 402;
		InputMap["joy3"] = 403;
		InputMap["joy4"] = 404;
		InputMap["joy5"] = 405;
		InputMap["joy6"] = 406;
		InputMap["joy7"] = 407;
		InputMap["joy8"] = 408;
		InputMap["joy9"] = 409;
		InputMap["joy10"] = 410;
		InputMap["joy11"] = 411;
		InputMap["joy12"] = 412;
		InputMap["joy13"] = 413;
		InputMap["joy14"] = 414;
		InputMap["joy15"] = 415;
		InputMap["joy16"] = 416;
		InputMap["joy17"] = 417;
		InputMap["joy18"] = 418;
		InputMap["joy19"] = 419;
		InputMap["joy20"] = 420;
		InputMap["joy21"] = 421;
		InputMap["joy22"] = 422;
		InputMap["joy23"] = 423;
		InputMap["joy24"] = 424;
		InputMap["joy25"] = 425;
		InputMap["joy26"] = 426;
		InputMap["joy27"] = 427;
		InputMap["joy28"] = 428;
		InputMap["joy29"] = 429;
		InputMap["joy30"] = 430;
		InputMap["joy31"] = 431;

		InputMap["joypov0n"] = 432;
		InputMap["joypov0e"] = 433;
		InputMap["joypov0s"] = 434;
		InputMap["joypov0w"] = 435;

		InputMap["joypov1n"] = 436;
		InputMap["joypov1e"] = 437;
		InputMap["joypov1s"] = 438;
		InputMap["joypov1w"] = 439;

		InputMap["joypov2n"] = 440;
		InputMap["joypov2e"] = 441;
		InputMap["joypov2s"] = 442;
		InputMap["joypov2w"] = 443;

		InputMap["joypov3n"] = 444;
		InputMap["joypov3e"] = 445;
		InputMap["joypov3s"] = 446;
		InputMap["joypov3w"] = 447;

		InputMap["joyslider0"] = 448;
		InputMap["joyslider1"] = 449;

		InputMap["joyx"] = 450;
		InputMap["joyy"] = 451;
		InputMap["joyz"] = 452;
		InputMap["joyrx"] = 453;
		InputMap["joyry"] = 454;
		InputMap["joyrz"] = 455;
	}
}