/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _DI_INPUT_H_INCLUDED_
#define _DI_INPUT_H_INCLUDED_

#pragma comment (lib, "dxguid")
#pragma comment (lib, "dinput8")
#define DIRECTINPUT_VERSION 0x0800
#include <dinput.h>

namespace Sith
{
	namespace Win32
	{
		// interface to query for key states
		// generic input types for the query, which map (through a multimap)
		// to one or more keys to check
		// separate input bindings for keyboard and mouse/joystick, to handle
		// things like forward/backward mapping
		// return type is float, 0 or 1 for keys and -1 to 1 for an axis
		class Input : public System::Task
		{
			friend class Sith::System::KernelProc;
			friend Sith::SITHERROR Sith::COG::VMHeap::SerializeGlobal(std::string);
		private:

			LPDIRECTINPUTDEVICE8	g_KDIDev;
			LPDIRECTINPUTDEVICE8	g_pMouse;

			DIMOUSESTATE2			dims;
			DIJOYSTATE				dijs;
			BYTE					keybuffer[256];

			int Loaded;

			std::map<std::string, int> InputMap; // maps input names to values

			// action, with key/scale pairs
			std::multimap<std::string, std::pair<std::string, float> > Bindings;

			void MapInput();

			static Input* StaticInterface;

			Input();
			~Input();

			Sith::SITHERROR Process();

		public:
			LPDIRECTINPUT8			g_DI;
			LPDIRECTINPUTDEVICE8	g_Joy;

			Sith::SITHERROR BindInput(std::string keyname, std::string action, float scale);
			float GetState(std::string action);

			static inline Input& Get()
			{
				return *StaticInterface;
			};

			COGLIBAPI verbind(COG::VMType* keyname, COG::VMType* action, COG::VMType* scale)
			{
				Get().BindInput(keyname->GetString(), action->GetString(), scale->GetFloat());
				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI bind(COG::VMType* keyname, COG::VMType* action)
			{
				Get().BindInput(keyname->GetString(), action->GetString(), 1.0f);
				return new COG::VMType(COG::VMTC_VOID);
			};
		};
	};
};

#endif // _DI_INPUT_H_INCLUDED_