/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "uiwidget.h"
#include "uimain.h"

#include "dx_render.h"

#include "worldstate.h"

const DWORD Sith::Win32::D3DUIVertex_FVF = D3DFVF_XYZRHW | D3DFVF_TEX1 | D3DFVF_DIFFUSE;
const DWORD Sith::Win32::D3DJKLVertex_FVF = D3DFVF_XYZ | D3DFVF_TEX1 | D3DFVF_DIFFUSE;

Sith::Win32::Renderer* Sith::Win32::Renderer::StaticInterface = 0;

Sith::Win32::Renderer::Renderer() : Task("Direct3D 9 Renderer")
{
	// Set static interface
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple renderer objects for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;

	// Add renderer settings
	Sith::COG::CogLib::InjectNamespace("video");
	
	// ****** NOTE ******** CHANGE THESE, THEY SHOULD BE AUTOMATICALLY GENERATED
	Sith::COG::CogLib::AddSetting("video::fullscreen", Sith::COG::VMType(false));
	Sith::COG::CogLib::AddSetting("video::x_res", Sith::COG::VMType(int(1024)));
	Sith::COG::CogLib::AddSetting("video::y_res", Sith::COG::VMType(int(768)));
	Sith::COG::CogLib::AddSetting("video::bitdepth", Sith::COG::VMType(int(32)));
	Sith::COG::CogLib::AddSetting("video::vsync", Sith::COG::VMType(true));
	Sith::COG::CogLib::AddSetting("video::triplebuffer", Sith::COG::VMType(true));
	Sith::COG::CogLib::AddSetting("video::clear", Sith::COG::VMType(true));
	return;
};

Sith::Win32::Renderer::~Renderer()
{
	if(lpD3D)
	{
		lpD3D->Release();
	};

	if(hMainWnd)
	{
		DestroyWindow(hMainWnd);

		// Handle last remaining messages
		while(GetMessage(&Msg, hMainWnd, 0U, 0U))
		{
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		};
	};

	return;
};

LRESULT CALLBACK Sith::Win32::Renderer::RenderWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_CLOSE:
		DestroyWindow(hWnd);
		break;

	case WM_DESTROY:
		Sith::Kernel::Get().SendSignal(Sith::KS_QUIT);
		break;

	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	};

	return 0;
};

Sith::SITHERROR Sith::Win32::Renderer::Startup()
{
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_OWNDC;
	wc.lpfnWndProc = RenderWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "SITH 2 D3DWND";
	wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

	if(!RegisterClassEx(&wc))
	{
		MessageBox(GetDesktopWindow(), "Failed to register renderer window class", "Error", MB_OK | MB_ICONERROR);
		Sith::ConOut("CRITICAL ERROR: Failed to register renderer window class.\n", ConOut::Critical);
		Sith::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	};

	if(Sith::COG::CogLib::GetSetting("video::fullscreen").GetBool())
	{
		Sith::ConOut("Starting up video in full screen mode...\n", ConOut::Pedantic);
		
		hMainWnd = CreateWindow("SITH 2 D3DWND", "Sith 2", WS_POPUP | WS_VISIBLE, 0, 0, Sith::COG::CogLib::GetSetting("video::x_res").GetInt(),
			Sith::COG::CogLib::GetSetting("video::y_res").GetInt(), GetDesktopWindow(), NULL, GetModuleHandle(NULL), NULL);
	}
	else
	{
		Sith::ConOut("Starting up video in windowed mode...\n", ConOut::Pedantic);

		RECT NewClientRect;
		NewClientRect.top = 0;
		NewClientRect.left = 0;
		NewClientRect.right = Sith::COG::CogLib::GetSetting("video::x_res").GetInt();
		NewClientRect.bottom = Sith::COG::CogLib::GetSetting("video::y_res").GetInt();

		AdjustWindowRect(&NewClientRect, WS_OVERLAPPED | WS_SYSMENU | WS_VISIBLE, FALSE);

		hMainWnd = CreateWindow("SITH 2 D3DWND", "Sith 2", WS_OVERLAPPED | WS_SYSMENU | WS_VISIBLE, CW_USEDEFAULT, CW_USEDEFAULT,
			NewClientRect.right - NewClientRect.left, NewClientRect.bottom - NewClientRect.top,
			GetDesktopWindow(), NULL, GetModuleHandle(NULL), NULL);
	};

	if(hMainWnd == NULL)
	{
		MessageBox(GetDesktopWindow(), "Failed to create renderer window\n", "Error", MB_OK | MB_ICONERROR);
		Sith::ConOut("CRITICAL ERROR: Failed to create renderer window.\n", ConOut::Critical);
		Sith::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	};

	ShowWindow(hMainWnd, SW_SHOWDEFAULT);
	UpdateWindow(hMainWnd);

	lpD3D = 0;
	lpD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if(!lpD3D)
	{
		MessageBox(GetDesktopWindow(), "Failed to obtain Direct3D interface\n", "Error", MB_OK | MB_ICONERROR);
		Sith::ConOut("CRITICAL ERROR: Failed to obtain Direct3D interface.\n", ConOut::Critical);
		Sith::Kernel::Get().SendSignal(Sith::KS_QUIT);

		return SE_GENERAL_CRITICAL;
	};

	ZeroMemory(&present_parameters, sizeof(D3DPRESENT_PARAMETERS));

	if(COG::CogLib::GetSetting("video::triplebuffer").GetBool() == true)
	{
		present_parameters.BackBufferCount = 2;
	}
	else
	{
		present_parameters.BackBufferCount = 1;
	};

	present_parameters.MultiSampleType = D3DMULTISAMPLE_NONE;
	present_parameters.MultiSampleQuality = 0;

	present_parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;

	present_parameters.hDeviceWindow = hMainWnd;

	present_parameters.Flags = 0;

	present_parameters.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;

	if(COG::CogLib::GetSetting("video::vsync").GetBool())
	{
        present_parameters.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	}
	else
	{
		present_parameters.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	};

	present_parameters.EnableAutoDepthStencil = FALSE;

	if(COG::CogLib::GetSetting("video::fullscreen").GetBool() == true)
	{
		int bpp = COG::CogLib::GetSetting("video::bitdepth").GetInt();

		if(bpp == 32)
		{
			present_parameters.BackBufferFormat = D3DFMT_X8R8G8B8;
		}
		else if(bpp = 16)
		{
			present_parameters.BackBufferFormat = D3DFMT_R5G6B5;
		}
		else
		{
			MessageBox(GetDesktopWindow(), "Unrecognized bitdepth set; adjust \'video::bitdepth\' and retry.", "Error", MB_OK | MB_ICONERROR);
			ConOut("CRITICAL ERROR: Unrecognized bitdepth set; adjust \'video::bitdepth\' and retry. Success not possible.\n", ConOut::Critical);
			Sith::Kernel::Get().SendSignal(KS_QUIT);

			return SE_GENERAL_CRITICAL;
		};

		present_parameters.Windowed = FALSE;
		present_parameters.BackBufferWidth = COG::CogLib::GetSetting("video::x_res").GetInt();
		present_parameters.BackBufferHeight = COG::CogLib::GetSetting("video::y_res").GetInt();
	}
	else
	{
		present_parameters.Windowed = TRUE;
		present_parameters.BackBufferFormat = D3DFMT_UNKNOWN; // Discard previous format and use desktop display format...
	};

	LastError = lpD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hMainWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &present_parameters, &Device);

	if(FAILED(LastError))
	{
		MessageBox(GetDesktopWindow(), "Failed to create Direct3D device", "Error", MB_OK | MB_ICONERROR);
		ConOut("CRITICAL ERROR: Failed to create Direct3D device. One or more settings in namespace \'video\' may be invalid, or your display device may not support Sith 2.\n",
			ConOut::Critical);
		Sith::Kernel::Get().SendSignal(KS_QUIT);

		return SE_GENERAL_CRITICAL;
	};

	ClearEveryFrame = COG::CogLib::GetSetting("video::clear").GetBool();
	ClearColor = 0x00000000;

	D3DADAPTER_IDENTIFIER9 d3dai;

	lpD3D->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0x0, &d3dai);

	Device->GetDeviceCaps(&device_capabilities);

	if((device_capabilities.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL) || !(device_capabilities.TextureCaps & D3DPTEXTURECAPS_POW2))
	{
		DeviceSupportsBMs = true;
	}
	else
	{
		DeviceSupportsBMs = false;
		Sith::ConOut("WARNING: Video hardware does not support BMs. User interface elements may not be rendered correctly.\n", ConOut::Warning);
	};

	Sith::ConOut("Using: " + std::string(d3dai.Description) + "\n", ConOut::Standard);

	Sith::ConOut("Video startup successful.\n\n", ConOut::Standard);
    
	return SE_OK;
};

Sith::SITHERROR Sith::Win32::Renderer::Process()
{
	while(PeekMessage(&Msg, hMainWnd, 0U, 0U, PM_REMOVE) > 0)
	{
		TranslateMessage(&Msg);
		DispatchMessage(&Msg);
	};

	if(ClearEveryFrame)
	{
		LastError = Device->Clear(0, NULL, D3DCLEAR_TARGET, ClearColor, 1.0f, 0);

		if(FAILED(LastError))
		{
			ConOut("WARNING: Failed to clear framebuffer.\n", ConOut::Warning);
		};
	};

	LastError = Device->BeginScene();
	if(FAILED(LastError))
	{
		ConOut("ERROR: Could not begin rendering process. BeginScene call failed.\n", ConOut::Error);
	};

	// TODO: ACTUAL RENDERING
	D3DJKLVertex Buffer[4];


	// Render User Interface:
	Device->SetFVF(D3DUIVertex_FVF);

	DrawUIWidget(Sith::UI::UIMain::Get().RootWidget);

	LastError = Device->EndScene();
	if(FAILED(LastError))
	{
		ConOut("ERROR: Could not end rendering process. EndScene call failed.\n", ConOut::Error);
	};

	LastError = Device->Present(NULL, NULL, NULL, NULL);
	if(FAILED(LastError))
	{
		ConOut("ERROR: Could not flip buffers; Present call failed.\n", ConOut::Error);
	};

	return Sith::SE_OK;
};

Sith::SITHERROR Sith::Win32::Renderer::DrawUIWidget(Sith::UI::BaseWidget &Widget)
{
	if(!Widget.IsVisible())
	{
		return SE_OK;
	};

	if(FAILED(Device->SetTexture(0, ResourceManager::Get().GetMaterial("06wstn1a.mat", "dflt.cmp")->GetHBM().TextureData)))
	{
		ConOut("WARNING: For some reason it won't let me set the texture to dflt.mat :( :( :(.\n", ConOut::Warning);
	};

	Device->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	Device->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	Device->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);

	D3DUIVertex UIVertex[4];

	UI::UIRect Target = Widget.GetRect();

	UIVertex[0].x = (float)(Target.x);					UIVertex[0].y = (float)(Target.y);					UIVertex[0].z = 1.0f; UIVertex[0].rhw = 1.0f; UIVertex[0].tu = 0.0f; UIVertex[0].tv = 0.0f;
	UIVertex[1].x = (float)(Target.width + Target.x);	UIVertex[1].y = (float)(Target.y);					UIVertex[1].z = 1.0f; UIVertex[1].rhw = 1.0f; UIVertex[1].tu = 2.0f; UIVertex[1].tv = 0.0f;
	UIVertex[2].x = (float)(Target.x);					UIVertex[2].y = (float)(Target.height + Target.y);	UIVertex[2].z = 1.0f; UIVertex[2].rhw = 1.0f; UIVertex[2].tu = 0.0f; UIVertex[2].tv = 2.0f;
	UIVertex[3].x = (float)(Target.width + Target.x);	UIVertex[3].y = (float)(Target.height + Target.y);	UIVertex[3].z = 1.0f; UIVertex[3].rhw = 1.0f; UIVertex[3].tu = 2.0f; UIVertex[3].tv = 2.0f;

	UI::UIColor RColor = Widget.GetColor(1);
	UI::UIColor LColor = Widget.GetColor(2);

	switch(Widget.GetColorState())
	{
	case UI::UICS_LRGRADIENT:
		UIVertex[0].color = RColor.RawColor;
		UIVertex[1].color = LColor.RawColor;
		UIVertex[2].color = RColor.RawColor;
		UIVertex[3].color = LColor.RawColor;
		break;

	case UI::UICS_TDGRADIENT:
		UIVertex[0].color = RColor.RawColor;
		UIVertex[1].color = RColor.RawColor;
		UIVertex[2].color = LColor.RawColor;
		UIVertex[3].color = LColor.RawColor;
		break;

	default:
	case UI::UICS_FLAT:
		UIVertex[0].color = RColor.RawColor;
		UIVertex[1].color = RColor.RawColor;
		UIVertex[2].color = RColor.RawColor;
		UIVertex[3].color = RColor.RawColor;
		break;
	};

	LastError = Device->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, UIVertex, sizeof(D3DUIVertex));
	if(FAILED(LastError))
	{
		ConOut("WARNING: Could not render GUI element.\n", ConOut::Warning);
	};

	for(std::list<UI::BaseWidget>::iterator it = Widget.ChildList.begin(); it != Widget.ChildList.end(); ++it)
	{
		DrawUIWidget(*it);
	};

	return SE_OK;
};
