/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "vfile.h"
#include "vdirectory.h"
#include "episode.h"

#include "episode.y.hpp"

extern int EPparse();
extern bool EPSetInputString(const char *Input);
extern bool EPCloseInputString();

Sith::VFS::Episode* GlobalEpisode;

Sith::VFS::Episode::EpisodeElement::EpisodeElement(int _LineNumber, int _CDNumber, int _LevelNumber, ElementType _Type, std::string _FileName,
												   int _LightPower, int _DarkPower, int _GotoA, int _GotoB) : LineNumber(_LineNumber),
												   CDNumber(_CDNumber), LevelNumber(_LevelNumber), Type(_Type), LightPower(_LightPower), DarkPower(_DarkPower),
												   GotoA(_GotoA), GotoB(_GotoB)
{
	FileName = std::string("jkl\\") + _FileName;
	for(unsigned int i = 0; i < _FileName.length(); ++i)
	{
		if(_FileName[i] == '.') break;
		else ShortName.push_back(_FileName[i]);
	};

	return;
};

Sith::VFS::Episode::Episode() : EpisodeContainer("EPISODE:\\")
{
	return;
};

Sith::SITHERROR Sith::VFS::Episode::Load(const char* AsciiResourceBuffer)
{
	GlobalEpisode = this;

	EPSetInputString(AsciiResourceBuffer);

	if(EPparse() == 1)
	{
		EPCloseInputString();
		GlobalEpisode = 0;
		return SE_FILE_CORRUPT;
	}
	else
	{
		EPCloseInputString();
		GlobalEpisode = 0;
		return SE_OK;
	};
};

Sith::SITHERROR Sith::VFS::Episode::AddLevel(int _LineNumber, int _CDNumber, int _LevelNumber, ElementType _Type, std::string _FileName, int _LightPower,
											 int _DarkPower, int _GotoA, int _GotoB)
{
	LevelList.push_back(EpisodeElement(_LineNumber, _CDNumber, _LevelNumber, _Type, _FileName, _LightPower, _DarkPower, _GotoA, _GotoB));
	return SE_OK;
};

Sith::VFS::Episode::EpisodeElement Sith::VFS::Episode::GotoFirstLevel()
{
	CurrentLevel = LevelList.begin();
	return *CurrentLevel;
};

Sith::VFS::Episode::EpisodeElement Sith::VFS::Episode::GotoLine(int NextLineNumber)
{
	// TODO: Add end-game behavior here?
	for(std::list<EpisodeElement>::iterator it = LevelList.begin(); it != LevelList.end(); ++it)
	{
		if(it->LineNumber == NextLineNumber)
		{
			CurrentLevel = it;
			return *CurrentLevel;
		};
	};

	// Still here? That means it doesn't exist! keke. Return start:

	CurrentLevel = LevelList.begin();
	return *CurrentLevel;
};