/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _EPISODE_H_INCLUDED_
#define _EPISODE_H_INCLUDED_

namespace Sith
{
	namespace VFS
	{
		class Episode
		{
		public:
			enum EpisodeType
			{
				SINGLEPLAYER,
				DEATHMATCH,
				CAPTURETHEFLAG,
				UNKNOWN
			};

			enum ElementType
			{
				CUT = 1,
				LEVEL = 2,
				DECIDE = 3,
			};

			class EpisodeElement
			{
			public:
				int LineNumber;
				int CDNumber;
				int LevelNumber;

				ElementType Type;

				std::string FileName;

				std::string ShortName;

				int LightPower;
				int DarkPower;

				int GotoA;
				int GotoB;

				EpisodeElement(int _LineNumber, int _CDNumber, int _LevelNumber, ElementType _Type, std::string _FileName, int _LightPower,
					int _DarkPower, int _GotoA, int _GotoB);
			};

			// ---------- File implementations ------------
			std::string Name;

			std::string ShortName;

			EpisodeType Type;

			std::list<EpisodeElement> LevelList;
			std::list<EpisodeElement>::iterator CurrentLevel;

			VDirectory EpisodeContainer;

			Episode();

			SITHERROR Load(const char* AsciiResourceBuffer);

			SITHERROR AddLevel(int _LineNumber, int _CDNumber, int _LevelNumber, ElementType _Type, std::string _FileName, int _LightPower,
				int _DarkPower, int _GotoA, int _GotoB);

			EpisodeElement GotoFirstLevel();
			EpisodeElement GotoLine(int NextLineNumber);
		};
	};
};

#endif // _EPISODE_H_INCLUDED_