/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

%{
#include "stdafx.h"

#include "vfile.h"
#include "vdirectory.h"
#include "episode.h"

#include "episode.y.hpp"

int EPerror(char *s);
int EPlex(void);

extern char *yytext;

extern int EPlineno;

extern Sith::VFS::Episode* GlobalEpisode;

%}

%union { int Integer; char *String; }

%token <String> EP_IDENTIFIER EP_LITERAL_STRING EP_FILENAME
%token <Integer> EP_INTEGER

%token EP_TYPE EP_SEQ EP_END EP_CUT EP_LEVEL EP_DECIDE EP_COLON

%type <String>	string_literal
%type <Integer> level_type

%start episodefile

%%

string_literal
	: EP_LITERAL_STRING
		{
			std::string StrTmp = std::string($1);
			std::string StrBuf;
			// Translate and output STRING LITERALS:
			int i = 1;
			while(StrTmp[i] != '\"' && StrTmp[i] != '\0')
			{
				if(StrTmp[i] == '\\')
				{
					++i;
					if(StrTmp[i] == 'n')
					{
						StrBuf.push_back('\n');
					}
					else if(StrTmp[i] == '\"')
					{
						StrBuf.push_back('\"');
					}
					else if(StrTmp[i] == 't')
					{
						StrBuf.push_back('\t');
					}
					else if(StrTmp[i] == '\0')
					{
						break;
					}
					
					++i;
					continue;
				}
				else
				{
					StrBuf.push_back(StrTmp[i++]);
					continue;
				};
			};
			$$ = strdup(StrBuf.c_str());
			free($1);
		}
	;

level_type
	: EP_CUT { $$ = 1; }
	| EP_LEVEL { $$ = 2; }
	| EP_DECIDE { $$ = 3; }
	;

level
	: EP_INTEGER EP_COLON EP_INTEGER EP_INTEGER level_type EP_FILENAME EP_INTEGER EP_INTEGER EP_INTEGER EP_INTEGER
			{
				GlobalEpisode->AddLevel($1, $3, $4, (Sith::VFS::Episode::ElementType)$5, $6, $7, $8, $9, $10);
				free($6);
			}
	;

level_list
	: level
	| level_list level
	;

episodefile
	: string_literal EP_TYPE EP_INTEGER EP_SEQ EP_INTEGER level_list EP_END
			{
				GlobalEpisode->Name = $1;
				
				switch($3)
				{
				case 1:
					GlobalEpisode->Type = Sith::VFS::Episode::SINGLEPLAYER;
					break;
				case 2:
					GlobalEpisode->Type = Sith::VFS::Episode::DEATHMATCH;
					break;
				case 8:
					GlobalEpisode->Type = Sith::VFS::Episode::CAPTURETHEFLAG;
					break;
				default:
					GlobalEpisode->Type = Sith::VFS::Episode::UNKNOWN;
				};
				
				free($1);
			};
	;
	
%%

int EPerror(char* s)
{
	extern int EPlineno;
	extern char *EPtext;
	
	std::stringstream ErrorString;
	ErrorString << "ERROR: \"" << std::string(s) << "\" at symbol \"" << std::string(EPtext) << "\" on line " << EPlineno << "\n";
	ConOut(ErrorString.str(), ConOut::Error);
	
	return 1;
};