/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "gamestate.h"
#include "worldstate.h"
#include "di_input.h"
#include "ogl_render.h"
#include "oal_sound.h"
#include "jkl_vis.h"
#include "vmproc.h"
#include "physics.h"

Sith::Gameplay::GameState* Sith::Gameplay::GameState::StaticInterface = 0;

Sith::Gameplay::GameState::GameState() : Task("Game State")
{
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple instances of the game state for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;

	Sith::COG::CogLib::InjectLibraryFunction("map", 1, &map);
	Sith::COG::CogLib::InjectLibraryFunction("ismulti", 0, &isMulti);
	Sith::COG::CogLib::InjectLibraryFunction("isserver", 0, &isServer);
	Sith::COG::CogLib::InjectLibraryFunction("noclip", 0, &noclip);

	Sith::COG::CogLib::InjectLibraryFunction("warp", 1, &Warp);
	Sith::COG::CogLib::InjectLibraryFunction("tele", 1, &Tele);

	Sith::COG::CogLib::InjectLibraryFunction("creatething", 2, &COGCreateThing);
	Sith::COG::CogLib::InjectLibraryFunction("createthingnr", 2, &COGCreateThing);
	Sith::COG::CogLib::InjectLibraryFunction("createthingatpos", 4, &COGCreateThingAtPos);
	Sith::COG::CogLib::InjectLibraryFunction("createthingatposnr", 4, &COGCreateThingAtPos);
	Sith::COG::CogLib::InjectLibraryFunction("destroything", 1, &COGDestroyThing);

	Sith::COG::CogLib::InjectLibraryFunction("surfacecenter", 1, &SurfaceCenter);
	Sith::COG::CogLib::InjectLibraryFunction("getsurfacecenter", 1, &SurfaceCenter);

	Sith::COG::CogLib::InjectLibraryFunction("findsectoratpos", 1, &findsectoratpos);

	Sith::COG::CogLib::InjectLibraryFunction("rand", 0, &rand);

	Sith::COG::CogLib::InjectLibraryFunction("getinv", 2, &GetInv);
	Sith::COG::CogLib::InjectLibraryFunction("setinv", 3, &SetInv);
	Sith::COG::CogLib::InjectLibraryFunction("changeinv", 3, &ChangeInv);
	Sith::COG::CogLib::InjectLibraryFunction("getinvmax", 2, &GetInvMax);
	Sith::COG::CogLib::InjectLibraryFunction("getinvmin", 2, &GetInvMin);
	Sith::COG::CogLib::InjectLibraryFunction("isinvactivated", 2, &GetInvActivated);
	Sith::COG::CogLib::InjectLibraryFunction("isinvavailable", 2, &GetInvAvailable);
	Sith::COG::CogLib::InjectLibraryFunction("setinvactivated", 3, &SetInvActivated);
	Sith::COG::CogLib::InjectLibraryFunction("setinvavailable", 3, &SetInvAvailable);

	Sith::COG::CogLib::InjectLibraryFunction("getdifficulty", 0, &GetDifficulty);

	Sith::COG::CogLib::InjectLibraryFunction("jkprintunistring", 2, &jkprintunistring);

	Sith::COG::CogLib::InjectLibraryFunction("createbackpack", 1, &cogcreatebackpack);
	Sith::COG::CogLib::InjectLibraryFunction("stopkey", 3, &cogstopkey);
	Sith::COG::CogLib::InjectLibraryFunction("playkey", 4, &cogplaykey);
	Sith::COG::CogLib::InjectLibraryFunction("jkstoppovkey", 3, &cogstoppovkey);
	Sith::COG::CogLib::InjectLibraryFunction("jkplaypovkey", 4, &cogplaypovkey);

	Sith::COG::CogLib::InjectLibraryFunction("applyforce", 2, &cogapplyforce);

	Sith::COG::CogLib::InjectLibraryFunction("getcurweapon", 1, &coggetcurweapon);

	Sith::COG::CogLib::InjectLibraryFunction("selectweapon", 2, &dumb);
	Sith::COG::CogLib::InjectLibraryFunction("autoselectweapon", 2, &dumb);
	Sith::COG::CogLib::InjectLibraryFunction("setfirewait", 2, &dumb);
	Sith::COG::CogLib::InjectLibraryFunction("setcurinvweapon", 2, &dumb);
	Sith::COG::CogLib::InjectLibraryFunction("setmountwait", 2, &dumb);
	Sith::COG::CogLib::InjectLibraryFunction("playmode", 2, &dumb);

	Sith::COG::CogLib::InjectLibraryFunction("getcurrentcamera", 0, &coggetcurrentcamera);

	// Hack:
	Sith::COG::CogLib::InjectLibraryFunction("jkendtarget", 0, &coggetcurrentcamera);

	Sith::COG::CogLib::InjectNamespace("game");
	Sith::COG::CogLib::AddSetting("game::difficulty", COG::VMType(1));

	GameIsActive = false;
	DoPlayerPhysics = true;

	// move this wherever you want
	Sith::Win32::Input *input = &Sith::Win32::Input::Get();
	// while mouse input and keyboard input can be bound to the same thing, mouse input
	// should not be scaled up or controlled by frametime
/*	input->BindInput("up", "keylook", 1);
	input->BindInput("down", "keylook", -1);
	input->BindInput("mousey", "look", .15f);
	input->BindInput("left", "keyturn", 1);
	input->BindInput("right", "keyturn", -1);
	input->BindInput("mousex", "turn", .15f);
	input->BindInput("w", "move", 1);
	input->BindInput("s", "move", -1);
	input->BindInput("a", "strafe", 1);
	input->BindInput("d", "strafe", -1);
	input->BindInput("e", "activate", 1);
	input->BindInput("lcontrol", "vertical", -1);
	input->BindInput("space", "vertical", 1);
	input->BindInput("mousez", "vertical", 4);

	input->BindInput("joy0", "move", 1);
	input->BindInput("joyx", "turn", 1);
	input->BindInput("joyy", "move", 1);

	input->BindInput("joyrz", "strafe", 1);

	input->BindInput("joypov0n", "move", 1);
	input->BindInput("joypov0s", "move", -1);
	input->BindInput("joypov0e", "strafe", -1);
	input->BindInput("joypov0w", "strafe", 1);*/

	return;
};

Sith::Gameplay::GameState::~GameState()
{
	return;
};

Sith::Math::Vector Sith::Gameplay::GameState::VecSurfaceCenter(Res::Level::SURFACE *Surface)
{
	if(Gameplay::WorldState::Get().CurrentLevel.isValid())
	{
		Math::Vector pos(0.0f, 0.0f, 0.0f);
		for(int i = 0; i < Surface->VertexCount; ++i)
		{
			pos = Math::VectorAdd(pos, Math::Vector(&Gameplay::WorldState::Get().CurrentLevel->GeovertexList[Surface->VertexList[i].GeovertexIndex].X));
		};
		pos.x = pos.x / Surface->VertexCount;
		pos.y = pos.y / Surface->VertexCount;
		pos.z = pos.z / Surface->VertexCount;
		return pos;
	};

	return Sith::Math::VectorSet(0.0f, 0.0f, 0.0f);
};


void Sith::Gameplay::GameState::GenSurfVis(int Sector, Math::Vector Pos, Math::Vector LookVector, std::vector<Res::Level::SURFACE*> &SurfVis)
{

	std::vector<Res::Level::SURFACE*>::iterator SurfIt;
	for (SurfIt = Gameplay::WorldState::Get().CurrentLevel->SectorList[Sector].SurfaceList.begin(); SurfIt != Gameplay::WorldState::Get().CurrentLevel->SectorList[Sector].SurfaceList.end(); SurfIt++)
	{		// is the surface visible? we might need to adjust this for tolerance
		if ((Math::VectorDot(LookVector, Math::Vector(&(*SurfIt)->Normal.X)) < 0) && (Math::VectorDist(Pos, VecSurfaceCenter(*SurfIt)) < 0.2) )
		{
			// is it an adjoin? do we need to recurse? do i look fat in this?
			if ((*SurfIt)->AdjoinIndex > -1)
			{
				// recurse
				GenSurfVis ((*SurfIt)->AdjoinedSector->SectorNumber, Pos, LookVector, SurfVis);
			}
			// add it to the list of surfaces
			else SurfVis.push_back(*SurfIt);
		}
	}
}

void Sith::Gameplay::GameState::Activated (int ActivateThingNum)
{
	std::vector<Res::Level::SURFACE*> SurfVis;
	if (ActivateThingNum >= Gameplay::WorldState::Get().ThingList.size())
	{
		return;
	}
	Gameplay::Thing *ActivateThing = &(Gameplay::WorldState::Get().ThingList[ActivateThingNum]);

	Math::Vector ActivateThingPos = ActivateThing->Pos;
	Math::Vector ActivateThingLook = Math::VectorSet(0.0f, 0.0f, 0.0f);
	
	// Rotate ActivateThingLook
	float xzscale = cos(ActivateThing->HeadPitch * PIOVER180);
	ActivateThingLook.x = -sin(ActivateThing->Orient.y * PIOVER180) * xzscale;
	ActivateThingLook.y = cos(ActivateThing->Orient.y * PIOVER180) * xzscale;
	ActivateThingLook.z = sin(ActivateThing->HeadPitch * PIOVER180);

	// find surfaces
	GenSurfVis(ActivateThing->Sector, ActivateThingPos, ActivateThingLook, SurfVis);
	for(std::vector<Res::Level::SURFACE*>::iterator SurfIt = SurfVis.begin(); SurfIt != SurfVis.end(); SurfIt++)
	{
		// send activated with sender = (*SurfIt)->SurfaceNumber and source = ActivateThingNum
		COG::VMProcess::Get().SendMessage("activated", COG::VMType(COG::VMTC_SURFACE, (*SurfIt)->SurfaceNumber, -1), 0, COG::VMType(COG::VMTC_THING, ActivateThingNum, -1));
		COG::VMProcess::Get().SendMessage("activate", COG::VMType(COG::VMTC_SURFACE, (*SurfIt)->SurfaceNumber, -1), 0, COG::VMType(COG::VMTC_THING, ActivateThingNum, -1));
	}

	// find things
	int ThingNum = 0;
	for (std::vector<Gameplay::Thing>::iterator ThingIt = Gameplay::WorldState::Get().ThingList.begin(); ThingIt != Gameplay::WorldState::Get().ThingList.end(); ThingIt++)
	{
		if ((Math::VectorDot(ActivateThingLook, Math::VectorSub(ThingIt->Pos, ActivateThingPos)) > 0.0f)
			&& (Math::VectorDist(ActivateThingPos, ThingIt->Pos) < 0.2) )
		{
			// send message with sender = ThingNum and source = ActivateThingNum
			COG::VMProcess::Get().SendMessage("activated", COG::VMType(COG::VMTC_THING, ThingNum, -1), 0, COG::VMType(COG::VMTC_THING, ActivateThingNum, -1));
			COG::VMProcess::Get().SendMessage("activate", COG::VMType(COG::VMTC_THING, ThingNum, -1), 0, COG::VMType(COG::VMTC_THING, ActivateThingNum, -1));
		}
		ThingNum++;
	}
}

int Sith::Gameplay::GameState::GetCameraSector(Sith::Math::Vector NewPlayer, int curSector)
{
	Gameplay::Thing* curPlayer = &Gameplay::WorldState::Get().ThingList[Gameplay::WorldState::Get().LocalPlayerThing];
	Gameplay::Camera* curCamera = &Gameplay::WorldState::Get().LocalCamera;

	Sith::Math::Vector OldPlayer = curPlayer->Pos;

	Math::Vector Rel = Math::VectorSub(NewPlayer, OldPlayer);

	for(std::vector<Res::Level::SURFACE*>::iterator cs = Gameplay::WorldState::Get().CurrentLevel->SectorList[curSector].SurfaceList.begin();
		cs != Gameplay::WorldState::Get().CurrentLevel->SectorList[curSector].SurfaceList.end(); ++cs)
	{
		if((*cs)->AdjoinIndex >= 0)
		{
			Math::Vector nrml((*cs)->Normal.X, (*cs)->Normal.Y, (*cs)->Normal.Z);

			if(Math::VectorDot(nrml, Rel) > 0) continue;

			Math::Vector csvx(Gameplay::WorldState::Get().CurrentLevel->GeovertexList[(*cs)->VertexList.back().GeovertexIndex].X,
				Gameplay::WorldState::Get().CurrentLevel->GeovertexList[(*cs)->VertexList.back().GeovertexIndex].Y,
				Gameplay::WorldState::Get().CurrentLevel->GeovertexList[(*cs)->VertexList.back().GeovertexIndex].Z);

			Math::Vector NewVx = Math::VectorSub(NewPlayer, csvx);
			Math::Vector OldVx = Math::VectorSub(OldPlayer, csvx);

			float NewDist = Math::VectorDot(nrml, NewVx);
			// if it's passed through the plane
			if(NewDist < 0)
			{
				// test to see if it's actually passed through the adjoin...
				// test against the surface
				// get vertex on the plane by subtracting the distance
				if(Physics::isPointWithinSurface((*cs), nrml, Math::VectorSub(NewPlayer, Math::VectorScale(nrml, NewDist))))
				{
					int Sec = GetCameraSector(NewPlayer, (*cs)->AdjoinedSector->SectorNumber);
					// Camera is in a different sector than the player thing;
					if(Sec == curSector) Sec = (*cs)->AdjoinedSector->SectorNumber;
					return Sec;
				};
			}
		};
	};
	return curPlayer->Sector;
};

Sith::SITHERROR Sith::Gameplay::GameState::Process()
{
	static Win32::Timer timer;
	if(Get().GameIsActive)
	{
		Gameplay::Thing* curPlayer = &Gameplay::WorldState::Get().ThingList[Gameplay::WorldState::Get().LocalPlayerThing];
		Gameplay::Camera* curCamera = &Gameplay::WorldState::Get().LocalCamera;

		// temporary control stuff
		Sith::Win32::Input *input = &Sith::Win32::Input::Get();
		Sith::Win32::Sound *sound = &Sith::Win32::Sound::Get();

		float frametime = timer.Query();

		// Reset player thrust to 0 - this is where the input comes from
		curPlayer->Thrust = Math::VectorSet(0.0f, 0.0f, 0.0f);

		// Input:
		if (input)
		{
			float speed = 1;
			float timestep = frametime;
			float piover180 = 0.0174532925f;

			float keylook, look, keyturn, turn, move, strafe, vertical, activate;
			look = input->GetState("look");
			turn = input->GetState("turn");
			keylook = input->GetState("keylook");
			keyturn = input->GetState("keyturn");
			move = input->GetState("move");
			strafe = input->GetState("strafe");
			vertical = input->GetState("vertical");
			activate = input->GetState("activate");

			curPlayer->Orient.y += keyturn * frametime * 180 + turn;
			// pitch is good for camera, bad for kyle
			curCamera->Pitch += keylook * frametime * 120 + look;

			if (curPlayer->Orient.y > 360)
				curPlayer->Orient.y -= 360;
			else if (curPlayer->Orient.y < 0)
				curPlayer->Orient.y += 360;

			if (curCamera->Pitch > 85)
				curCamera->Pitch = 85;
			else if (curCamera->Pitch < -85)
				curCamera->Pitch = -85;

			// set the head pitch to...something
			curPlayer->HeadPitch = curCamera->Pitch;

			//float xzscale = cos(curCamera->Pitch * piover180);
			float xzscale = 1;

			static bool JustJumped;
			static bool JustActivated;
			if(vertical <= 0) JustJumped = false;
			if(activate <= 0) JustActivated = false;

			if(DoPlayerPhysics)
			{
				curPlayer->Thrust.x += (strafe) * curPlayer->MaxThrust;
				curPlayer->Thrust.y += (move) * curPlayer->MaxThrust;

				if((vertical > 0) && (curPlayer->AttachState != Gameplay::Thing::FREE) && (!JustJumped))
				{
					curPlayer->Velocity = Math::VectorAdd(curPlayer->Velocity, Math::VectorSet(0.0f, 0.0f, curPlayer->JumpSpeed));
					curPlayer->AttachState = Gameplay::Thing::FREE;
					JustJumped = true;
				};

				if((activate > 0) && (!JustActivated))
				{
					Activated(Gameplay::WorldState::Get().LocalPlayerThing);
					JustActivated = true;
				};
			}
			else
			{
				curPlayer->Pos.x -= move * sinf(curPlayer->Orient.y * piover180) * speed * timestep * xzscale;
				curPlayer->Pos.y += move * cosf(curPlayer->Orient.y * piover180) * speed * timestep * xzscale;
				// comment out this line, and make xzscale = 1 to remove flying
				//curPlayer->Z += move * sin(curCamera->Pitch * piover180) * speed * timestep;

				curPlayer->Pos.x -= strafe * sinf((curPlayer->Orient.y + 90) * piover180) * speed * timestep;
				curPlayer->Pos.y += strafe * cosf((curPlayer->Orient.y + 90) * piover180) * speed * timestep;

				curPlayer->Pos.z += vertical * speed * timestep;
			};
		};

		Physics::UpdateThings();

		// offset the camera to the thing's correct look position
		curCamera->X = curPlayer->EyeOffset.x + curPlayer->Pos.x;// + sin(curPlayer->Yaw * piover180) * xzscale * .2f;
		curCamera->Y = curPlayer->EyeOffset.y + curPlayer->Pos.y;// - cos(curPlayer->Yaw * piover180) * xzscale * .2f;
		curCamera->Z = curPlayer->EyeOffset.z + curPlayer->Pos.z;// - sin(.5f * curCamera->Pitch * piover180) * .2f;
		curCamera->Yaw = curPlayer->Orient.y;
		curCamera->Roll = 0;

		Res::Level::SECTOR* camSector = &Gameplay::WorldState::Get().CurrentLevel->SectorList[GetCameraSector(Math::Vector(&curCamera->X), curPlayer->Sector)];

		Win32::Renderer::Get().SetCamera(curCamera->X, curCamera->Y, curCamera->Z, curCamera->Pitch, curCamera->Yaw, camSector->SectorNumber, Math::VectorSet(camSector->Tint[0], camSector->Tint[1], camSector->Tint[2]), Math::VectorSet(0.0f, 0.0f, 0.0f), VFS::ResourceManager::Get().GetModel("bryv.3do"));
		Win32::Sound::Get().SetSectorSound(camSector->SoundName, 1.0f, camSector->SoundVolume);
	};

	return SE_OK;
};

int Sith::Gameplay::GameState::CreateThing(Gameplay::Template &tpl, Math::Vector Position, int Sector, bool DoScripts)
{
	int index = 0;
	for(std::vector<Gameplay::Thing>::iterator it = Gameplay::WorldState::Get().ThingList.begin(); it != Gameplay::WorldState::Get().ThingList.end(); ++it)
	{
		if(it->Type == Gameplay::Template::TT_FREE)
		{
			Gameplay::Thing* NewThing;
			NewThing = (Gameplay::Thing*)calloc(1, sizeof(Gameplay::Thing));

			NewThing->CopyTemplate(&tpl);
			NewThing->Pos = Position;
			NewThing->Sector = Sector;
			NewThing->Instanced = true;
			NewThing->Parent = index; // it's its own parent

			memcpy(&(*it), NewThing, sizeof(Gameplay::Thing));
			free(NewThing);

			if(DoScripts)
			{
                if(it->COG.isValid())
				{
					COG::VMProcess::Get().AddScript(it->COG->GetScript(), COG::VMType(COG::VMTC_THING, index, -1));
				};
			};

			return index;
		};

		++index;
	};

	// Still here? Allocate a new thing or three.
	for(int i = 0; i < 64; ++i)
	{
		Gameplay::WorldState::Get().ThingList.push_back(Gameplay::Thing());
		Gameplay::WorldState::Get().ThingList.back().Type = Gameplay::Template::TT_FREE;
	};

	return CreateThing(tpl, Position, Sector);
};

void Sith::Gameplay::GameState::DestroyThing(int ThingNum)
{
	if((ThingNum >= 0) && (ThingNum < Gameplay::WorldState::Get().ThingList.size()))
	{
		Gameplay::WorldState::Get().ThingList[ThingNum].Instanced = false;
		Gameplay::WorldState::Get().ThingList[ThingNum].Type = Gameplay::Template::TT_FREE;
	};

	return;
};

COGLIB Sith::Gameplay::GameState::COGCreateThing(COG::VMType* tpl, COG::VMType *thingpos)
{
	if((thingpos->GetInt() >= 0) && (thingpos->GetInt() < Gameplay::WorldState::Get().ThingList.size()))
	{
		return new COG::VMType(Get().CreateThing(*Gameplay::WorldState::Get().GetTemplate(tpl->GetTpl()), Math::Vector(Gameplay::WorldState::Get().ThingList[thingpos->GetInt()].Pos),
			Gameplay::WorldState::Get().ThingList[thingpos->GetInt()].Sector));
	}
	else return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::GameState::COGCreateThingAtPos(COG::VMType* tpl, COG::VMType *sector, COG::VMType *pos, COG::VMType *ori)
{
	int NewThing = Get().CreateThing(*Gameplay::WorldState::Get().GetTemplate(tpl->GetTpl()), pos->GetVector(), sector->GetInt());
	Gameplay::WorldState::Get().ThingList[NewThing].Orient = ori->GetVector();

	return new COG::VMType(NewThing);
};

COGLIB Sith::Gameplay::GameState::COGDestroyThing(COG::VMType* thingnum)
{
	Get().DestroyThing(thingnum->GetInt());

	return new COG::VMType(COG::VMTC_VOID);
};

// HACK: IS THIS RIGHT?
COGLIB Sith::Gameplay::GameState::SurfaceCenter(COG::VMType* surface)
{
	if(Gameplay::WorldState::Get().CurrentLevel.isValid())
	{
		if((surface->GetInt() >= 0) && (surface->GetInt() < Gameplay::WorldState::Get().CurrentLevel->SurfaceList.size()))
		{
			Math::Vector pos(0.0f, 0.0f, 0.0f);
			for(int i = 0; i < Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].VertexCount; ++i)
			{
				pos = Math::VectorAdd(pos, Math::Vector(&Gameplay::WorldState::Get().CurrentLevel->GeovertexList[Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].VertexList[i].GeovertexIndex].X));
			};

			pos.x = pos.x / Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].VertexCount;
			pos.y = pos.y / Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].VertexCount;
			pos.z = pos.z / Gameplay::WorldState::Get().CurrentLevel->SurfaceList[surface->GetInt()].VertexCount;

			return new COG::VMType(pos);
		};
	};

	return new COG::VMType(Math::VectorSet(0.0f, 0.0f, 0.0f));
};

COGLIB Sith::Gameplay::GameState::map(COG::VMType *MapName)
{
	Get().GameIsActive = false;

	// Flush COGs.
	Sith::COG::VMProcess::Get().EndLevel();

	Sith::VFS::WisePointer<Res::Level> NextLevel = VFS::ResourceManager::Get().GetLevel(MapName->GetString() + ".jkl");

	if(NextLevel.isValid())
	{
        // Set WorldState to new map:
		Gameplay::WorldState::Get().SetLevel(NextLevel);

		// Okay. Now, spawn in new things and sort out walkplayer/spawnpoints:
		Get().Walkplayers.clear();

		for(std::vector<Gameplay::Thing>::iterator it = NextLevel->Things.begin(); it != NextLevel->Things.end(); ++it)
		{
			if(it->Type == Gameplay::Template::TT_PLAYER)
			{
				Get().Walkplayers.push_back(*it);

				Gameplay::WorldState::Get().ThingList.push_back(*it);
				Gameplay::WorldState::Get().ThingList.back().Instanced = false;
				Gameplay::WorldState::Get().ThingList.back().Type = Gameplay::Template::TT_FREE;
			}
			else
			{
				Gameplay::WorldState::Get().ThingList.push_back(*it);
				Gameplay::WorldState::Get().ThingList.back().Instanced = true;

				if(it->COG.isValid())
				{
					COG::VMProcess::Get().AddScript(it->COG->GetScript());
				};
			};
		};

		if(Get().Walkplayers.size() == 0)
		{
			ConOut("ERROR: Level \'" + MapName->GetString() + "\' contains no spawn points. Gameplay is impossible.\n", ConOut::Error);
			Gameplay::WorldState::Get().SetLevel(Sith::VFS::WisePointer<Sith::Res::Level>(0));
			return new COG::VMType(COG::VMTC_VOID);
		};

		// Spawn local player thing:
		Gameplay::Thing NewPlayer(Get().Walkplayers.front());

		// SEND FORTH ALL LEGIONS
		Gameplay::WorldState::Get().LocalPlayerThing = Gameplay::GameState::Get().CreateThing(NewPlayer, NewPlayer.Pos, NewPlayer.Sector, false);

		// Load up kyle.cog:
		COG::VMProcess::Get().AddScript(Gameplay::WorldState::Get().ThingList[Gameplay::WorldState::Get().LocalPlayerThing].COG->GetScript());

		// done spawning local player thing...

		// load items.dat
		VFS::WisePointer<Res::Inventory> ItemsDat = VFS::ResourceManager::Get().GetInventory("items.dat");
		if (ItemsDat.isValid()) {
			Get().BinList = ItemsDat->BinList;
		}

		// Add level cogs:
		for(std::vector<COG::StackMachine>::iterator it = NextLevel->Cogs.begin(); it != NextLevel->Cogs.end(); ++it)
		{
			COG::VMProcess::Get().AddScript(*it);
		};

		// I WILL BREAK HIM
		COG::VMProcess::Get().SendMessage("newplayer", Gameplay::WorldState::Get().LocalPlayerThing, -1, Gameplay::WorldState::Get().LocalPlayerThing);

		Get().GameIsActive = true;		
	}
	else
	{
		ConOut("ERROR: Failed to load level \'" + MapName->GetString() + "\'.\n", ConOut::Error);
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::GameState::Warp(COG::VMType *wplyr)
{
	if(Gameplay::WorldState::Get().CurrentLevel.isValid())
	{
		if((wplyr->GetInt() >= 0) && (wplyr->GetInt() < Get().Walkplayers.size()))
		{
			Gameplay::WorldState::Get().ThingList[Gameplay::WorldState::Get().LocalPlayerThing].Pos = Get().Walkplayers[wplyr->GetInt()].Pos;
			Gameplay::WorldState::Get().ThingList[Gameplay::WorldState::Get().LocalPlayerThing].Sector = Get().Walkplayers[wplyr->GetInt()].Sector;
		};
	};

	return new COG::VMType(COG::VMTC_VOID);
};

COGLIB Sith::Gameplay::GameState::Tele(COG::VMType *surface)
{
	// Activates surface
	COG::VMProcess::Get().SendMessage("touched", COG::VMType(COG::VMTC_SURFACE, surface->GetInt(), -1), 0, COG::VMType(COG::VMTC_THING, WorldState::Get().LocalPlayerThing, -1));

	return new COG::VMType();
};

COGLIB Sith::Gameplay::GameState::findsectoratpos(COG::VMType *Pos)
{
	Math::Vector PosVec = Pos->GetVector();
	float PosArray[3];
	PosArray[0] = PosVec.x;
	PosArray[1] = PosVec.y;
	PosArray[2] = PosVec.z;

	return new COG::VMType(JKL::FindSectorAtPos(Gameplay::WorldState::Get().CurrentLevel, PosArray));
}