/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "jkl_vis.h"
#include "random.h"


#define PIOVER180			0.0174532925f
#define TWOPI				6.283185307f
#define ERRORBIAS			.00001f

namespace Sith
{
	namespace JKL
	{


bool InsidePolygon(const float *p, const float *poly, int vertices)
{
	float angle = 0;
	float v1[3];
	float v2[3];
	for (int i = 0; i < vertices; i++)
	{
		v1[0] = poly[i * 3 + 0] - p[0];
		v1[1] = poly[i * 3 + 1] - p[1];
		v1[2] = poly[i * 3 + 2] - p[2];
		v2[0] = poly[((i + 1) % vertices) * 3 + 0] - p[0];
		v2[1] = poly[((i + 1) % vertices) * 3 + 1] - p[1];
		v2[2] = poly[((i + 1) % vertices) * 3 + 2] - p[2];
		angle += acos((v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2])
			/ (sqrt(v1[0] * v1[0] + v1[1] * v1[1] + v1[2] * v1[2])
				* sqrt(v2[0] * v2[0] + v2[1] * v2[1] + v2[2] * v2[2])));
	}
	if (angle > (.95f * (TWOPI)) )
		return 1;
	return 0;
}

int ClipSurface(const CleaveData &cd, int vertices, float *in, float *&out, float error)
{
	int newcount;
	float *s1;
	float *s2;
	s1 = in;
	for (int p = 0; p < cd.plane.size(); p++)
	{
		s2 = new float [(vertices * 2 + 2) * 3];
		newcount = 0;
		float *vd;
		vd = new float [vertices];
		int reject = 1;
		for (int v = 0; v < vertices; v++)
		{
			vd[v] = cd.plane[p].n[0] * s1[v * 3 + 0] + cd.plane[p].n[1] * s1[v * 3 + 1] + cd.plane[p].n[2] * s1[v * 3 + 2] - cd.plane[p].d;
			if (vd[v] > -error)
				reject = 0;
		}
		if (!reject)
		{
			for (v = 0; v < vertices; v++)
			{
				int i2 = (v + 1) % vertices;

				if (vd[i2] < -error)
				{
					if (vd[v] > error)
					{
						float dist = (-vd[v]) / (vd[i2] - vd[v]);
						s2[newcount * 3 + 0] = s1[v * 3 + 0] + (s1[i2 * 3 + 0] - s1[v * 3 + 0]) * dist;
						s2[newcount * 3 + 1] = s1[v * 3 + 1] + (s1[i2 * 3 + 1] - s1[v * 3 + 1]) * dist;
						s2[newcount * 3 + 2] = s1[v * 3 + 2] + (s1[i2 * 3 + 2] - s1[v * 3 + 2]) * dist;
						newcount++;
					}
				}
				else
				{
					if (vd[i2] > error && vd[v] < -error)
					{
						float dist = (-vd[v]) / (vd[i2] - vd[v]);
						s2[newcount * 3 + 0] = s1[v * 3 + 0] + (s1[i2 * 3 + 0] - s1[v * 3 + 0]) * dist;
						s2[newcount * 3 + 1] = s1[v * 3 + 1] + (s1[i2 * 3 + 1] - s1[v * 3 + 1]) * dist;
						s2[newcount * 3 + 2] = s1[v * 3 + 2] + (s1[i2 * 3 + 2] - s1[v * 3 + 2]) * dist;
						newcount++;
					}

					s2[newcount * 3 + 0] = s1[i2 * 3 + 0];
					s2[newcount * 3 + 1] = s1[i2 * 3 + 1];
					s2[newcount * 3 + 2] = s1[i2 * 3 + 2];
					newcount++;
				}
			}
		}
		delete [] vd;

		vertices = newcount;
		if (s1 != in)
			delete [] s1;
		s1 = s2;
	}
	if (s1 == in)
		return -1;
	out = s1;
	return vertices;
}

int FindSectorAtPos(const VFS::WisePointer<Res::Level> &level, const float *pos)
{
	for (std::vector<Res::Level::SECTOR>::iterator s = level->SectorList.begin(); s != level->SectorList.end(); ++s)
	{
		for (std::vector<Res::Level::SURFACE>::iterator f = &level->SurfaceList[s->FirstSurfaceIndex]; f != &level->SurfaceList[s->FirstSurfaceIndex + s->SurfaceCount]; ++f)
		{
			float d1, d2;
			d1 = f->Normal.X * level->GeovertexList[f->VertexList[0].GeovertexIndex].X
				+ f->Normal.Y * level->GeovertexList[f->VertexList[0].GeovertexIndex].Y
				+ f->Normal.Z * level->GeovertexList[f->VertexList[0].GeovertexIndex].Z;
			d2 = f->Normal.X * pos[0] + f->Normal.Y * pos[1] + f->Normal.Z * pos[2];
			if (d2 - d1 < 0) // behind surface
				break;
		}
		if (f == &level->SurfaceList[s->FirstSurfaceIndex + s->SurfaceCount]) // didn't break, so inside the sector
		{
			return s->SectorNumber;
		}
	}
	return -1;
}

// should also check vs rendering flags, like geo != 0 on an adjoin and no transparency
void VisSector(const VFS::WisePointer<Res::Level> &level, const float *pos, std::vector<unsigned char> &sectorvis, int sector, int previous, int depth, const CleaveData &cd)
{
	int maxdepth = 2000;

	Res::Level::SECTOR *sectorrec = &level->SectorList[sector];
	int surfaces = sectorrec->SurfaceCount;

	if (depth < maxdepth) // limit the maximum sectors to run through
	{
		// for each adjoin, find new cleave surfaces and follow into the next sector
		for (int s = 0; s < surfaces; s++)
		{
			//tjedsurfacerec surface;
			Res::Level::SURFACE *surface = &level->SurfaceList[sectorrec->FirstSurfaceIndex + s];
			// skip non-adjoins
			if (!surface->AdjoinedSector)
				continue;
			// don't trace backwards to the previous sector
			if (surface->AdjoinedSector->SectorNumber == previous)
				continue;


			// simple distance check
			float adjoinnormal[3];
			float adjoinvert[3];
			adjoinvert[0] = level->GeovertexList[surface->VertexList[0].GeovertexIndex].X;
			adjoinvert[1] = level->GeovertexList[surface->VertexList[0].GeovertexIndex].Y;
			adjoinvert[2] = level->GeovertexList[surface->VertexList[0].GeovertexIndex].Z;
			adjoinnormal[0] = surface->Normal.X;
			adjoinnormal[1] = surface->Normal.Y;
			adjoinnormal[2] = surface->Normal.Z;
			float adjoindist = adjoinnormal[0] * adjoinvert[0] + adjoinnormal[1] * adjoinvert[1] + adjoinnormal[2] * adjoinvert[2];
			float lightdist = adjoinnormal[0] * pos[0] + adjoinnormal[1] * pos[1] + adjoinnormal[2] * pos[2];

			// backfacing adjoin check
			if (lightdist < adjoindist)
				continue;

			// if we are on the same plane as the adjoin
			if (lightdist - adjoindist < ERRORBIAS)
			{
				CleaveData newcd = cd;
				newcd.adjoinsc = surface->AdjoinedSector->SectorNumber;

				// create a new position slightly inside the other sector
				float newpos[3];
				newpos[0] = pos[0] - surface->Normal.X * ERRORBIAS;
				newpos[1] = pos[1] - surface->Normal.Y * ERRORBIAS;
				newpos[2] = pos[2] - surface->Normal.Z * ERRORBIAS;
				VisSector(level, newpos, sectorvis, newcd.adjoinsc, sector, depth + 1, newcd);

				continue;
			}



			float *originalsurface;
			float *clippedsurface;
			int vertices = surface->VertexCount;
			int vcount = 0;
			originalsurface = new float [vertices * 3];
			// removes duplicate vertices
			for (int v = 0; v < vertices; v++)
			{
				float vertd[3];
				vertd[0] = level->GeovertexList[surface->VertexList[v].GeovertexIndex].X;
				vertd[1] = level->GeovertexList[surface->VertexList[v].GeovertexIndex].Y;
				vertd[2] = level->GeovertexList[surface->VertexList[v].GeovertexIndex].Z;
				if (v > 0)
				{
					float dist = sqrt((originalsurface[(v - 1) * 3 + 0] - vertd[0]) * (originalsurface[(v - 1) * 3 + 0] - vertd[0])
						+ (originalsurface[(v - 1) * 3 + 1] - vertd[1]) * (originalsurface[(v - 1) * 3 + 1] - vertd[1])
						+ (originalsurface[(v - 1) * 3 + 2] - vertd[2]) * (originalsurface[(v - 1) * 3 + 2] - vertd[2]));
					if (dist > ERRORBIAS)
					{
						originalsurface[vcount * 3 + 0] = vertd[0];
						originalsurface[vcount * 3 + 1] = vertd[1];
						originalsurface[vcount * 3 + 2] = vertd[2];
						vcount++;
					}
				}
				else
				{
					originalsurface[vcount * 3 + 0] = vertd[0];
					originalsurface[vcount * 3 + 1] = vertd[1];
					originalsurface[vcount * 3 + 2] = vertd[2];
					vcount++;
				}
			}
			// take out the last vertex if it is a duplicate of the first
			if (sqrt((originalsurface[0] - originalsurface[(vcount - 1) * 3 + 0]) * (originalsurface[0] - originalsurface[(vcount - 1) * 3 + 0])
						+ (originalsurface[1] - originalsurface[(vcount - 1) * 3 + 1]) * (originalsurface[1] - originalsurface[(vcount - 1) * 3 + 1])
						+ (originalsurface[2] - originalsurface[(vcount - 1) * 3 + 2]) * (originalsurface[2] - originalsurface[(vcount - 1) * 3 + 2]))
						< ERRORBIAS)
			{
				vcount--;
			}
			vertices = vcount;
			if (vcount < 3)
			{
				// too few original vertices
				continue;
			}

			// clip with positive error (favors keeping vertices)
			int clippedcount = ClipSurface(cd, vertices, originalsurface, clippedsurface, ERRORBIAS);
			if (clippedcount == -1)
			{
				clippedsurface = originalsurface;
			}
			else
			{
				vertices = clippedcount;
				delete [] originalsurface;
			}


			CleaveData newcd;

			if (vertices < 3)
			{
				delete [] clippedsurface;
				continue;
			}




			int skipsurface = 0;
			for (v = 0; v < vertices; v++)
			{
				Plane p;

				int i1, i2;
				i1 = v;
				i2 = (v + 1) % vertices;

				float edge[3];
				edge[0] = clippedsurface[i2 * 3 + 0] - clippedsurface[i1 * 3 + 0];
				edge[1] = clippedsurface[i2 * 3 + 1] - clippedsurface[i1 * 3 + 1];
				edge[2] = clippedsurface[i2 * 3 + 2] - clippedsurface[i1 * 3 + 2];
				if (sqrt(edge[0] * edge[0] + edge[1] * edge[1] + edge[2] * edge[2]) < ERRORBIAS)
				{
					do
					{
						v++;
						if (v >= vertices)
							break;
						i1 = v % vertices;
						i2 = (v + 1) % vertices;
						edge[0] = clippedsurface[i2 * 3 + 0] - clippedsurface[i1 * 3 + 0];
						edge[1] = clippedsurface[i2 * 3 + 1] - clippedsurface[i1 * 3 + 1];
						edge[2] = clippedsurface[i2 * 3 + 2] - clippedsurface[i1 * 3 + 2];
					} while (sqrt(edge[0] * edge[0] + edge[1] * edge[1] + edge[2] * edge[2]) < ERRORBIAS);
					if (v >= vertices)
					{
						skipsurface = 1;
						break;
					}
				}

				float lightv1[3];
				lightv1[0] = pos[0] - clippedsurface[i1 * 3 + 0];
				lightv1[1] = pos[1] - clippedsurface[i1 * 3 + 1];
				lightv1[2] = pos[2] - clippedsurface[i1 * 3 + 2];

				p.n[0] = lightv1[1] * edge[2] - lightv1[2] * edge[1];
				p.n[1] = lightv1[2] * edge[0] - lightv1[0] * edge[2];
				p.n[2] = lightv1[0] * edge[1] - lightv1[1] * edge[0];

				float invlen = 1 / sqrt(p.n[0] * p.n[0] + p.n[1] * p.n[1] + p.n[2] * p.n[2]);
				p.n[0] *= invlen;
				p.n[1] *= invlen;
				p.n[2] *= invlen;

				p.d = p.n[0] * pos[0] + p.n[1] * pos[1] + p.n[2] * pos[2];

				newcd.AddPlane(p);
			}

			delete [] clippedsurface;


			if (skipsurface)
			{
				continue;
			}

			// if we are here, the adjoin is visible

			newcd.adjoinsc = surface->AdjoinedSector->SectorNumber;
			// if planes == 0 then we have clipped out the entire light volume
			if (newcd.plane.size() > 0)
			{
				// otherwise continue on
				VisSector(level, pos, sectorvis, newcd.adjoinsc, sector, depth + 1, newcd);
			}
		}
	}

	// if we are in this function, the sector is visible
	// reverse order stuff goes here
	// in order stuff goes at top

	sectorvis[sector] = 1;
}

int VisStart(const std::vector<int> &startsector, const VFS::WisePointer<Res::Level> &level, const float *pos, const float *forward, const float *right, const float *up, const float fov, const float aspect, std::vector<unsigned char> &sectorvis)
{
	std::vector<int> sectorlist = startsector;

	// remove -1 entries
	for (int n = 0; n < sectorlist.size(); n++)
	{
		if (sectorlist[n] == -1)
		{
			sectorlist.erase(&sectorlist[n]);
			n--;
		}
	}
	if (sectorlist.size() == 0)
	{
		// for now set everything to visible
		memset(&sectorvis[0], 1, sizeof(unsigned char) * sectorvis.size());
		return -1;
	}


	CleaveData cd;
	Plane p[5];

	float t = 1 / (tan(fov * PIOVER180 * .5f) * aspect);
	float t2 = 1 / tan(fov * PIOVER180 * .5f);
	// left = forward + right * 1 / (tan(fov * PIOVER180 * .5f) * (float)wsizex / (float)wsizey)
	p[0].n[0] = forward[0] + right[0] * t;
	p[0].n[1] = forward[1] + right[1] * t;
	p[0].n[2] = forward[2] + right[2] * t;
	p[0].d = pos[0] * p[0].n[0] + pos[1] * p[0].n[1] + pos[2] * p[0].n[2];
	cd.AddPlane(p[0]);
	// right = forward - right * 1 / (tan(fov * PIOVER180 * .5f) * (float)wsizex / (float)wsizey)
	p[1].n[0] = forward[0] - right[0] * t;
	p[1].n[1] = forward[1] - right[1] * t;
	p[1].n[2] = forward[2] - right[2] * t;
	p[1].d = pos[0] * p[1].n[0] + pos[1] * p[1].n[1] + pos[2] * p[1].n[2];
	cd.AddPlane(p[1]);
	// bottom = forward + up * 1 / tan(fov * PIOVER180 * .5f)
	p[2].n[0] = forward[0] + up[0] * t2;
	p[2].n[1] = forward[1] + up[1] * t2;
	p[2].n[2] = forward[2] + up[2] * t2;
	p[2].d = pos[0] * p[2].n[0] + pos[1] * p[2].n[1] + pos[2] * p[2].n[2];
	cd.AddPlane(p[2]);
	// top = forward - up * 1 / tan(fov * PIOVER180 * .5f)
	p[3].n[0] = forward[0] - up[0] * t2;
	p[3].n[1] = forward[1] - up[1] * t2;
	p[3].n[2] = forward[2] - up[2] * t2;
	p[3].d = pos[0] * p[3].n[0] + pos[1] * p[3].n[1] + pos[2] * p[3].n[2];
	cd.AddPlane(p[3]);
	// back plane = forward
	p[4].n[0] = forward[0];
	p[4].n[1] = forward[1];
	p[4].n[2] = forward[2];
	p[4].d = pos[0] * p[4].n[0] + pos[1] * p[4].n[1] + pos[2] * p[4].n[2];
	cd.AddPlane(p[4]);

	for (int n = 0; n < sectorlist.size(); n++)
	{
		VisSector(level, pos, sectorvis, sectorlist[n], sectorlist[n], 0, cd);
	}

	return 1;
}

// returns the sector the point is in, or -1 if it is outside of a sector
int FindVis(const std::vector<int> &startsector, const VFS::WisePointer<Res::Level> &level, const float *pos, const float *forward, const float *right, const float *up, const float fov, const float aspect, std::vector<unsigned char> &sectorvis)
{
	sectorvis.clear();
	sectorvis.resize(level->SectorList.size(), 0);

	return VisStart(startsector, level, pos, forward, right, up, fov, aspect, sectorvis);
}


	};
};