/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _JKL_VIS_H_INCLUDED_
#define _JKL_VIS_H_INCLUDED_

namespace Sith
{
	namespace JKL
	{
		struct Plane
		{
			float n[3];
			float d;
		};

		class CleaveData
		{
		public:
			std::vector<Plane> plane;

			int adjoinsc;


			// returns -1 if the point is inside all planes, otherwise the first plane it is outside
			int IsOutside(const float *point)
			{
				for (int n = 0; n < plane.size(); n++)
				{
					float d2 = plane[n].n[0] * point[0] + plane[n].n[1] * point[1] + plane[n].n[2] * point[2];
					if (d2 < plane[n].d - .00001f)
						return n;
				}
				return -1;
			}

			void AddPlane(const Plane &p)
			{
				plane.push_back(p);
			}
			void DeletePlane(int p)
			{
				plane.erase(&plane[p]);
			}

			CleaveData()
			{
				adjoinsc = -1;
			}
		};

		bool InsidePolygon(const float *p, const float *poly, int vertices);
		int FindSectorAtPos(const VFS::WisePointer<Res::Level> &level, const float *pos);
		// out gets allocated inside ClipSurface
		int ClipSurface(const CleaveData &cd, int vertices, float *in, float *&out, float error);
		int FindVis(const std::vector<int> &startsector, const VFS::WisePointer<Res::Level> &level, const float *pos, const float *forward, const float *right, const float *up, const float fov, const float aspect, std::vector<unsigned char> &adjoinvis);
	};
};

#endif