/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _KERNEL_H_INCLUDED_
#define _KERNEL_H_INCLUDED_

namespace Sith
{
	namespace System
	{
		// KernelProc should listen for a QUIT signal
		// - when it gets it, it should stay alive until
		// - all other processes have shut down, then
		// - it should exit gracefully
		class KernelProc : public System::Task
		{
			friend class Kernel;
		private:
			bool QuitMessageRecieved;

			SITHERROR Process();
			SITHERROR Startup();
			SITHERROR MessageHandler(KSIGNAL Signal);

			KernelProc();
			~KernelProc();

			COG::VMHeap* VMNamespace;
		};

		class Kernel
		{
			friend class System::KernelProc;
		private:
			std::list<Task*> TaskList;
			int TaskCount;

			Kernel();
			~Kernel();

		public:
			static inline Kernel& Get()
			{
				static Kernel theKernel;
				return theKernel;
			};

			inline std::list<Task*>::iterator RegisterTask(Task *theTask)
			{
				TaskList.push_back(theTask);
				theTask->SendSignal(KS_START);
				return TaskList.end();
			};

			inline SITHERROR SendSignal(KSIGNAL Signal)
			{
				for(std::list<Task*>::iterator it = TaskList.begin(); it != TaskList.end(); ++it)
				{
					(*it)->SendSignal(Signal);
				};

				return SE_OK;
			};

			SITHERROR Preempt();

			inline SITHERROR Execute()
			{
				SITHERROR ReturnValue = SE_OK;

				while(1)
				{
					if(Preempt() == SE_KERNEL_QUIT)
					{
						return SE_KERNEL_QUIT;
					};
				};

				return SE_OK;
			};

			// COG API BINDINGS ----
			COGLIBAPI lsproc()
			{
				std::stringstream Stream;
				Stream << "The Sith Kernel is currently operating " << (unsigned int)System::Kernel::Get().TaskList.size() << " processes.\n";
				ConOut(Stream.str(), ConOut::Standard);

				for(std::list<Task*>::iterator it = System::Kernel::Get().TaskList.begin(); it != System::Kernel::Get().TaskList.end(); ++it)
				{
					Stream.str("");
					Stream << "\t| " << (*it)->TaskName << "\n";
					ConOut(Stream.str(), ConOut::Standard);
				};

				ConOut("End of list\n\n", ConOut::Standard);

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI quit()
			{
				ConOut("Shutting down...\nSending all tasks KS_TERM... ", ConOut::Standard);
				Sith::System::Kernel::Get().SendSignal(Sith::KS_QUIT);
				ConOut("Done\n", ConOut::Standard);

				return new COG::VMType(COG::VMTC_VOID);
			};
		};
	};
};
#endif // _KERNEL_H_INCLUDED_
