/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _OGL_HWBM_H_INCLUDED
#define _OGL_HWBM_H_INCLUDED

#include <gl\gl.h>

namespace Sith
{
	namespace Win32
	{
		enum HBMFormat
		{
			HBMF_BGRA8, // this is the preferred (fastest) way to upload
			HBMF_BGR8,
			HBMF_RGBA8,
			HBMF_RGB8,
			HBMF_LUMINANCE8,
			HBMF_LUMINANCE8_ALPHA8,
			HBMF_ALPHA8,
			HBMF_COMPRESSED,

			HBMF_BADFORMAT=0
		};

		class HBMGeometry
		{
		public:
			unsigned int xdim;
			unsigned int ydim;

			HBMFormat format;

			bool Pow2;

			void* bitdata;

			HBMGeometry()
			{
				bitdata = 0;
				Pow2 = true;
				return;
			};

			HBMGeometry(unsigned int x, unsigned int y, HBMFormat type, bool pow2)
			{
				xdim = x;
				ydim = y;
				format = type;
				Pow2 = pow2;
				bitdata = 0;
				return;
			};
		};

		class Hardware_Bitmap
		{
		private:
			unsigned int x_res;
			unsigned int y_res;
			HBMFormat format;
			bool Pow2;

			bool Locked;
			void *lock_pointer;

		public:
			std::vector<unsigned int> texid;
			unsigned int diffuse, bump, gloss;

			std::vector<std::string> meta;

			Hardware_Bitmap()
			{
				Locked = false;
				lock_pointer = 0;
				diffuse = 0;
				bump = 0;
				gloss = 0;
				return;
			};

			~Hardware_Bitmap()
			{
				for (int n = 0; n < texid.size(); n++)
				{
					if (texid[n])
					{
						glDeleteTextures(1, &texid[n]);
					}
					texid[n] = 0;
				}

				return;
			};

			inline HBMGeometry GetGeometry()
			{
				return HBMGeometry(x_res, y_res, format, Pow2);
			};

			SITHERROR CreateDDS(FILE *file, char *metaname = 0);

			SITHERROR Create(HBMGeometry* GeoData);
			SITHERROR Lock(HBMGeometry* GeoDataBuffer);
			SITHERROR Unlock();
		};
	};
};			

#endif // _OGL_HWBM_H_INCLUDED