/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "jkl_vis.h"

// if fancy water is on, don't call this with drawwater = true more than once per frame, as the reflections update each time
void Sith::Win32::Renderer::DrawJKLBasic(int startsector, std::vector<unsigned char> sectorvis)
{
	int n;


	std::vector<std::vector<Res::Level::SECTOR>::iterator> sectorlist;
	std::vector<unsigned char> sectorvisited;
	if (startsector != -1) // if we have a starting sector
	{
		sectorlist.push_back(&level->SectorList[startsector]);
		// need to keep track of already visited sectors if we are walking through
		sectorvisited.resize(level->SectorList.size(), 0);
		sectorvisited[startsector] = 1;
	}
	else // otherwise put in every visible sector
	{
		for (n = 0; n < sectorvis.size(); n++)
		{
			if (sectorvis[n]) // visible
			{
				sectorlist.push_back(&level->SectorList[n]);
			}
		}
	}


	memset(clevel.p, 0, sizeof(int) * clevel.drawbins);
	int baseindex, texindex;

	std::vector<std::vector<Res::Level::SURFACE>::iterator> transparentlist;
	std::vector<std::vector<Res::Level::SURFACE>::iterator> parallaxskylist;
	std::vector<std::vector<Res::Level::SURFACE>::iterator> ceilingskylist;

	std::vector<int> transparentthings;
	std::vector<std::vector<int> > sectorthings;
	sectorthings.resize(level->SectorList.size());
	for (n = 0; n < Gameplay::WorldState::Get().ThingList.size(); n++)
	{
		if (Gameplay::WorldState::Get().ThingList[n].Model3D.isValid() && Gameplay::WorldState::Get().ThingList[n].Instanced && Gameplay::WorldState::Get().ThingList[n].Sector != -1)// && !(WorldState->ThingList[n].ThingFlags & 0x10))
			sectorthings[Gameplay::WorldState::Get().ThingList[n].Sector].push_back(n);
	}

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	for (int sn = 0; sn < sectorlist.size(); sn++)
	{
		std::vector<Res::Level::SECTOR>::iterator s = sectorlist[sn];
		sectors_drawn++;

		// draw things
		for (std::vector<int>::iterator t = sectorthings[s->SectorNumber].begin(); t != sectorthings[s->SectorNumber].end(); ++t)
		{
			Gameplay::Thing *thing = &Gameplay::WorldState::Get().ThingList[*t];

			if (thing->Model3D->Transparency)
			{
				transparentthings.push_back(*t);
				continue;
			}

			glPushMatrix();
			glTranslatef(thing->Pos.x, thing->Pos.y, thing->Pos.z);
			glRotatef(thing->Orient.x, 1, 0, 0);
			glRotatef(thing->Orient.y, 0, 0, 1);
			glRotatef(thing->Orient.z, 0, 1, 0);
			Render3do(thing->Model3D, thing->Sector);
			glPopMatrix();
		}

		std::vector<Res::Level::SURFACE>::iterator surface_end = &level->SurfaceList[s->FirstSurfaceIndex + s->SurfaceCount];
		for (std::vector<Res::Level::SURFACE>::iterator f = &level->SurfaceList[s->FirstSurfaceIndex]; f != surface_end; ++f)
		{
			// if adjoined
			if (f->AdjoinedSector)
			{
				// if it is adjoined and the adjoing sector isn't visible, the adjoin can't be
				if (!sectorvis[f->AdjoinedSector->SectorNumber])
					continue;

				// if we have a starting point, add sectors as we find them, if not already visited
				if (startsector != -1)
				{
					if (!sectorvisited[f->AdjoinedSector->SectorNumber])
					{
						sectorvisited[f->AdjoinedSector->SectorNumber] = 1;
						sectorlist.push_back(&level->SectorList[f->AdjoinedSector->SectorNumber]);
					}
				}
			}


			if (!f->GeoMode)
				continue;
			// transparent
			if (f->FaceFlags & 0x2)
			{
				transparentlist.push_back(f);
				continue;
			}
			if (f->SurfaceFlags & 0x400) // ceiling sky
			{
				if (wstencilbits > 0)
				{
					ceilingskylist.push_back(f);
					continue;
				}
			}
			if (f->SurfaceFlags & 0x200) // parallax sky
			{
				if (wstencilbits > 0)
				{
					parallaxskylist.push_back(f);
					continue;
				}
			}

			texindex = f->TextureIndex;
			if (texindex < 0)
				continue;

			baseindex = clevel.surfaceoffset[f->SurfaceNumber];
			for (int v = 2; v < f->VertexCount; v++)
			{
				clevel.drawindex[texindex][clevel.p[texindex]] = baseindex;
				clevel.p[texindex]++;
				clevel.drawindex[texindex][clevel.p[texindex]] = baseindex + v - 1;
				clevel.p[texindex]++;
				clevel.drawindex[texindex][clevel.p[texindex]] = baseindex + v;
				clevel.p[texindex]++;
			}
		}
	}

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, clevel.drawbuffer);
//	glVertexPointer(3, GL_FLOAT, sizeof(CompiledVertex), clevel.vertex[0].pos);
//	glNormalPointer(GL_FLOAT, sizeof(CompiledVertex), clevel.vertex[0].normal);
//	glTexCoordPointer(2, GL_FLOAT, sizeof(CompiledVertex), clevel.vertex[0].tex);
//	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(CompiledVertex), clevel.vertex[0].color);
	glVertexPointer(3, GL_FLOAT, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, pos));
	glNormalPointer(GL_FLOAT, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, normal));
	glTexCoordPointer(2, GL_FLOAT, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, tex));
	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(CompiledVertex), (void *)offsetof(CompiledVertex, color));
	for (n = 0; n < clevel.drawbins; n++)
	{
		if (clevel.p[n] > 0)
		{
			glBindTexture(GL_TEXTURE_2D, level->MatList[n]->GetHBM(0).texid[0]);
			glDrawElements(GL_TRIANGLES, clevel.p[n], GL_UNSIGNED_INT, clevel.drawindex[n]);
		}
	}


	// ceiling sky
	if (ceilingskylist.size() > 0)
	{
		glDisable(GL_TEXTURE_2D);
		glColorMask(0, 0, 0, 0);
		glEnable(GL_STENCIL_TEST);
		// only supporting one 0x400 sky texture for now
		glStencilFunc(GL_ALWAYS, 2, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
		for (unsigned int n = 0; n <ceilingskylist.size(); n++)
		{
			std::vector<Res::Level::SURFACE>::iterator f = ceilingskylist[n];
			glBegin(GL_POLYGON);
			for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
			{
				glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
			}
			glEnd();
		}
		glColorMask(1, 1, 1, 1);
		glDisable(GL_DEPTH_TEST);
		glStencilFunc(GL_EQUAL, 2, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		VFS::WisePointer<Res::Material> mat;
		mat = level->MatList[ceilingskylist[0]->TextureIndex];
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, mat->GetHBM(ceilingskylist[0]->CurrentCel).texid[0]);

		float umod, vmod;
		umod =  level->Header.CeilingSkyOffset[0] / mat->GetHBM(ceilingskylist[0]->CurrentCel).GetGeometry().xdim;
		vmod =  level->Header.CeilingSkyOffset[1] / mat->GetHBM(ceilingskylist[0]->CurrentCel).GetGeometry().ydim;

		float infiniteedge = 2000;
		glColor4f(1, 1, 1, 1);
		glBegin(GL_QUADS);
		glTexCoord2f(-infiniteedge * 0.0625f + umod, infiniteedge * 0.0625f + vmod);
		glVertex3f(-infiniteedge, infiniteedge, level->Header.CeilingSkyZ);
		glTexCoord2f(infiniteedge * 0.0625f + umod, infiniteedge * 0.0625f + vmod);
		glVertex3f(infiniteedge, infiniteedge, level->Header.CeilingSkyZ);
		glTexCoord2f(infiniteedge * 0.0625f + umod, -infiniteedge * 0.0625f + vmod);
		glVertex3f(infiniteedge, -infiniteedge, level->Header.CeilingSkyZ);
		glTexCoord2f(-infiniteedge * 0.0625f + umod, -infiniteedge * 0.0625f + vmod);
		glVertex3f(-infiniteedge, -infiniteedge, level->Header.CeilingSkyZ); // add camera.z?
		glEnd();

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
	}

	// parallax sky
	if (parallaxskylist.size() > 0)
	{
		glDisable(GL_TEXTURE_2D);
		glColorMask(0, 0, 0, 0);
		glEnable(GL_STENCIL_TEST);
		// only supporting one 0x200 sky texture for now
		glStencilFunc(GL_ALWAYS, 1, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
		for (unsigned int n = 0; n < parallaxskylist.size(); n++)
		{
			std::vector<Res::Level::SURFACE>::iterator f = parallaxskylist[n];
			glBegin(GL_POLYGON);
			for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
			{
				glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
			}
			glEnd();
		}
		glColorMask(1, 1, 1, 1);
		glDisable(GL_DEPTH_TEST);
		glStencilFunc(GL_EQUAL, 1, 0xFF);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		VFS::WisePointer<Res::Material> mat = level->MatList[parallaxskylist[0]->TextureIndex];
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, mat->GetHBM(parallaxskylist[0]->CurrentCel).texid[0]);

		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		float umod, vmod;
		umod = (-Camera.yaw / 360 * level->Header.HorizonPixelsPerRev + level->Header.HorizonSkyOffset[0]) / mat->GetHBM(parallaxskylist[0]->CurrentCel).GetGeometry().xdim;
		vmod = (Camera.pitch / 180 * level->Header.HorizonPixelsPerRev + level->Header.HorizonSkyOffset[1]) / mat->GetHBM(parallaxskylist[0]->CurrentCel).GetGeometry().ydim;

		glColor4f(1, 1, 1, 1);
		glBegin(GL_QUADS);
		glTexCoord2f(umod, vmod);
		glVertex3f(-1, -1, 0);
		glTexCoord2f(1 + umod, vmod);
		glVertex3f(1, -1, 0);
		glTexCoord2f(1 + umod, 1 + vmod);
		glVertex3f(1, 1, 0);
		glTexCoord2f(umod, 1 + vmod);
		glVertex3f(-1, 1, 0);
		glEnd();

		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
	}

	// draw transparent 3dos
	if (transparentthings.size() > 0)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		for (int n = transparentthings.size() - 1; n >= 0; n--)
		{
			glPushMatrix();
			glTranslatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Pos.x, Gameplay::WorldState::Get().ThingList[transparentthings[n]].Pos.y, Gameplay::WorldState::Get().ThingList[transparentthings[n]].Pos.z);
			glRotatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Orient.x, 1, 0, 0);
			glRotatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Orient.y, 0, 0, 1);
			glRotatef(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Orient.z, 0, 1, 0);
			Render3do(Gameplay::WorldState::Get().ThingList[transparentthings[n]].Model3D, Gameplay::WorldState::Get().ThingList[transparentthings[n]].Sector);
			glPopMatrix();
		}
		glDisable(GL_BLEND);
	}

	// draw transparent surfaces
	if (transparentlist.size() > 0)
	{
		glAlphaFunc(GL_GREATER, .5f * .352941f);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_ALPHA_TEST);
		glEnable(GL_BLEND);
		for (int n = transparentlist.size() - 1; n >= 0; n--)
		{
			std::vector<Res::Level::SURFACE>::iterator f = transparentlist[n];
			if (!f->GeoMode)
				continue;
			VFS::WisePointer<Res::Material> mat;
			mat = level->MatList[f->TextureIndex];
			glBindTexture(GL_TEXTURE_2D, mat->GetHBM(f->CurrentCel).texid[0]);
			float uscale, vscale;
			uscale = ((float)1 / mat->GetHBM(f->CurrentCel).GetGeometry().xdim) * mat->xscale;
			vscale = ((float)1 / mat->GetHBM(f->CurrentCel).GetGeometry().ydim) * mat->yscale;
			glBegin(GL_POLYGON);
			for (std::vector<Res::Level::SURFACEVERTEX>::iterator v = f->VertexList.begin(); v != f->VertexList.end(); ++v)
			{
				float intensity = 1; // using intensity = 1 for now because it looks too dark otherwise in LEC levels
				glColor4f(intensity, intensity, intensity, .352941f); // 90 / 255
				glTexCoord2f((f->TexShiftU + level->TexvertexList[v->TexvertexIndex].U) * uscale, (f->TexShiftV + level->TexvertexList[v->TexvertexIndex].V) * vscale);
				glVertex3fv(&level->GeovertexList[v->GeovertexIndex].X);
			}
			glEnd();
		}
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
	}


	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
}

void Sith::Win32::Renderer::DrawLevelBasic()
{
		float fov = Sith::COG::CogLib::GetSetting("video::fov").GetFloat();

		// determine sector visibility
		std::vector<unsigned char> camera_vis;
		std::vector<int> sector_vis_start;
		sector_vis_start.push_back(Camera.sector);
		Sith::JKL::FindVis(sector_vis_start, level, &Camera.x, Camera.forward, Camera.right, Camera.up, fov, (float)wsizex / wsizey, camera_vis);

		// draw the level
		DrawJKLBasic(sector_vis_start[0], camera_vis);
}