/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "coglib.h"

Sith::Win32::Renderer* Sith::Win32::Renderer::StaticInterface = 0;

Sith::Win32::Renderer::Renderer() : Task("OpenGL Renderer")
{
	resetting = 0;

	// Set static interface
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple renderer objects for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;

	// Add renderer settings
	Sith::COG::CogLib::InjectNamespace("video");

	// ****** NOTE ******** CHANGE THESE, THEY SHOULD BE AUTOMATICALLY GENERATED
	Sith::COG::CogLib::AddSetting("video::fullscreen", Sith::COG::VMType(false));
	Sith::COG::CogLib::AddSetting("video::x_res", Sith::COG::VMType(int(1024)));
	Sith::COG::CogLib::AddSetting("video::y_res", Sith::COG::VMType(int(768)));
	Sith::COG::CogLib::AddSetting("video::bitdepth", Sith::COG::VMType(int(32)));
	Sith::COG::CogLib::AddSetting("video::depthbits", Sith::COG::VMType(int(24)));
	Sith::COG::CogLib::AddSetting("video::stencilbits", Sith::COG::VMType(int(8)));
	Sith::COG::CogLib::AddSetting("video::vsync", Sith::COG::VMType(true));
	Sith::COG::CogLib::AddSetting("video::clear", Sith::COG::VMType(true));
	Sith::COG::CogLib::AddSetting("video::fov", Sith::COG::VMType(float(70)));
	Sith::COG::CogLib::AddSetting("video::anisotropy", Sith::COG::VMType(int(16)));
	Sith::COG::CogLib::AddSetting("video::renderpath", Sith::COG::VMType(int(0)));


	Sith::COG::CogLib::InjectLibraryFunction("videoreset", 0, &VideoReset);

	return;
};

Sith::Win32::Renderer::~Renderer()
{
	return;
};

void Sith::Win32::Renderer::Shutdown()
{
	for (int n = 0; n < reflect_tex.size(); n++)
	{
		if (reflect_tex[n].texid)
			glDeleteTextures(1, &reflect_tex[n].texid);
	}

    if (fullscreen)
	{
		ChangeDisplaySettings(&dmOriginalScreenSettings, 0);
	}

	if (hRC)
	{
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(hRC);
		hRC = NULL;
	}
	if (hDC)
		ReleaseDC(hMainWnd, hDC);
	if (hMainWnd)
		DestroyWindow(hMainWnd);
	UnregisterClass("SITH 2 OPENGL", GetModuleHandle(NULL));
}

void Sith::Win32::Renderer::ReloadTextures()
{
	if(level.isValid())
	{
		for(std::vector<VFS::WisePointer<Res::Material> >::iterator it = level->MatList.begin(); it != level->MatList.end(); ++it)
		{
			(*it)->LoadFromFile(VFS::VirtualFileSystem::Get().Open(std::string("/mat/" + (*it)->GetName()), false), level->ColormapList[0]);
		};
	};
};

Sith::SITHERROR Sith::Win32::Renderer::Startup()
{
	SITHERROR rval = CreateOGLWindow();
	if (rval != SE_OK)
		return rval;

	InitExtensions();
	// vsync
	if (wglSwapIntervalEXT)
	{
		if (Sith::COG::CogLib::GetSetting("video::vsync").GetBool())
			wglSwapIntervalEXT(1);
		else
			wglSwapIntervalEXT(0);
	}

	float max_anisotropy, anisotropy_level = Sith::COG::CogLib::GetSetting("video::anisotropy").GetInt();
	glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &max_anisotropy);
	if (max_anisotropy > anisotropy_level)
		anisotropy = anisotropy_level;
	else
		anisotropy = max_anisotropy;

	if (Sith::COG::CogLib::GetSetting("video::renderpath").GetInt())
	{
		AdvancedStartup();
	}

	glViewport(0, 0, wsizex, wsizey);

	SetCamera(0, 0, 0, 0, 0, 0, Math::VectorSet(1, 1, 1), Math::VectorSet(0, 0, 0));

	ConOut("Video startup successful.\n\n", ConOut::Standard);

	timer.Query();

	nearclip = .02f;

	ClearEveryFrame = Sith::COG::CogLib::GetSetting("video::clear").GetBool();

	titleupdate = 0;

	return SE_OK;
};