/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "ogl_render.h"
#include "uimain.h"

// add flags for widgets for things such as blending
Sith::SITHERROR Sith::Win32::Renderer::DrawUIWidget(Sith::UI::BaseWidget &Widget, Sith::UI::BaseWidget &Parent)
{
	if(!Widget.IsVisible())
	{
		return SE_OK;
	};

	glDisable(GL_TEXTURE_2D);

	float bmx = 1.0f, bmy = 1.0f;

	if(Widget.GetBitmap().isValid())
	{
		int num = Widget.GetBitmapNumber();
		if((num < 0) || (num >= Widget.GetBitmap()->Images.size()))
		{
			num = 0;
		};

		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, Widget.GetBitmap()->Images[num].OpenGLTexture);
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_TEXTURE_RECTANGLE_ARB);
		glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glAlphaFunc(GL_GREATER, .7f);

		bmx = Widget.GetBitmap()->Images[0].Width;
		bmy = Widget.GetBitmap()->Images[0].Height;
	}
	else
	{
		glDisable(GL_TEXTURE_RECTANGLE_ARB);
	};

	UI::UIRect Target = Widget.GetRect();

	switch(Widget.GetRectState())
	{
	case UI::UIRS_BOTTOMRIGHT:
		Target.x = Parent.GetRect().x + Parent.GetRect().width - Target.width - Target.x;
		Target.y = Parent.GetRect().y + Parent.GetRect().height - Target.height - Target.y;
		break;
	case UI::UIRS_TOPRIGHT:
		Target.x = Parent.GetRect().x + Parent.GetRect().width - Target.width - Target.x;
		Target.y += Parent.GetRect().y;
		break;
	case UI::UIRS_BOTTOMLEFT:
		Target.x += Parent.GetRect().x;
		Target.y = Parent.GetRect().y + Parent.GetRect().height - Target.height - Target.y;
		break;
	case UI::UIRS_TOPLEFT:
	default:
		Target.x += Parent.GetRect().x;
		Target.y += Parent.GetRect().y;
		break;
	};

	OGLUIVertex UIVertex[4];
	UIVertex[0].p[0] = (float)(Target.x);					UIVertex[0].p[1] = (float)(Target.y);					UIVertex[0].p[2] = 0.0f; UIVertex[0].t[0] = 0.0f;							UIVertex[0].t[1] = 0.0f;
	UIVertex[1].p[0] = (float)(Target.width + Target.x);	UIVertex[1].p[1] = (float)(Target.y);					UIVertex[1].p[2] = 0.0f; UIVertex[1].t[0] = ((Widget.GetBitmapState() == UI::UIBS_FIT) ? (bmx) : (Target.width));	UIVertex[1].t[1] = 0.0f;
	UIVertex[2].p[0] = (float)(Target.width + Target.x);	UIVertex[2].p[1] = (float)(Target.height + Target.y);	UIVertex[2].p[2] = 0.0f; UIVertex[2].t[0] = ((Widget.GetBitmapState() == UI::UIBS_FIT) ? (bmx) : (Target.width));	UIVertex[2].t[1] = ((Widget.GetBitmapState() == UI::UIBS_FIT) ? (bmy) : (Target.height));
	UIVertex[3].p[0] = (float)(Target.x);					UIVertex[3].p[1] = (float)(Target.height + Target.y);	UIVertex[3].p[2] = 0.0f; UIVertex[3].t[0] = 0.0f;							UIVertex[3].t[1] = ((Widget.GetBitmapState() == UI::UIBS_FIT) ? (bmy) : (Target.height));

	UI::UIColor RColor = Widget.GetColor(1);
	UI::UIColor LColor = Widget.GetColor(2);

	switch(Widget.GetColorState())
	{
	case UI::UICS_LRGRADIENT:
		*(unsigned __int32 *)UIVertex[0].c = RColor.RawColor;
		*(unsigned __int32 *)UIVertex[1].c = LColor.RawColor;
		*(unsigned __int32 *)UIVertex[2].c = LColor.RawColor;
		*(unsigned __int32 *)UIVertex[3].c = RColor.RawColor;
		break;

	case UI::UICS_TDGRADIENT:
		*(unsigned __int32 *)UIVertex[0].c = RColor.RawColor;
		*(unsigned __int32 *)UIVertex[1].c = RColor.RawColor;
		*(unsigned __int32 *)UIVertex[2].c = LColor.RawColor;
		*(unsigned __int32 *)UIVertex[3].c = LColor.RawColor;
		break;

	default:
	case UI::UICS_FLAT:
		*(unsigned __int32 *)UIVertex[0].c = RColor.RawColor;
		*(unsigned __int32 *)UIVertex[1].c = RColor.RawColor;
		*(unsigned __int32 *)UIVertex[2].c = RColor.RawColor;
		*(unsigned __int32 *)UIVertex[3].c = RColor.RawColor;
		break;
	};

	glDisable(GL_CULL_FACE);
	glBegin(GL_QUADS);
		glColor4ubv(UIVertex[0].c);
		glTexCoord2fv(UIVertex[0].t);
		glVertex3fv(UIVertex[0].p);

		glColor4ubv(UIVertex[1].c);
		glTexCoord2fv(UIVertex[1].t);
		glVertex3fv(UIVertex[1].p);

		glColor4ubv(UIVertex[2].c);
		glTexCoord2fv(UIVertex[2].t);
		glVertex3fv(UIVertex[2].p);

		glColor4ubv(UIVertex[3].c);
		glTexCoord2fv(UIVertex[3].t);
		glVertex3fv(UIVertex[3].p);
	glEnd();

	glDisable(GL_TEXTURE_RECTANGLE_ARB);
	glEnable(GL_TEXTURE_2D);

	if(Widget.GetFont().isValid() && (Widget.GetText().length() > 0))
	{
		DrawUIText(Widget.GetText(), Target.x, Target.y, Widget.GetFont(), Widget.GetFontScale());
	};

	for(std::list<UI::BaseWidget>::iterator it = Widget.ChildList.begin(); it != Widget.ChildList.end(); ++it)
	{
		DrawUIWidget(*it, Widget);
	};

	return SE_OK;
};

Sith::SITHERROR Sith::Win32::Renderer::DrawUIText (std::string Text, unsigned int x, unsigned int y, VFS::WisePointer<Res::Font> Font, float Scale)
{
	unsigned int i, CharOffset, CharWidth, CharHeight;
	glColor4f(1, 1, 1, 1); 
	glBindTexture (GL_TEXTURE_RECTANGLE_ARB, Font->FontBitmap->Images[0].OpenGLTexture);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_RECTANGLE_ARB);
	if(Scale > 1.0f) glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glBegin (GL_QUADS);
		CharHeight = Font->FontBitmap->Images[0].Height;
		for (i = 0; i < Text.length(); i++)
		{
			if (Font->AllCharDefs[Text[i]] == NULL)
			{		// this character is not in the font
				continue;
			}
			CharOffset = Font->AllCharDefs[Text[i]]->Offset;
			CharWidth = Font->AllCharDefs[Text[i]]->Width;
			// all the fonts have a blocky space character for some weird reason; skip it
			if (Text[i] != ' ')
			{
				glTexCoord2f(CharOffset, 0);
				glVertex2f(x, y);
				glTexCoord2f(CharOffset, CharHeight);
				glVertex2f(x, y + (CharHeight * Scale));
				glTexCoord2f(CharOffset + CharWidth, CharHeight);
				glVertex2f(x + (CharWidth * Scale), y + (CharHeight * Scale));
				glTexCoord2f(CharOffset + CharWidth, 0);
				glVertex2f(x + (CharWidth * Scale), y);
			}
			x += (CharWidth * Scale);
		}
	glEnd ();
	glDisable(GL_TEXTURE_RECTANGLE_ARB);
	glEnable(GL_TEXTURE_2D);

	return SE_OK;
}