/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "resman.h"

Sith::Res::Particle::Particle (std::string FileName) : Sith::VFS::Resource(FileName) {
	return;
}

Sith::SITHERROR Sith::Res::Particle::LoadFromFile(const char* FileBuffer)
{
	CurrentOffset = 0;
	LineNumber = 0;
	EndOfStream = false;
	Buffer = FileBuffer;

	unsigned int i;

	BufferLen = strlen(FileBuffer);

	// Start parsing
	// Find the HEADER section
	SkipToNextSection();
	// Is this the header section?
	CurrentOffset += 9;
	if (GetWordToken() != "header") {
		ConOut("ERROR: First section in PAR file is not HEADER\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}

	if (GetWordToken() != "par") {
		ConOut("ERROR: PAR not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	Version = GetFloatToken();
	
	if (GetWordToken() != "size") {
		ConOut("ERROR: SIZE not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	Size = GetFloatToken();

	if (GetWordToken() != "material") {
		ConOut("ERROR: PAR MATERIAL not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	MaterialName = GetFilenameToken();
	// try to open the mat
	Material = VFS::ResourceManager::Get().GetMaterial (MaterialName, "dflt.cmp");
	if (!Material.isValid()) {
		ConOut("ERROR: PAR MATERIAL could not be loaded\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}

	if (GetWordToken() != "lightingmode") {
		ConOut("ERROR: LIGHTINGMODE not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	LightingMode = GetIntegerToken();

	// Find the GEOMETRYDEF section
	SkipToNextSection();
	// Is this the GEOMETRY section?
	CurrentOffset += 9;
	if (GetWordToken() != "geometrydef") {
		ConOut("ERROR: Second section in PAR file is not GEOMETRYDEF\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	SkipToNewLine();

	if (GetWordToken() != "radius") {
		ConOut("ERROR: RADIUS not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	Radius = GetFloatToken();

	if (GetWordToken() != "insert" || GetWordToken() != "offset") {
		ConOut("ERROR: INSERT OFFSET not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	Offset.x = GetFloatToken();
	Offset.y = GetFloatToken();
	Offset.z = GetFloatToken();

	if (GetWordToken() != "vertices") {
		ConOut("ERROR: VERTICES not found in PAR\n", ConOut::Error);
		return Sith::SE_GENERAL_ERROR;
	}
	Vertices = GetIntegerToken();

	Points.resize (Vertices);

	for (i = 0; i < Vertices; ++i) {
		if (GetIntegerToken() != i) {
			ConOut("ERROR: Incorrect point index in PAR\n", ConOut::Error);
			return Sith::SE_GENERAL_ERROR;
		}
		GetPuncToken();
		Points[i].x = GetFloatToken();
		Points[i].y = GetFloatToken();
		Points[i].z = GetFloatToken();
		Points[i].i = GetFloatToken();
	}

	return SE_OK;
};