/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "resman.h"

Sith::VFS::ResourceManager* Sith::VFS::ResourceManager::StaticInterface = 0;

Sith::VFS::ResourceManager::ResourceManager()
{
	if(StaticInterface)
	{
		ConOut("WARNING: Attempted to create multiple instances of the Resource Manager for a single instance of the engine.\n", ConOut::Warning);
	};
	StaticInterface = this;

	Sith::COG::CogLib::InjectLibraryFunction("ls", 0, &ls);
	Sith::COG::CogLib::InjectLibraryFunction("lsepisode", 0, &lsepisode);
	Sith::COG::CogLib::InjectLibraryFunction("mount", 1, &mount);
	Sith::COG::CogLib::InjectLibraryFunction("umount", 0, &umount);
	Sith::COG::CogLib::InjectLibraryFunction("lsmap", 0, &lsmap);
	Sith::COG::CogLib::InjectLibraryFunction("mapinfo", 1, &mapinfo);

	Sith::COG::CogLib::InjectNamespace("VFS");
	Sith::COG::CogLib::AddSetting("VFS::JKResourceDirectory", Sith::COG::VMType(std::string("Resource")));
	Sith::COG::CogLib::AddSetting("VFS::JKEpisodeDirectory", Sith::COG::VMType(std::string("Episode")));
	Sith::COG::CogLib::AddSetting("VFS::JKPathDirectory", Sith::COG::VMType(Sith::COG::VMTC_VOID));
	return;
};

Sith::VFS::ResourceManager::~ResourceManager()
{
	StaticInterface = 0;
	return;
};

Sith::SITHERROR Sith::VFS::ResourceManager::Startup()
{
	ConOut("Starting up file system...\n", ConOut::Pedantic);

	if(Sith::COG::CogLib::GetSetting("VFS::JKResourceDirectory").GetType() == COG::VMTC_STRING)
	{
		VFS::VirtualFileSystem::Get().InitializeResource(Sith::COG::CogLib::GetSetting("VFS::JKResourceDirectory").GetString());
	}
	else
	{
		ConOut("ERROR: Selected resource path is not a string.\n", ConOut::Error);
	};

	if(Sith::COG::CogLib::GetSetting("VFS::JKPathDirectory").GetType() == COG::VMTC_STRING)
	{
		VFS::VirtualFileSystem::Get().InitializeResource(Sith::COG::CogLib::GetSetting("VFS::JKPathDirectory").GetString());
	};

	if(Sith::COG::CogLib::GetSetting("VFS::JKEpisodeDirectory").GetType() == COG::VMTC_STRING)
	{
		VFS::VirtualFileSystem::Get().InitializeEpisode(Sith::COG::CogLib::GetSetting("VFS::JKEpisodeDirectory").GetString());
	}
	else
	{
		ConOut("ERROR: Selected episode path is not a string.\n", ConOut::Error);
	};

	Sith::COG::CogLib::LoadJKStrings("ui\\jkstrings.uni");
	Sith::COG::CogLib::LoadSithStrings("misc\\sithstrings.uni");

	ConOut("File system initialized\n\n", ConOut::Pedantic);

	return SE_OK;
};

Sith::SITHERROR Sith::VFS::ResourceManager::Process()
{
	return SE_OK;
};


// RESOURCE LOADERS:
Sith::VFS::WisePointer<Sith::Res::Colormap> Sith::VFS::ResourceManager::GetColormap(std::string SystemName)
{
	// search the colormap list for a name matching this one
	for(std::list<Res::Colormap>::iterator it = ColormapList.begin(); it != ColormapList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Colormap>(&(*it));
		};
	};

	// still here? Not loaded yet. So load it dumbass.
	ColormapList.push_back(Res::Colormap(SystemName));

	FILE* FileHandle = VFS::VirtualFileSystem::Get().Open("misc\\cmp\\" + SystemName, false);
	if(!FileHandle)
	{
		ColormapList.pop_back();
		ConOut("ERROR: Colormap file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::Colormap>(); // return null pointer
	};

	if(Failed(ColormapList.back().LoadFromFile(FileHandle)))
	{
		fclose(FileHandle);
		ColormapList.pop_back();
		ConOut("ERROR: Failed to load colormap \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::Colormap>(); // return null pointer
	};

	fclose(FileHandle);
	return VFS::WisePointer<Res::Colormap>(&ColormapList.back());
};

Sith::VFS::WisePointer<Sith::Res::Level> Sith::VFS::ResourceManager::GetLevel(std::string SystemName)
{
	// search the level list for a name matching this one
	for(std::list<Res::Level>::iterator it = LevelList.begin(); it != LevelList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Level>(&(*it));
		};
	};

	// level not in list, load it
	LevelList.push_back(Res::Level(SystemName));
	LevelList.back().AddRef();

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("jkl\\" + SystemName);

	if(!FileBuffer)
	{
		LevelList.pop_back();
		ConOut("ERROR: Level file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::Level>(); // return null ptr
	};

	if(Failed(LevelList.back().LoadFromFile(FileBuffer)))
	{
		delete[] FileBuffer;
		LevelList.pop_back();
		ConOut("ERROR: Failed to load level \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::Level>(); // return null
	};

	delete[] FileBuffer;

	LevelList.back().Release();

	return VFS::WisePointer<Res::Level>(&LevelList.back());
};


Sith::VFS::WisePointer<Sith::Res::COGScript> Sith::VFS::ResourceManager::GetScript(std::string SystemName)
{
	// search the script list for a name matching this one
	for(std::list<Res::COGScript>::iterator it = ScriptList.begin(); it != ScriptList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::COGScript>(&(*it));
		};
	};

	// still here? not loaded yet, asshat.
	ScriptList.push_back(Res::COGScript(SystemName));

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("cog\\" + SystemName);

	if(!FileBuffer)
	{
		ScriptList.pop_back();
		ConOut("ERROR: COG script \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::COGScript>();	// return null
	};

	if(Failed(ScriptList.back().LoadFromFile(FileBuffer)))
	{
		delete[] FileBuffer;
		ScriptList.pop_back();
		ConOut("ERROR: Failed to load script \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::COGScript>();	// return null
	};

	delete[] FileBuffer;
	return VFS::WisePointer<Res::COGScript>(&ScriptList.back());
};

Sith::VFS::WisePointer<Sith::Res::Material> Sith::VFS::ResourceManager::GetMaterial(std::string SystemName, std::string Palette)
{
	// search the material list for a name matching this one
	for(std::list<Res::Material>::iterator it = MaterialList.begin(); it != MaterialList.end(); ++it)
	{
		if((it->GetName() == SystemName) && ((it->GetColormapName() == Palette) || (it->GetColormapName().length() == 0)))
		{
			return VFS::WisePointer<Res::Material>(&(*it));
		};
	};

	// still here? Not loaded yet. So load it lollers.
	MaterialList.push_back(Res::Material(SystemName));

	FILE* FileHandle = VFS::VirtualFileSystem::Get().Open("mat\\" + SystemName, false);
	if(!FileHandle)
	{
		FileHandle = VFS::VirtualFileSystem::Get().Open("3do\\mat\\" + SystemName, false);
	};

	if(!FileHandle)
	{
		MaterialList.pop_back();
		ConOut("WARNING: Material file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Warning);
		return GetMaterial("dflt.mat", "dflt.cmp"); // return null pointer
	};

	if(Failed(MaterialList.back().LoadFromFile(FileHandle, Palette)))
	{
		fclose(FileHandle);
		MaterialList.pop_back();
		ConOut("WARNING: Failed to load material \'" + SystemName + "\'.\n", ConOut::Warning);
		return GetMaterial("dflt.mat", "dflt.cmp"); // return null pointer
	};

	fclose(FileHandle);
	return VFS::WisePointer<Res::Material>(&MaterialList.back());
};

Sith::VFS::WisePointer<Sith::Res::Keyframe> Sith::VFS::ResourceManager::GetKeyframe(std::string SystemName)
{
	// search the keyframe list for a name matching this one
	for(std::list<Res::Keyframe>::iterator it = KeyframeList.begin(); it != KeyframeList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Keyframe>(&(*it));
		};
	};

	// keyframe not in list, load it
	KeyframeList.push_back(Res::Keyframe(SystemName));

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("3do\\key\\" + SystemName);

	if(!FileBuffer)
	{
		KeyframeList.pop_back();
		ConOut("ERROR: Keyframe file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::Keyframe>(); // return null ptr
	};

	if(Failed(KeyframeList.back().LoadFromFile(FileBuffer)))
	{
		delete[] FileBuffer;
		KeyframeList.pop_back();
		ConOut("ERROR: Failed to load keyframe \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::Keyframe>(); // return null
	};

	delete[] FileBuffer;
	return VFS::WisePointer<Res::Keyframe>(&KeyframeList.back());
};


Sith::VFS::WisePointer<Sith::Res::Bitmap> Sith::VFS::ResourceManager::GetBitmap(std::string SystemName)
{
	// search the bitmap list for a name matching this one
	for(std::list<Res::Bitmap>::iterator it = BitmapList.begin(); it != BitmapList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Bitmap>(&(*it));
		};
	};

	// bitmap not in list, load it
	BitmapList.push_back(Res::Bitmap(SystemName));

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("ui\\bm\\" + SystemName);

	if(!FileBuffer)
	{
		BitmapList.pop_back();
		ConOut("ERROR: Bitmap file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::Bitmap>(); // return null ptr
	};

	if(Failed(BitmapList.back().LoadFromFile(FileBuffer)))
	{
		delete[] FileBuffer;
		BitmapList.pop_back();
		ConOut("ERROR: Failed to load bitmap \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::Bitmap>(); // return null
	};

	delete[] FileBuffer;
	return VFS::WisePointer<Res::Bitmap>(&BitmapList.back());
};

Sith::VFS::WisePointer<Sith::Res::Font> Sith::VFS::ResourceManager::GetFont(std::string SystemName)
{
	// search the font list for a name matching this one
	for(std::list<Res::Font>::iterator it = FontList.begin(); it != FontList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Font>(&(*it));
		};
	};

	// font not in list, load it
	FontList.push_back(Res::Font(SystemName));

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("ui\\sft\\" + SystemName);

	if(!FileBuffer)
	{
		FontList.pop_back();
		ConOut("ERROR: Font file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::Font>(); // return null ptr
	};

	if(Failed(FontList.back().LoadFromFile(FileBuffer)))
	{
		delete[] FileBuffer;
		FontList.pop_back();
		ConOut("ERROR: Failed to load font \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::Font>(); // return null
	};

	delete[] FileBuffer;
	return VFS::WisePointer<Res::Font>(&FontList.back());
};

Sith::VFS::WisePointer<Sith::Res::Inventory> Sith::VFS::ResourceManager::GetInventory(std::string SystemName)
{
	// search the nventory list for a name matching this one
	for(std::list<Res::Inventory>::iterator it = InventoryList.begin(); it != InventoryList.end(); ++it)
	{
		if(it->GetName() == SystemName)
		{
			return VFS::WisePointer<Res::Inventory>(&(*it));
		};
	};

	// inventory not in list, load it
	InventoryList.push_back(Res::Inventory(SystemName));

	const char* FileBuffer = VFS::VirtualFileSystem::Get().OpenBuffer("misc\\" + SystemName);

	if(!FileBuffer)
	{
		InventoryList.pop_back();
		ConOut("ERROR: Inventory file \'" + SystemName + "\' was not found in the Virtual File System.\n", ConOut::Error);
		return VFS::WisePointer<Res::Inventory>(); // return null ptr
	};

	if(Failed(InventoryList.back().LoadFromFile(FileBuffer)))
	{
		delete[] FileBuffer;
		InventoryList.pop_back();
		ConOut("ERROR: Failed to load nventory \'" + SystemName + "\'.\n", ConOut::Error);
		return VFS::WisePointer<Res::Inventory>(); // return null
	};

	delete[] FileBuffer;
	return VFS::WisePointer<Res::Inventory>(&InventoryList.back());
};
