/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _RESMAN_H_INCLUDED_
#define _RESMAN_H_INCLUDED_

namespace Sith
{
	namespace VFS
	{
		class ResourceManager
		{
			friend class Sith::System::KernelProc;
		private:
			static ResourceManager* StaticInterface;

			ResourceManager();
			~ResourceManager();

			SITHERROR Process();
			SITHERROR Startup();
			SITHERROR Shutdown();

			std::list<Res::Colormap> ColormapList;
			std::list<Res::Material> MaterialList;
			std::list<Res::Level> LevelList;
			std::list<Res::COGScript> ScriptList;
			std::list<Res::Model> ModelList;
			std::list<Res::Keyframe> KeyframeList;
			std::list<Res::Bitmap> BitmapList;
			std::list<Res::Font> FontList;
			std::list<Res::Inventory> InventoryList;

		public:
			static ResourceManager& Get()
			{
				static ResourceManager rm;
				return rm;
			};

			VFS::WisePointer<Res::Colormap> GetColormap(std::string SystemName);
			VFS::WisePointer<Res::Material> GetMaterial(std::string Shortname, std::string Palette);
			VFS::WisePointer<Res::Level> GetLevel(std::string Shortname);
			VFS::WisePointer<Res::COGScript> GetScript(std::string Shortname);
			VFS::WisePointer<Res::Model> GetModel(std::string SystemName, std::string Palette="dflt.cmp");
			VFS::WisePointer<Res::Keyframe> GetKeyframe (std::string SystemName);
			VFS::WisePointer<Res::Bitmap> GetBitmap (std::string SystemName);
			VFS::WisePointer<Res::Font> GetFont (std::string SystemName);
			VFS::WisePointer<Res::Inventory> GetInventory (std::string SystemName);

			// ******* COG BINDINGS *******
			COGLIBAPI ls()
			{
				VFS::VirtualFileSystem::Get().ls();

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI lsepisode()
			{
				VFS::VirtualFileSystem::Get().lsepisode();

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI mount(COG::VMType* InString)
			{
				VFS::VirtualFileSystem::Get().MountEpisode(InString->GetString());

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI umount()
			{
				VFS::VirtualFileSystem::Get().UnmountEpisode();

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI lsmap()
			{
				VFS::VirtualFileSystem::Get().lsmap();

				return new COG::VMType(COG::VMTC_VOID);
			};

			COGLIBAPI mapinfo(COG::VMType* InString)
			{
				VFS::VirtualFileSystem::Get().mapinfo(InString->GetString());

				return new COG::VMType(COG::VMTC_VOID);
			};
		};
	};
};
#endif // _RESMAN_H_INCLUDED_