/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _STACKMACHINE_H_INCLUDED_
#define _STACKMACHINE_H_INCLUDED_

namespace Sith
{
	namespace COG
	{
		class StackMachine
		{
		public:
			std::vector<Instruction> Code;		// Code array
			int InstructionCount;

			// Adds an instruction to the list and increments the instruction count
			// Really just meant for yyparse()
			inline int AddInstruction(Opcode OPCODE, VMType DATA, int Line)
			{
				Code.push_back(Instruction(OPCODE, DATA, Line));
				return InstructionCount++;
			};

			inline int RemoveInstruction()
			{
				Code.pop_back();
				return --InstructionCount;
			};

			double Pulse;					// Pulse timer - COG recieves pulse message once this expires
			double SleepTime;

			double PulseAccumulator;
			double SleepAccumulator;

			int SleepJump;
			COG::VMType SleepSender;
			COG::VMType SleepSource;
			COG::VMType SleepPar1, SleepPar2, SleepPar3, SleepPar4;
			int SleepSenderID;

			VMType Host;

			VMHeap Heap;					// Local heap

			std::stack<VMType> Stack;		// Execution stack

			VMType ReturnRegister;			// Return value register

			std::string Temp;				// TODO: Temporary debug store for the script name

			StackMachine();
			~StackMachine();

			SITHERROR Parse(std::string Script);

			COG::VMType SendMessage(std::string MessageName, COG::VMType Sender, int OverrideSenderID, COG::VMType Source, COG::VMType Par1=COG::VMType(), COG::VMType Par2=COG::VMType(),
				COG::VMType Par3=COG::VMType(), COG::VMType Par4=COG::VMType());
			COG::VMType Execute(int StartOffset, COG::VMType Sender, int SenderID, COG::VMType Source, COG::VMType Par1, COG::VMType Par2, COG::VMType Par3, COG::VMType Par4);

			inline void Clear()
			{
				Code.clear();
				InstructionCount = 0;
				Heap.Clear();
				
				while(!Stack.empty())
				{
					Stack.pop();
				};

				return;
			};

			StackMachine(const StackMachine& Cpy)
			{
				Code = Cpy.Code;
				InstructionCount = Cpy.InstructionCount;
				Pulse = Cpy.Pulse;
				SleepTime = Cpy.SleepTime;
				PulseAccumulator = Cpy.PulseAccumulator;
				SleepAccumulator = Cpy.SleepAccumulator;

				Host = -1;

				Heap.Clear();
				Heap = VMHeap(Cpy.Heap);

				Temp = Cpy.Temp;

				while(!Stack.empty())
				{
					Stack.pop();
				};

				ReturnRegister = Cpy.ReturnRegister;

				return;
			};
		};
	};
};
#endif // _STACKMACHINE_H_INCLUDED_
