/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _TEMPLATE_H_INCLUDED_
#define _TEMPLATE_H_INCLUDED_

namespace Sith
{
	namespace Gameplay
	{
		class Template
		{
		public:
			std::string TemplateName;

			int ThingFlags;

			int TypeFlags;

			enum THINGTYPE
			{
				TT_FREE = 0,
				TT_CAMERA = 1,
				TT_ACTOR = 2,
				TT_WEAPON = 3,
				TT_DEBRIS = 4,
				TT_ITEM = 5,
				TT_EXPLOSION = 6,
				TT_COG = 7,
				TT_GHOST = 8,
				TT_CORPSE = 9,
				TT_PLAYER = 10,
				TT_PARTICLE = 11
			} Type;

			// Physics:
			enum COLLIDETYPE
			{
				CT_GHOST = 0,
				CT_SPHERE = 1,
				CT_POLY = 2
			} Collide;

			enum MOVETYPE
			{
				MT_PATH = 0,
				MT_PHYSICS = 1,
				MT_NONE = 2
			} Move;
			float MoveSize;
			float Size;

			float AirDrag;
			float StaticDrag;
			float SurfDrag;

			float Buoyancy;
			float Mass;

			Sith::Math::Vector AngVel;

			float MaxRotThrust;
			float MaxRotVel;

			Sith::Math::Vector Velocity;

			float MaxThrust;
			float MaxVel;

			Sith::Math::Vector Orient;

			int PhysFlags;

			float PitchRange;
			float YawRange;

			// Gameplay:
			Sith::VFS::WisePointer<Res::COGScript> COG;

			// 'spodes
			float BlastTime;
			Template* Explode;
			Template* Debris;
			Sith::Math::Vector FlashRGB;

			Template* FleshHit;

			float Force;

			Sith::VFS::WisePointer<Res::Model> Model3D;
			// Sith::VFS::WisePointer<Res::Particle> Particle
			// Sith::VFS::WisePointer<Res::Material> Material
			// Sith::VFS::WisePointer<Res::Sprite> Sprites
			int Count;
			float ElementSize;
			float MinSize;
			float Range;
			float Rate;

			Template* CreateThing;

			float Damage;
			float MinDamage;

			int DamageClass;

			// Sith::VFS::WisePointer<Res::AIClass> Aiclass;
			// Sith::VFS::WisePointer<Res::Puppet> Puppet;
			// Sith::VFS::WisePointer<Res::Soundclass> Soundclass;
			float Chance;
			float Error;
			Sith::Math::Vector EyeOffset;
			Sith::Math::Vector FireOffset;
			float FOV;
			float Health;
			float MaxHealth;
			float Height;
			float JumpSpeed;
			float MaxHeadPitch;
			float MinHeadPitch;

			Template* Weapon;
			Template* Weapon2;

			float Light;
			float MaxLight;
			float LightIntensity;
			Sith::Math::Vector LightOffset;

			float Respawn;
			float Timer;
			float TrailRandAngle;
			float TrailCylRadius;
			Template* TrailThing;

			Template();
			Template(std::string Name);
			Template(std::string Name, Template *Parent);
			void CopyTemplate(Sith::Gameplay::Template *t);
			~Template();
		};
	};
};

#endif // _TEMPLATE_H_INCLUDED_