/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _UIWIDGET_H_INCLUDED_
#define _UIWIDGET_H_INCLUDED_

namespace Sith
{
	namespace UI
	{
		class BaseWidget
		{
		protected:
			UIRect Geometry;
			UIRectState RState;

			UIColorState CSState;
			UIColor Color1;
			UIColor Color2;

			UIBitmapState BSState;
			VFS::WisePointer<Res::Bitmap> Image;
			int ImageNum;

			std::string WindowText;
			VFS::WisePointer<Res::Font> Font;
			float FontScale;

			bool Visible;

		public:
			std::list<BaseWidget> ChildList;

			BaseWidget();

			virtual void SetColorState(UIColorState _State);
			virtual void SetPosition(int x, int y);
			virtual void SetSize(int width, int height);
			virtual void SetRect(UIRect Rect);
			virtual void SetRectState(UIRectState _State);
			virtual void SetColor(UIColor _Color, int ColorNumber = 0);
			virtual void SetBitmapState(UIBitmapState _State);
			virtual void SetBitmapNumber(int BitNum);
			virtual void SetBitmap(std::string BitmapName);
			virtual void SetText(std::string Text);
			virtual void SetFont(std::string FontName);
			virtual void SetFontScale(float Scale);
			virtual void Hide();
			virtual void Show();
			virtual UIRect GetRect();
			virtual UIRectState GetRectState();
			virtual UIColor GetColor(int ColorNumber = 0);
			virtual UIColorState GetColorState();
			virtual UIBitmapState GetBitmapState();
			virtual int GetBitmapNumber();
			virtual VFS::WisePointer<Res::Bitmap> GetBitmap();
			virtual std::string GetText();
			virtual VFS::WisePointer<Res::Font> GetFont();
			virtual float GetFontScale();
			virtual bool IsVisible();
			virtual BaseWidget& CreateChild(BaseWidget Base=BaseWidget());
		};
	};
};

#endif // _UIWIDGET_H_INCLUDED_