/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"

#include "vfile.h"
#include "vdirectory.h"

Sith::VFS::VDirectory::VDirectory(std::string theName) : Name(theName)
{
	return;
};

Sith::VFS::VDirectory* Sith::VFS::VDirectory::GetDirectory(std::string theName)
{
	std::transform(theName.begin(), theName.end(), theName.begin(), tolower);

	for(std::list<VDirectory>::iterator it = SubDirectories.begin(); it != SubDirectories.end(); ++it)
	{
		if(it->Name == theName) return &(*it);
	};
	return 0;
};

Sith::VFS::VDirectory* Sith::VFS::VDirectory::MakeDirectory(std::string theName)
{
	std::transform(theName.begin(), theName.end(), theName.begin(), tolower);

	for(std::list<VDirectory>::iterator it = SubDirectories.begin(); it != SubDirectories.end(); ++it)
	{
		if(it->Name == theName) return &(*it);
	};

	SubDirectories.push_back(VDirectory(theName));
	return &SubDirectories.back();
};

Sith::VFS::VFile* Sith::VFS::VDirectory::GetFile(std::string theName)
{
	std::transform(theName.begin(), theName.end(), theName.begin(), tolower);

	std::string pathbuffer;
	VDirectory *CurrentDirectory = this, *NextDirectory = 0;
	for(unsigned int i = 0; i <= theName.length(); ++i)
	{
		if(theName[i] == '\\' || theName[i] == '/')
		{
			NextDirectory = CurrentDirectory->GetDirectory(pathbuffer);
			if(NextDirectory) CurrentDirectory = NextDirectory;
			pathbuffer.clear();
		}
		else if(theName[i] == '\0')
		{
			for(std::list<VFile>::iterator it = CurrentDirectory->SubFiles.begin(); it != CurrentDirectory->SubFiles.end(); ++it)
			{
				if(it->Name == pathbuffer) return &(*it);
			};
			return 0;
		}
		else pathbuffer.push_back(theName[i]);
	};

	return 0;
};

Sith::VFS::VFile* Sith::VFS::VDirectory::MakeFile(std::string theName, std::string theContainerFile, unsigned int theOffset, unsigned int theLength)
{
	std::transform(theName.begin(), theName.end(), theName.begin(), tolower);

	for(std::list<VFile>::iterator it = SubFiles.begin(); it != SubFiles.end(); ++it)
	{
		if(it->Name == theName) return &(*it);
	};

	SubFiles.push_back(VFile(theName, theContainerFile, theOffset, theLength));
	return &SubFiles.back();
};

int Sith::VFS::VDirectory::ls(std::string PrevName)
{
	int ReturnCount = 0;
	int Batch = 0;

	std::ostringstream Stream;

	Stream << PrevName << Name;
	std::string DPath = Stream.str();
	Stream.str("");

	for(std::list<VFile>::iterator it = SubFiles.begin(); it != SubFiles.end(); ++it)
	{
		Stream << DPath << "\\" << it->Name << "\n";
		ConOut(Stream.str(), ConOut::Standard);
		Stream.str("");
		++ReturnCount;
		++Batch;
		if(Batch == 32)
		{
			// ******************** DANGER DANGER DANGER ***********************
			// - PREEMPT THE LS CALL TO PREVENT THE ENGINE FROM STALLING
			// - DO NOT DO THIS, EVER. ONLY EVER PREEMPT A FUNCTION IF YOU KNOW
			// - WHAT YOU'RE DOING AND EXACTLY WHAT THE CONSEQUENCES ARE.
			// - READ AS "DO NOT USE THE PREEMPTIVE FEATURES UNLESS YOU ARE JON`C".
			// - IF YOU THINK YOU NEED THEM TALK TO JON`C BEFORE YOU USE THEM.
			// - SERIOUSLY, JUST DON'T CALL PREEMPT EVER!!!!
			Sith::System::Kernel::Get().Preempt();
			// Preempting in batches to keep the console moving. If you allow too much to
			// accumulate in console i/o it causes the entire engine to stall for several minutes
			Batch = 0;
		};
	};

	for(std::list<VDirectory>::iterator jt = SubDirectories.begin(); jt != SubDirectories.end(); ++jt)
	{
		ReturnCount += jt->ls(DPath + "\\");
	};

	return ReturnCount;
};