/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "stdafx.h"
#include "vmproc.h"

#include "resman.h"

#include "worldstate.h"

Sith::COG::VMProcess* Sith::COG::VMProcess::StaticInterface = 0;

Sith::COG::VMProcess::VMProcess() : Task("Virtual Machine")
{
	if(StaticInterface)
	{
		ConOut("WARNING: Created multiple instances of the Virtual Machine for a single instance of the engine.\n", ConOut::Warning);
	};

	StaticInterface = this;

	COG::CogLib::InjectLibraryFunction("setpulse", 1, &setpulse);

	COG::CogLib::InjectLibraryFunction("settimer", 1, &settimer);
	COG::CogLib::InjectLibraryFunction("settimerex", 4, &settimerex);
	COG::CogLib::InjectLibraryFunction("killtimerex", 1, &killtimerex);

	COG::CogLib::InjectLibraryFunction("lscog", 0, &lscog);

	return;
};

Sith::COG::VMProcess::~VMProcess()
{
	return;
};

int Sith::COG::VMProcess::AddScript(StackMachine &CopyScript, COG::VMType Host)
{
	ScriptList.push_back(CopyScript);
	ScriptList.back().Host = Host;

	if(Gameplay::WorldState::Get().CurrentLevel.isValid())
	{
		if((Host >= 0) && (ScriptList.back().Heap.Local.HeapCount > 0))
		{
			switch(Host.GetType())
			{
			case COG::VMTC_SECTOR:
				if(Host.GetInt() < Gameplay::WorldState::Get().CurrentLevel->SectorList.size())
				{
					Gameplay::WorldState::Get().CurrentLevel->SectorList[Host.GetInt()].LinkTable.push_back(&ScriptList.back());
				};
				break;

			case COG::VMTC_SURFACE:
				if(Host.GetInt() < Gameplay::WorldState::Get().CurrentLevel->SurfaceList.size())
				{
					Gameplay::WorldState::Get().CurrentLevel->SurfaceList[Host.GetInt()].LinkTable.push_back(&ScriptList.back());
				};
				break;

			case COG::VMTC_THING:
				if(Host.GetInt() < Gameplay::WorldState::Get().ThingList.size())
				{
					Gameplay::WorldState::Get().ThingList[Host.GetInt()].LinkTable.push_back(&ScriptList.back());
				};
				break;

			default:
				break;
			};
		};

		for(int i = 0; i < ScriptList.back().Heap.Local.HeapCount; ++i)
		{
			int IntVal = ScriptList.back().Heap.Get(i, -1).GetInt();

			switch(ScriptList.back().Heap.Get(i, -1).GetType())
			{
			case COG::VMTC_SECTOR:
				if((IntVal >= 0) && (IntVal < Gameplay::WorldState::Get().CurrentLevel->SectorList.size()))
				{
					Gameplay::WorldState::Get().CurrentLevel->SectorList[IntVal].LinkTable.push_back(&ScriptList.back());
				}
				break;
			case COG::VMTC_SURFACE:
				if((IntVal >=0) && (IntVal < Gameplay::WorldState::Get().CurrentLevel->SurfaceList.size()))
				{
					Gameplay::WorldState::Get().CurrentLevel->SurfaceList[IntVal].LinkTable.push_back(&ScriptList.back());
				}
				break;
			case COG::VMTC_THING:
				if((IntVal >= 0) && (IntVal < Gameplay::WorldState::Get().ThingList.size()))
				{
					Gameplay::WorldState::Get().ThingList[IntVal].LinkTable.push_back(&ScriptList.back());
				};
				break;
			default:
				break;
			};
		};
	};

	ScriptList.back().SendMessage("startup", COG::VMType(), Host.GetInt(), COG::VMType());
	return (ScriptList.size() - 1);
};

void Sith::COG::VMProcess::SendMessage(std::string MessageName, Sith::COG::VMType Sender, int OverrideID, Sith::COG::VMType Source, COG::VMType Par1, COG::VMType Par2, COG::VMType Par3, COG::VMType Par4)
{
	// Transform message name:
	std::transform(MessageName.begin(), MessageName.end(), MessageName.begin(), tolower);

	// Special-case this for different types of messages.
	// Or not. Whatever.
	if((MessageName == "activated") || (MessageName == "activate") || (MessageName == "arrived") || (MessageName == "blocked") || (MessageName == "crossed") ||
		(MessageName == "damaged") || (MessageName == "entered") || (MessageName == "touched"))
	{		
		// Cycle through the link lists
		switch(Sender.GetType())
		{
		case COG::VMTC_SECTOR:
			if((Sender.GetInt() >= 0) && (Sender.GetInt() < Gameplay::WorldState::Get().CurrentLevel->SectorList.size()))
			{
				for(std::vector<COG::StackMachine*>::iterator it = Gameplay::WorldState::Get().CurrentLevel->SectorList[Sender.GetInt()].LinkTable.begin();
					it != Gameplay::WorldState::Get().CurrentLevel->SectorList[Sender.GetInt()].LinkTable.end(); ++it)
				{
					(*it)->SendMessage(MessageName, Sender, OverrideID, Source, Par1, Par2, Par3, Par4);
				};
			};
			break;
		case COG::VMTC_SURFACE:
			if((Sender.GetInt() >= 0) && (Sender.GetInt() < Gameplay::WorldState::Get().CurrentLevel->SurfaceList.size()))
			{
				for(std::vector<COG::StackMachine*>::iterator it = Gameplay::WorldState::Get().CurrentLevel->SurfaceList[Sender.GetInt()].LinkTable.begin();
					it != Gameplay::WorldState::Get().CurrentLevel->SurfaceList[Sender.GetInt()].LinkTable.end(); ++it)
				{
					(*it)->SendMessage(MessageName, Sender, OverrideID, Source, Par1, Par2, Par3, Par4);
				};
			};
			break;
		case COG::VMTC_THING:
			if((Sender.GetInt() >= 0) && (Sender.GetInt() < Gameplay::WorldState::Get().ThingList.size()))
			{
				for(std::vector<COG::StackMachine*>::iterator it = Gameplay::WorldState::Get().ThingList[Sender.GetInt()].LinkTable.begin();
					it != Gameplay::WorldState::Get().ThingList[Sender.GetInt()].LinkTable.end(); ++it)
				{
					(*it)->SendMessage(MessageName, Sender, OverrideID, Source, Par1, Par2, Par3, Par4);
				};
			};
			break;
		default:
			break;
		};
	}
	else
	{
		// Add unmanaged/unspecialized messages
		for(std::list<COG::StackMachine>::iterator it = ScriptList.begin(); it != ScriptList.end(); ++it)
		{
			it->SendMessage(MessageName, Sender, OverrideID, Source, Par1, Par2, Par3, Par4);
		};
		return;
	};

	return;
};

Sith::SITHERROR Sith::COG::VMProcess::Process()
{
	dT = theTimer.Query();

	if(Timers.size() > 0)
	{
		for(std::list<TimerData>::iterator it = Timers.begin(); it != Timers.end();)
		{
			it->TimeAccumulator += dT;
			if(it->TimeAccumulator >= it->Time)
			{
				it->vm->SendMessage("timer", COG::VMType(), it->ID, COG::VMType(), it->Par1, it->Par2);
				it = Timers.erase(it);
			}
			else
			{
				++it;
			};
		};
	};

	for(std::list<COG::StackMachine>::iterator it = ScriptList.begin(); it != ScriptList.end(); ++it)
	{
		it->SleepAccumulator += dT;
		if(it->SleepAccumulator > it->SleepTime)
		{
			if(it->SleepTime > 0.0)
			{
                it->SleepAccumulator = 0.0;
				it->SleepTime = 0.0;

				it->Execute(it->SleepJump, it->SleepSender, it->SleepSenderID, it->SleepSource, it->SleepPar1, it->SleepPar2, it->SleepPar3, it->SleepPar4);
			};

			it->PulseAccumulator += dT;

			if((it->PulseAccumulator >= it->Pulse) && (it->Pulse > 0))
			{
				it->PulseAccumulator = 0.0;
				it->SendMessage("pulse", COG::VMType(), 0, COG::VMType());
			};
		};
	};

	return SE_OK;
};