/***************************************************************************
 *   Copyright (C) 2005 by Jonathan Clark                                  *
 *   Copyright (C) 2005 by The Sith 2 Project                              *
 *   captain.cliche@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _VMPROC_H_INCLUDED_
#define _VMPROC_H_INCLUDED_

namespace Sith
{
	namespace COG
	{
		class VMProcess : public System::Task
		{
			friend class System::KernelProc;

		private:
			static VMProcess* StaticInterface;

			VMProcess();
			~VMProcess();

			SITHERROR Process();

			std::list<COG::StackMachine> ScriptList;

			Win32::Timer theTimer;
			double dT;

			struct TimerData
			{
				COG::StackMachine* vm;

				double Time;
				double TimeAccumulator;

				int ID;

				COG::VMType Par1, Par2;

				TimerData(COG::StackMachine* _vm, double _Time, COG::VMType _Par1, COG::VMType _Par2, int _ID)
				{
					vm = _vm;
					Time = _Time;
					Par1 = _Par1;
					Par2 = _Par2;
					ID = _ID;
					TimeAccumulator = 0.0;
				};
			};

			std::list<TimerData> Timers;

		public:
			static inline VMProcess& Get()
			{
				return *StaticInterface;
			};

			void SendMessage(std::string Name, COG::VMType Sender, int OverrideID, COG::VMType Source, COG::VMType Par1=COG::VMType(), COG::VMType Par2=COG::VMType(),
				COG::VMType Par3=COG::VMType(), COG::VMType Par4=COG::VMType());

			inline void EndLevel()
			{
				// Call this at the end of the level - [JC]
				for(std::list<COG::StackMachine>::iterator it = ScriptList.begin(); it != ScriptList.end(); ++it)
				{
					it->SendMessage("shutdown", COG::VMType(COG::VMTC_VOID), 0, COG::VMType(COG::VMTC_VOID));
				};

				Timers.clear();

				ScriptList.clear();
				return;
			};

			int AddScript(COG::StackMachine& CopyScript, COG::VMType Host=COG::VMType(COG::VMTC_VOID));

			COGLIBAPI setpulse(COG::VMType* pulsetime, COG::StackMachine* gsm)
			{
				gsm->Pulse = pulsetime->GetFloat();
				gsm->PulseAccumulator = 0.0;
				return new COG::VMType();
			};

			COGLIBAPI settimer(COG::VMType* timertime, COG::StackMachine* gsm)
			{
				Get().Timers.push_front(TimerData(gsm, timertime->GetFloat(), COG::VMType(), COG::VMType(), 0xFFFFFFFF));
				return new COG::VMType();
			};

			COGLIBAPI settimerex(COG::VMType* timertime, COG::VMType* id, COG::VMType* Par1, COG::VMType* Par2, COG::StackMachine* gsm)
			{
				Get().Timers.push_front(TimerData(gsm, timertime->GetFloat(), &Par1, &Par2, id->GetInt()));
				return new COG::VMType();
			};

			COGLIBAPI killtimerex(COG::VMType* id)
			{
				for(std::list<TimerData>::iterator it = Get().Timers.begin(); it != Get().Timers.end();)
				{
					if(it->ID == id->GetInt())
					{
						it = Get().Timers.erase(it);
					}
					else
					{
						++it;
					};
				};
				return new COG::VMType();
			};

			COGLIBAPI lscog()
			{
				std::ostringstream Stream;
				Stream << "There are " << Get().ScriptList.size() << " scripts successfully running in this level.\n";
				ConOut(Stream.str(), ConOut::Standard);
				return new COG::VMType(COG::VMTC_VOID);
			};
		};
	};
};

#endif // _VMPROC_H_INCLUDED_